/* $Id: interp.c,v 1.60 2002/07/15 22:04:13 aa5779 Exp $ */
/*
	Copyright (C) 2001, Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
/// \define[tlo][\note Top level only]
/// \define[ncc][\error[no current chunk]]
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include FNMATCH_H
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <time.h>
#include <signal.h>
#include "lists.h"
#include "dstring.h"
#include "util.h"
#include "interp.h"
#include "automat.h"
#include "array.h"
#include "table.h"
#include "plist.h"
#include "seq.h"
#include "uds.h"
#ifdef USE_SQL
#include "sqlquery.h"
#endif
#include "bstream.h"
#include "sstream.h"
#include "hstream.h"
#include "pstream.h"
#include "cgistream.h"
#include "trap.h"
#ifdef USE_LTDL
#include "modules.h"
#endif

extern char *nsl_location;
static trap_handler *all_traps;

static void trap_destroy(trap_handler *trap)
{
	free(trap->trapname);
}

static int trap_find(trap_handler *trap, char *trapname)
{
	return !strcmp(trap->trapname, trapname);
}

static int decl_find(declaration *d, char *name)
{
	return !strcmp(d->name, name);
}

typedef int (*fence_condition)(int ch);

static int quote_fc(int ch) { return ch == '\"'; }
static int bquote_fc(int ch) { return ch == '`'; }
static int paren_fc(int ch) { return ch == ')'; }
static int bracket_fc(int ch) { return ch == ']'; }
static int normal_fc(int ch) { return isspace(ch) || ch == ')'; }
static int toplevel_fc(int ch) { return isspace(ch) || ch == '}' || ch == ';'; }
static int zero_fc(int ch) { return !ch; }
static int eol_fc(int ch) { return ch == '\n'; }

chunk *parse_argument(interpreter *i, fence_condition fc, char *start, char **end);

static shared_pool *all_pools;
static procedure *all_procs;
static autoseqdef *all_asds;
static storage *all_storages;
static stream *all_streams;
static interpreter *all_ints;

static procstack *pstack;

procstack *get_pstack(void) { return pstack; }

static int pool_find(shared_pool *pool, char *name)
{
	return !strcmp(pool->name, name);
}

static int var_find(variable *var, char *name)
{
	return !strcmp(var->name, name);
}

static void destroy_variable(variable *var);
static void destroy_pool(shared_pool *pool)
{
	free(pool->name);
	list_delete(pool->vars, (list_destroy_t)destroy_variable, sizeof(*pool->vars));
}

static void destroy_chunk(chunk *ch);
static void destroy_argument(argument *arg)
{
	list_delete(arg->arg, (list_destroy_t)destroy_chunk, sizeof(chunk));
}

static void destroy_branch(branch *br)
{
	list_delete(br->condition, (list_destroy_t)destroy_argument, sizeof(argument));
}

static void bind_chunk(chunk *ch)
{
	ch->lno = set_line_no(-1);
	ch->file = set_current_filename(NULL);
}

static void destroy_chunk(chunk *ch)
{
	switch(ch->type)
	{
		case arg_literal:
			ds_unfix(ch->x.literal);
			break;
		case arg_variable:
			free(ch->x.var.name);
			break;
		case arg_embedded:
			list_delete(ch->x.func.f.emb, (list_destroy_t)destroy_chunk, sizeof(chunk));
			list_delete(ch->x.func.args, (list_destroy_t)destroy_argument, sizeof(argument));
			break;
		case arg_funcall:
			free(ch->x.func.f.name);
		case arg_builtin:
			list_delete(ch->x.func.args, (list_destroy_t)destroy_argument, sizeof(argument));
			break;
		case arg_statement:
			list_delete(ch->x.s.ctl->header, (list_destroy_t)destroy_argument, sizeof(argument));
			list_delete(ch->x.s.ctl->branches, (list_destroy_t)destroy_branch, sizeof(branch));
			FREE(ch->x.s.ctl);
			break;
		case arg_parameter:
			break;
		default:
			recover_error("internal error at destroy_chunk");
	}
}

static void destroy_value(value *val)
{
	ds_unfix(val->val);
}

static void destroy_variable(variable *var)
{
	list_delete(var->vals, (list_destroy_t)destroy_value, sizeof(value));
	free(var->name);
}

static void destroy_ctl(ctlstack *ctl)
{
	if(ctl->master->x.s.statement == s_foreach)
	{
		if(ctl->x.loop.hangon && ctl->x.loop.param)
			ctl->x.loop.hangon->class->enditer(NULL, ctl->x.loop.hangon, &ctl->x.loop.param);
	}
	else if(ctl->master->x.s.statement == s_switch)
		ds_unfix(ctl->x.condition);
}

static int local_sto_find(shared_int *si, procstack *frame)
{
	return ((storage *)si->obj)->local_to == frame;
}


static void destroy_execstate(execstack *state) 
{
	if(state->buffer)
		ds_unfix(state->buffer);
	list_delete(state->out_vals, (list_destroy_t)destroy_value, sizeof(execstack));
}

static void destroy_to_remove(to_remove_list *trl)
{
	ds_unfix(trl->key);
}

static int to_remove_find(to_remove_list *trl, char *key)
{
	return !strcmp(DS_BODY(trl->key), key);
}


static void destroy_decl(declaration *decl)
{
	free(decl->name);
}

static void destroy_procframe(procstack *frame)
{
	if(frame->cur_pool)
		frame->cur_pool->refcnt--;
	list_delete(frame->state, (list_destroy_t)destroy_execstate, sizeof(execstack));
	list_delete(frame->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
	while(frame->vars != frame->start_vars)
		frame->vars = list_pop(frame->vars, (list_destroy_t)destroy_variable, sizeof(variable));
	list_delete(frame->args, (list_destroy_t)destroy_value, sizeof(value));
	frame->where->stos = list_remove_all(frame->where->stos, frame, (list_predicate_t)local_sto_find, (list_destroy_t)destroy_shared_int, sizeof(shared_int));
}


static void so_destroy_storage(storage *s)
{
	s->class->destroy(NULL, s);
	list_delete(s->to_remove, (list_destroy_t)destroy_to_remove, sizeof(to_remove_list));
	destroy_so(&s->sh);
}

static void so_destroy_stream(stream *s)
{
	s->class->close(NULL, s);
	destroy_so(&s->sh);
}

static void so_destroy_aseq(autoseqdef *asd)
{
	remove_autoseq(asd->seq);
	destroy_so(&asd->sh);
}

static shared_object_class soc_table[] = {{"storage", (void *)&all_storages, (void *)so_destroy_storage, sizeof(storage), offsetof(storage, sh), offsetof(interpreter, stos)},
							  {"stream", (void *)&all_streams, (void *)so_destroy_stream, sizeof(stream), offsetof(stream, sh), offsetof(interpreter, streams)},
							  {"autoseq", (void *)&all_asds, (void *)so_destroy_aseq, sizeof(autoseqdef), offsetof(autoseqdef, sh), offsetof(interpreter, aseqs)},
							  {"proc", (void *)&all_procs, NULL, sizeof(procedure), offsetof(procedure, sh), offsetof(interpreter, procs)}, // Attention: no destroyer for procedure!
							  {"subint", (void *)&all_ints, NULL, sizeof(interpreter), offsetof(interpreter, sh), offsetof(interpreter, subints)},
							  {NULL}
							};

shared_object_class *get_so_class(char *cls_name)
{
	shared_object_class *iter;

	for(iter = soc_table; iter->id; iter++)
	{
		if(!strcmp(iter->id, cls_name))
			return iter;
	}
	recover_error("unknown shared object class: %s", cls_name);
	return NULL; // never reached
}

chunk *create_literal(dstring *literal)
{
	chunk *nc = ALLOC(chunk);
	bind_chunk(nc);
	nc->type = arg_literal;
	nc->x.literal = ds_fix(literal);
	return nc;
}

chunk *create_parameter(int paramno)
{
	chunk *nc = ALLOC(chunk);
	bind_chunk(nc);
	nc->type = arg_parameter;
	nc->x.paramno = paramno;
	return nc;
}

chunk *create_bi_call(instruction_t action, ...)
{
	chunk *nc = ALLOC(chunk), *arg;
	va_list args;
	argument *na;
	
	va_start(args, action);
	nc->type = arg_builtin;
	bind_chunk(nc);
	nc->x.func.f.bi = action;
	for(;;)
	{
		arg = va_arg(args, chunk *);
		if(!arg)
			break;
		na = ALLOC(argument);
		na->arg = arg;
		nc->x.func.args = list_append(nc->x.func.args, na);
	}
	va_end(args);
	return nc;
}

value *copy_values(value *vals)
{
	value *nl = NULL;
	value *tmp;
	
	for(; vals; vals = list_next(vals))
	{
		tmp = ALLOC(value);
		tmp->val = ds_fix(vals->val);
		nl = list_add(nl, tmp);
	}
	return list_reversip(nl);
}

#ifdef ENABLE_REF
static variable *dereference_var(char *name)
{
	return *name == '#' && name[1] == '#' ? (void *)strtoul(name + 2, NULL, 16) : NULL;
}
#else
#define dereference_var(name) (NULL)
#endif

dstring *get_shared_value(char *poolname, char *name)
{
	shared_pool *p = list_find(all_pools, poolname, (list_predicate_t)pool_find);
	variable *v;
	if(!p)
		return NULL;
	v = list_find(p->vars, name, (list_predicate_t)var_find);
	if(!v)
		return NULL;
	if(v->readonly)
		recover_error("attempt to modify a readonly variable");
	return v->vals->val;
}


static variable *get_var(interpreter *i, char *name)
{
	variable *var = dereference_var(name);
	if(var)
		return var;
	if(pstack)
	{
		var = list_find(pstack->vars, name, (list_predicate_t)var_find);
		if(!var && pstack->cur_pool)
			var = list_find(pstack->cur_pool->vars, name, (list_predicate_t)var_find);
	}
	if(!var)
		var = list_find(i->globvars, name, (list_predicate_t)var_find);
	return var;
}

variable *safe_get_var(interpreter *i, char *name, int writing)
{
	variable *v = get_var(i, name);
	if(!v)
		recover_error("variable '%s' undefined", name);
	if(v->readonly && writing)
		recover_error("attempt to modify a readonly variable %s", name);
	return v;
}

static dstring *get_variable(interpreter *i, int is_tag, char *name)
{
	if(is_tag)
	{
		if(!i->this_chunk)
			recover_error("no input tag");
		if(HTML_CHUNK_TYPE(i->this_chunk) == hct_char)
		{
			if(!*name)
				return ds_create(NULL);
			if(!strcasecmp("value", name))
				return ds_createch(get_html_char(i->this_chunk));
			else if(!strcasecmp("ivalue", name))
				return ds_fromint(get_html_char(i->this_chunk), 10, 0);
			else if(*name == '#' && !name[1])
				return ds_fromint(2, 10, 0);
			else if(*name == '*' && !name[1])
			{
				dstring *temp = ds_create("value\t");
				ds_appendch(temp, get_html_char(i->this_chunk));
				ds_appendch(temp, '\n');
				ds_appendstr(temp, "ivalue\t");
				ds_append(temp, ds_fromint(get_html_char(i->this_chunk), 10, 0));
				return temp;
			}
		}
		else
		{
			char *val;
			if(!*name)
				return ds_create(get_tag_name(i->this_chunk));
			if(*name == '#' && !name[1])
				return ds_fromint(list_length(i->this_chunk->params), 10, 0);
			if(*name == '*' && !name[1])
			{
				html_param *parm;
				dstring *temp = ds_create(NULL);
				for(parm = i->this_chunk->params; parm; parm = list_next(parm))
				{
					ds_append(temp, parm->name);
					ds_appendch(temp, '\t');
					if(parm->val)
						ds_append(temp, parm->val);
					if(list_next(parm))
						ds_appendch(temp, '\n');
				}
				return temp;
			}
			val = get_html_param(i->this_chunk, name);
			if(!val)
				return NULL;
			return ds_create(val);
		}
	}
	else
	{
		variable *var = get_var(i, name);
		if(!var)
			return NULL;
		return var->vals->val;
	}
	return NULL;
}

void eval_funcallx(interpreter *i, chunk *curop, argument *args)
{
	execstack *es = ALLOC(execstack);
	es->curfunc = curop;
	es->out_vals = es->last_val = NULL;
	es->nextarg = args;
	if(!es->nextarg)
	{
		es->curop = NULL;
		es->buffer = NULL;
	}
	else
	{
		es->curop  = es->nextarg->arg;
		es->nextarg = list_next(es->nextarg);
		es->buffer  = ds_fix(ds_create(NULL));
	}
	pstack->state = list_add(pstack->state, es);
	return;
}

void eval_funcall(interpreter *i, chunk *curop)
{
	eval_funcallx(i, curop, curop->type == arg_statement ? curop->x.s.ctl->header : curop->x.func.args);
}

static void trace_procedure(char *procname, value *args);
void eval_proc(interpreter *i, chunk *body, char *procname, shared_pool *pool, value *args)
{
	procstack *np;

	if(pstack && pstack->where->to_trace && procname)
		trace_procedure(procname, args);
	
	np = ALLOC(procstack);
	np->where = i;
	np->args = args;
	np->nexttop = list_next(body);
	np->cur_pool = pool;
	if(pool)
		pool->refcnt++;
	pstack = list_add(pstack, np);
	eval_funcall(i, body);
}

void eval_chunk(void)
{
	chunk *curop = pstack->state->curop;
	interpreter *i = pstack->where;
	if(i->line_tracing && i->prev_line != curop->lno)
	{
		i->prev_line = curop->lno;
		fprintf(stderr, "[%d] ", curop->lno);
	}
	set_line_no(curop->lno);
	set_current_filename(curop->file);
	switch(curop->type)
	{
		case arg_literal:
			ds_append(pstack->state->buffer, curop->x.literal);
			break;
		case arg_variable:
		{
			dstring *ds = get_variable(i, curop->x.var.is_tag, curop->x.var.name);
			if(!ds)
				recover_error("undefined variable: %s", curop->x.var.name);
			ds_append(pstack->state->buffer, ds);
			break;
		}	
		case arg_parameter:
		{
			if(curop->x.paramno == switch_header)
			{
				if(pstack->ctl_stack && pstack->ctl_stack->master->x.s.statement == s_switch)
					ds_append(pstack->state->buffer, pstack->ctl_stack->x.condition);
			}
			else if(curop->x.paramno == param_count)
				ds_append(pstack->state->buffer, ds_fromint(list_length(pstack->args), 10, 0));
			else if(!curop->x.paramno)
			{
				if(list_next(pstack) && ((procstack *)list_next(pstack))->state)
					ds_append(pstack->state->buffer, ((procstack *)list_next(pstack))->state->buffer);
			}
			else
			{
				value *arg = list_nth(pstack->args, curop->x.paramno - 1);
				if(arg)
					ds_append(pstack->state->buffer, arg->val);
			}
			break;
		}
		case arg_funcall:
		case arg_builtin:
		case arg_embedded:
			pstack->state->curop = list_next(curop);
			eval_funcall(i, curop);
			return;	
		default:
			recover_error("internal error in eval_chunk");
	}
	pstack->state->curop = list_next(curop);
}

static int trace_find(tracelist *t, char *key)
{
	return !fnmatch(t->pattern, key, 0);
}

static void trace_procedure(char *procname, value *args)
{
	if(list_find(pstack->where->to_trace, procname, (list_predicate_t)trace_find))
	{
		fprintf(stderr, "called %s(", procname);
		for(; args; args = list_next(args))
			fprintf(stderr, "%s, ", DS_BODY(args->val));
		fputs(")\n", stderr);
		pstack->trace_flag = 1;
	}
}

static void error_recovery_procedure(char *desc)
{
//	procstack *np;
	interpreter *i;
	shared_int *si;
	value *nv, *nv1;
	if(!pstack)
	{
		unwind_error();
		recover_error("%s", DS_BODY(ds_create(desc)));
	}
	else
	{
		i = pstack->where;
		si = list_find(i->procs, i->error_handler, (list_predicate_t)si_find_byso);
		nv = ALLOC(value);
		nv->val = ds_fix(ds_create(set_current_filename(NULL)));
		nv1 = ALLOC(value);
		nv1->val = ds_fix(ds_fromint(set_line_no(-1), 10, 0));
		nv = list_add(nv1, nv);
		nv1 = ALLOC(value);
		nv1->val = ds_fix(ds_create(desc));
		nv = list_add(nv, nv1);
		if(!i->error_handler || !i->error_handler->body)
		{
			unwind_error();
			while(pstack && !pstack->boundary)
				pstack = list_pop(pstack, (list_destroy_t)destroy_procframe, sizeof(procstack));
			if(pstack)
				pstack->boundary = 0;
			recover_error("%s", DS_BODY(ds_create(desc)));
		}
//	list_delete(pstack->state, (list_destroy_t)destroy_execstate, sizeof(execstack));
//	pstack->state = NULL;
		eval_proc(i->error_handler->sh.owner, i->error_handler->body, si->intname, i->error_handler->pool, nv);
	}
}

static void handle_trap(char *trapname)
{
	trap_handler *handler = list_find(all_traps, trapname, (list_predicate_t)trap_find);
	value *val;
	if(!handler)
		return;
	val = ALLOC(value);
	val->val = ds_fix(ds_create(trapname));
	eval_proc(handler->proc->sh.owner, handler->proc->body, NULL, handler->proc->pool, val);
}

int squee_interval = 3000;

#define SQUEE_INTERVAL (squee_interval)

static void stmt_if(chunk *current, interpreter *i, value *args);
static void stmt_foreach(chunk *current, interpreter *i, value *args);
static void stmt_while(chunk *current, interpreter *i, value *args);
static void stmt_switch(chunk *current, interpreter *i, value *args);

static dstring *ret_val;

/// \section[EXECUTION MODEL] 
// The NSL interpreter is a tree of \em2[executon environment][.] Each
// execution environment is comparably independent from others and in
// principle communicate with them via \em2[sharable objects][.] The root
// of this tree is the environment holding the script loaded at startup.
// Strictly speaking, the structure may be not a tree, since enviroments are
// sharable objects, so the parent may export itself and its child will
// import it as a subordinated environment, thus yielding a loop. An
// environment has no implicit reference to its parent, so it may freely
// consider itself the topmost.
//
// Sharable objects include streams, storages, procedures, automata
// sequences and environments (also called \em2[subinterpreters][).]
// Variables are with the help of \em[pools] -- sets of variables available
// to all the environments.
//
// At a given moment each environment executes a certain procedure.
// Procedures are sequences of calls to other procedures, builtins and
// control statements. Builtin calls are atomic in the sense that trap and
// error handling do not occur when a builtin is executed. A call have a
// certain number of arguments which consist of \em2[chunks][.] (see SYNTAX OVERVIEW section 
// below). Arguments are always evaluated prior to the call itself. Chunks are evaluated 
// left to right, the result of their evaluation being added to the
// \em[current execution buffer]. New buffers are allocated when a
// subordinate call or a block is evaluated. When the call is done, its
// result is added to the upper-level buffer. However, unlike in other
// languages, NSL builtins and functions may affect the parent buffer in
// other way than just adding to it. It may e. g. completely replace the
// contents of the buffer. Functions with such behaviour are supposed to
// have their names started with '.'

static int process_short_circuit(void);
dstring *eval_script(interpreter *i, chunk *body, char *name, shared_pool *pool, value *args)
{
//	procstack *np;
	int squee_cnt = SQUEE_INTERVAL;
	char *trap_name;
	static instruction_t st_table[] = {stmt_if, stmt_foreach, stmt_while, stmt_switch};

	ret_val = NULL;
	if(body)
	{
		if(pstack)
			pstack->boundary = 1;
		eval_proc(i, body, name, pool, args);
	}
	register_error_handler(NULL);
	if(CATCH())
	{
		squee_cnt = SQUEE_INTERVAL;
		error_recovery_procedure(get_error_desc());
	}
	for(;;)
	{
		if((trap_name = check_for_trap()))
			handle_trap(trap_name);
		if(!--squee_cnt)
		{
			ds_squeeze(-1);
			squee_cnt = SQUEE_INTERVAL;	
		}
		if(!pstack || pstack->boundary)
			break;
		if(!pstack->state)
		{
			if(pstack->nexttop)
			{
				eval_funcall(pstack->where, pstack->nexttop);
				pstack->nexttop = list_next(pstack->nexttop);
			}
			else
			{
				if(pstack->ctl_stack)
					recover_error("unfinished control statement (internal error?)");
				if(pstack->trace_flag)
					fprintf(stderr, "return by end\n");
				pstack = list_pop(pstack, (list_destroy_t)destroy_procframe, sizeof(procstack));
			}
		}
		else if(pstack->state->curop)
			eval_chunk();
		else
		{
			if(pstack->state->buffer)
			{
				value *newval = ALLOC(value);
				newval->val = pstack->state->buffer;
				if(pstack->state->last_val)
				{
					list_add(newval, pstack->state->last_val);
					pstack->state->last_val = newval; 
				}
				else pstack->state->last_val = pstack->state->out_vals = newval;
// 				pstack->state->out_vals = list_append(pstack->state->out_vals, newval);
			}
			process_short_circuit(); 
			if(pstack->state->nextarg)
			{
				pstack->state->curop = pstack->state->nextarg->arg;
				pstack->state->nextarg = list_next(pstack->state->nextarg);
				pstack->state->buffer = ds_fix(ds_create(NULL));
			}
			else
			{
				value *args = pstack->state->out_vals;
				chunk *curf = pstack->state->curfunc;
				pstack->state = list_pop(pstack->state, NULL, sizeof(execstack));
				if(curf->type == arg_builtin || curf->type == arg_statement)
				{
					(curf->type == arg_builtin ? curf->x.func.f.bi : st_table[curf->x.s.statement])(curf, pstack->where, args);
					list_delete(args, (list_destroy_t)destroy_value, sizeof(value));
				}
				else if(curf->type == arg_embedded)
				{
					eval_proc(pstack->where, curf->x.func.f.emb, NULL, pstack->cur_pool, args);
					pstack->vars = pstack->start_vars = ((procstack *)list_next(pstack))->vars;
				}
				else
				{
					shared_int *si_proc = list_find(pstack->where->procs, curf->x.func.f.name, (list_predicate_t)si_find_exact);
					procedure *proc;
					if(!si_proc)
						recover_error("undefined procedure name: '%s'", curf->x.func.f.name);
					check_si_perm(i, si_proc);
					proc = si_proc->obj;
					eval_proc(proc->callback ? pstack->where : proc->sh.owner, proc->body, si_proc->intname, proc->pool, args);
				}
			}
		}	
	}
	if(pstack)
		pstack->boundary = 0;
	unwind_error();
	return ret_val;
}

static void bi_cond(chunk *current, interpreter *i, value *args);
static void bi_and(chunk *current, interpreter *i, value *args);
static void bi_or(chunk *current, interpreter *i, value *args);

static int process_short_circuit(void)
{
	chunk *curf = pstack->state->curfunc;
	if(curf->type != arg_builtin)
		return 0;
	if(curf->x.func.f.bi == bi_cond)
	{
		if(list_next(pstack->state->out_vals))
			pstack->state->nextarg = NULL;
		else if(ds_isempty(pstack->state->buffer))
			pstack->state->nextarg = list_next(pstack->state->nextarg);
	}
	else if(curf->x.func.f.bi == bi_and)
	{
		if(ds_isempty(pstack->state->buffer))
		{
			pstack->state->nextarg = NULL;
			list_delete(pstack->state->out_vals, (list_destroy_t)destroy_value, sizeof(value));
			pstack->state->out_vals = NULL;
		}
	}
	else if(curf->x.func.f.bi == bi_or)
	{
		if(!ds_isempty(pstack->state->buffer))
		{
			pstack->state->nextarg = NULL;
			list_delete(pstack->state->out_vals, (list_destroy_t)destroy_value, sizeof(value));
			pstack->state->out_vals = NULL;
		}
	}
	return 0;
}

static void bi_dispatch(chunk *current, interpreter *i, value *args);

dstring *eval_procedure(interpreter *i, char *pname, value *args)
{
	shared_int *si_p = list_find(i->procs, pname, (list_predicate_t)si_find_exact);
	if(si_p)
	{
		procedure *p = si_p->obj;
		if(p->body)
			return eval_script(p->sh.owner, p->body, si_p->intname, p->pool, args);
	}
	return NULL;
}

dstring *eval_procedurev(interpreter *i, char *pname, ...)
{
	value *procargs = NULL, *narg;
	va_list args;
	dstring *nextarg;
	va_start(args, pname);
	for(;;)
	{
		nextarg = va_arg(args, dstring *);
		if(!nextarg)
			break;
		narg = ALLOC(value);
		narg->val = ds_fix(nextarg);
		procargs = list_append(procargs, narg);
	}
	va_end(args);
	return eval_procedure(i, pname, procargs);
}

void discard_value(dstring *ds)
{
	if(ds)
		ds_unfix(ds);
}

void dispatch_input(interpreter *i)
{
	if(pstack)
		pstack->boundary = 1;
	bi_dispatch(NULL, i, NULL);
	i->already_read = 0;
	discard_value(eval_script(i, NULL, NULL, NULL, NULL));
}

int squee_long_string = 10000;
static volatile int exit_code;
static jmp_buf exit_jmp;

int script_main_loop(interpreter *i)
{
	if(setjmp(exit_jmp))
		return exit_code;
	discard_value(eval_procedure(i, "main", NULL));
	if(i->tags || i->filters)
	{
		for(;;)
		{
			if(squee_long_string > 0)
				ds_squeeze(squee_long_string); 
			if(!i->already_read)
			{
				if(!i->instream)
				{
					exit_code = EXIT_SUCCESS;
					break;
				}
				release_html_chunk(i->this_chunk);
				i->this_chunk = copy_html_chunk(i->instream->class->read_chunk(i, i->instream));
			}
			if(!i->this_chunk)
			{
				i->instream = NULL;
				i->already_read = 0;
				discard_value(eval_procedure(i, "eof", NULL));
				continue;
			}
			dispatch_input(i);
		}
	}
	discard_value(eval_procedure(i, "atexit", NULL));
	return exit_code;
}

	
void cleanup_interpreter(interpreter *i)
{
	list_delete(i->globvars, (list_destroy_t)destroy_variable, sizeof(variable));
	i->globvars = NULL;
	i->already_read = 0;
	list_delete(i->stos, (list_destroy_t)destroy_shared_int, sizeof(shared_int));
	i->stos = NULL;
	list_delete(i->streams, (list_destroy_t)destroy_shared_int, sizeof(shared_int));
	i->streams = NULL;
	list_delete(i->aseqs, (list_destroy_t)destroy_shared_int, sizeof(shared_int));
	i->aseqs = NULL;
}

/// \section[COMMAND SUMMARY]

////
// -(arg)
// \returns arg
static void bi_ident(chunk *current, interpreter *i, value *args) 
{
	ds_append(pstack->state->buffer, args->val);
}

////
// tag(tag-name)
// \returns a value of a tag parameter with the name \em2[tag-name][;]
// if \em[tag-name] is empty, the name of the tag;
// \em[tag-name] = '#' -- number of tag params;
// \em[tag-name] = '*' -- a string of pairs 'name\ <TAB>\ value\ <NL>' for every tag 
// parameter, where <TAB> means 
// a tab character, <NL> -- a newline character, spaces are for clarity only
// if the current tag is a character, two pseudo names are recognized:
// \item 'value': the character itself
// \item 'ivalue': the numeric code of the character.
// \endlist
// If there is no current tag, empty string
static void bi_tag(chunk *current, interpreter *i, value *args) 
{
	dstring *var = get_variable(i, 1, DS_BODY(args->val));
	if(var)
		ds_append(pstack->state->buffer, var);
}

///
// settag(tag-name value)
// sets the value of a tag parameter \em[tag-name] to \em2[value][.]
// If there is no such parameter, it is added
// \use[ncc]
static void bi_settag(chunk *current, interpreter *i, value *args) 
{
	if(!i->this_chunk)
		recover_error("no current chunk");
	if(ds_isempty(args->val))
		set_tag_name(i->this_chunk, next_value(args)->val);
	else
	{
		if(HTML_CHUNK_TYPE(i->this_chunk) == hct_tag)
			set_html_param(i->this_chunk, args->val, next_value(args)->val);
		else
		{
			if(!ds_comparestr(args->val, "value", ds_p_casefold, NULL))
				set_html_char(i->this_chunk, *DS_BODY(next_value(args)->val));
			else if(!ds_comparestr(args->val, "ivalue", ds_p_casefold, NULL))
				set_html_char(i->this_chunk, convert_to_int(DS_BODY(next_value(args)->val)));
		}
	}
}

///
// settagchar(tag-char)
// sets the type of the current chunk to \em[char] with the value
// \em[tag-char].
// \use[ncc]
static void bi_settagchar(chunk *current, interpreter *i, value *args) 
{
	if(!i->this_chunk)
		recover_error("no current chunk");
	set_html_char(i->this_chunk, convert_to_int(DS_BODY(args->val)));
}

///
// cleartag()
/// clears the current chunk. If there was no current chunk, it is created.
static void bi_cleartag(chunk *current, interpreter *i, value *args) 
{
	if(!i->this_chunk)
		i->this_chunk = alloc_html_chunk();
	reset_html_chunk(i->this_chunk);
}

/// rmtag(tag-name)
// removes a tag parameter \em[tag-name] from the current chunk
// \use[ncc]
static void bi_rmtag(chunk *current, interpreter *i, value *args)
{
	if(!i->this_chunk)
		recover_error("no current chunk");
	remove_html_param(i->this_chunk, args->val);
}

/// passtag(subint)
// replaces the current chunk of a \em[subint] with a copy of the current
// chunk of the master interpreter.
// If there is no current chunk, the current chunk of the \em[subint] is
// deleted altogether.
static void bi_passtag(chunk *current, interpreter *i, value *args)
{
	interpreter *in = find_subint(i, DS_BODY(args->val));
	if(in->this_chunk)
		release_html_chunk(in->this_chunk);
	if(i->this_chunk)
		in->this_chunk = copy_html_chunk(i->this_chunk);
	else in->this_chunk = NULL;
}

///
// var(varname)
// \returns the value of a variable \em2[varname][;] 
// if no such variable exists, a null string is returned
static void bi_var(chunk *current, interpreter *i, value *args) 
{
	dstring *var = get_variable(i, 0, DS_BODY(args->val));
	if(var)
		ds_append(pstack->state->buffer, var);
}

///
// arg(arg-no)
// \returns the value of \em2[arg-no]th argument of the current procedure.
// \error[invalid parameter index: arg-no] if arg-no <= 0
static void bi_arg(chunk *current, interpreter *i, value *args) 
{
	int argno = convert_to_int(DS_BODY(args->val));
	value *arg;
	if(argno <= 0)
		recover_error("invalid parameter index: %d", argno);
	arg = list_nth(pstack->args, argno - 1);
	if(arg)
		ds_append(pstack->state->buffer, arg->val);
}

static void bi_passargs(chunk *current, interpreter *i, value *args) 
{
	int argno = convert_to_int(DS_BODY(args->val));
	value *arg, *nv;
	if(argno <= 0)
		recover_error("invalid parameter index: %d", argno);
	for(arg = list_nth(pstack->args, argno - 1); arg; arg = list_next(arg))
	{
		nv = ALLOC(value);
		nv->val = ds_fix(arg->val);
		pstack->state->out_vals = list_append(pstack->state->out_vals, nv);
	}
	ds_unfix(pstack->state->buffer);
	pstack->state->buffer = NULL;
}


#ifdef ENABLE_REF
///
// ref(varname)
// \returns a reference to a \em2[varname][.] This value may be used in any
// place where a variable name is acceptable, e. g. as an argument to \em2[var][,]
// a first argument to \em2[set][,] etc. This allows to implement a
// pass-by-refererence technique, since the properties of a reference are
// the same as of the original variable, incl. the state of its stack
// \note \em[ref] should be used with care when applied to local variables,
// since if the reference is accessed after the end of the variable
// lifetime, the result is unpredictable.
// \note applications should treat references as opaque data and not try to
// analyse or construct them besides calling \em2[ref][.]
// \note since references are implemented as pointers, their presence is a
// considerable security hole even if there are no 'ref' calls in the
// script. Therefore, they may be disabled at all at compile-time by
// passing --disable-ref to ./configure
static void bi_ref(chunk *current, interpreter *i, value *args) 
{
	variable *var = safe_get_var(i, DS_BODY(args->val), 0);
	ds_appendstr(pstack->state->buffer, "##");
	ds_append(pstack->state->buffer, ds_fromint((int)var, 16, 0));
}
#endif

///
// pool(name)
// creates a new shared pool named \em2[name][.] 
// If such pool already exists, nothing happens.
static void bi_pool(chunk *current, interpreter *i, value *args) 
{
	shared_pool *np = list_find(all_pools, DS_BODY(args->val), (list_predicate_t)pool_find);
	if(!np)
	{
		np = ALLOC(shared_pool);
		np->name = xstrdup(DS_BODY(args->val));
		all_pools = list_add(all_pools, np);
	}
}

///
// rmpool(name)
// deletes a pool \em2[name][.] If the pool does not exist, does nothing
static void bi_rmpool(chunk *current, interpreter *i, value *args) 
{
	shared_pool *np = list_find(all_pools, DS_BODY(args->val), (list_predicate_t)pool_find);
	if(np && !np->refcnt)
	{	
		all_pools = list_remove(all_pools, DS_BODY(args->val), (list_predicate_t)pool_find,
								(list_destroy_t)destroy_pool, sizeof(shared_pool));
	}
}
	
///
// using([pool-name])
// selects the shared pool \em[pool-name] for further use.
// If \em[pool-name] is empty, no pool is selected.
// If \em[pool-name] is omitted, the current pool is not changed.
// \returns the name of a pool previously in use
// \error[pool 'pool-name' undefined]
static void bi_using(chunk *current, interpreter *i, value *args) 
{
	shared_pool *p = NULL;
	shared_pool *prev = pstack->cur_pool;
	if(args)
	{
		if(!ds_isempty(args->val))
		{
			p = list_find(all_pools, DS_BODY(args->val), (list_predicate_t)pool_find);
			if(!p)
				recover_error("pool '%s' undefined", DS_BODY(args->val));
		}
		if(pstack->cur_pool)
			pstack->cur_pool->refcnt--;
		pstack->cur_pool = p;
		p->refcnt++;
	}
	if(prev)
		ds_appendstr(pstack->state->buffer, prev->name);
}


///
// eval(arg)
// Parses and evaluates arg as a valid NSL argument expression
// \returns the value of evaluated arg
static void bi_return(chunk *current, interpreter *i, value *args);
static void bi_eval(chunk *current, interpreter *i, value *args)
{
	chunk *ch = parse_argument(i, zero_fc, DS_BODY(args->val), NULL);
	chunk *func = ALLOC(chunk);
	dstring *res;
	
	func->lno = current->lno;
	func->file = current->file;
	func->type = arg_builtin;
	func->x.func.f.bi = bi_return;
	func->x.func.args = ALLOC(argument);
	func->x.func.args->arg = ch;
	pstack->boundary = 1;
	eval_proc(i, func, NULL, pstack->cur_pool, NULL);
	pstack->vars = pstack->start_vars = ((procstack *)list_next(pstack))->vars;
	res = eval_script(i, NULL, NULL, pstack->cur_pool, NULL);
	if(res)
	{
		ds_unfix(res);
		ds_append(pstack->state->buffer, res);
	}
	list_delete(func, (list_destroy_t)destroy_chunk, sizeof(chunk));
}

///
// void(arg)
// Only discard the value of arg. It is hardly to be used explicitly,
// though the interpreter inserts it implicitly every time a built-in
// function which returns a value is used at the top level
// \use[tlo]
static void bi_dummy(chunk *current, interpreter *i, value *args)
{
	// dummy
}

///
// if cond1 ops1 [elsif cond2 ops2]... [else opsn] endif
// If \em[condi] is non-empty, \em[opsi] are evaluated, otherwise \em[opsn] if present
// \use[tlo]
// \error['if'/'else'/'elsif' without 'endif']
static void bi_if(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error at post-parsing");
}

static void stmt_if(chunk *current, interpreter *i, value *args)
{
	if(!pstack->ctl_stack || pstack->ctl_stack->master != current)
	{
		ctlstack *ctl = ALLOC(ctlstack);
		ctl->master = current;
		ctl->cur_branch = current->x.s.ctl->branches;
		pstack->ctl_stack = list_add(pstack->ctl_stack, ctl);
		eval_funcallx(i, current, pstack->ctl_stack->cur_branch->condition);
		return;
	}
	if(!ds_isempty(args->val))
		pstack->nexttop = pstack->ctl_stack->cur_branch->loc;
	else
	{
		pstack->ctl_stack->cur_branch = list_next(pstack->ctl_stack->cur_branch);
		if(!pstack->ctl_stack->cur_branch)
		{
			pstack->nexttop = current->x.s.ctl->ctl_end;
			pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
		}
		else if(!pstack->ctl_stack->cur_branch->condition)
			pstack->nexttop = pstack->ctl_stack->cur_branch->loc;
		else eval_funcallx(i, current, pstack->ctl_stack->cur_branch->condition);
	}
}


static void bi_else(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error");
}

static void bi_elsif(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error");
}

static void bi_endif(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error");
}

static void bi_jump_to_endif(chunk *current, interpreter *i, value *args)
{
	pstack->nexttop = pstack->ctl_stack->master->x.s.ctl->ctl_end;
	pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
}

static void bi_popdown_ctl(chunk *current, interpreter *i, value *args)
{
	pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
}

/// 
// switch arg ... case pattern ops ... caserange arg1 arg2 ops ... breaksw ... endsw
// \em[ops] are evaluated following \em3['][case]['] where \em[arg] matches \em2[pattern][,] or
// \em3['][caserange]['] where \em[arg] is between \em[arg1] and \em[arg2]
// Evaluation stops after encountering \em3['][breaksw]['] or \em3['][endsw][']
// \use[tlo]
// \error[switch without endsw]
static void bi_switch(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error");
}

static void stmt_switch(chunk *current, interpreter *i, value *args)
{
	if(!pstack->ctl_stack || pstack->ctl_stack->master != current)
	{
		ctlstack *ctl = ALLOC(ctlstack);
		ctl->master = current;
		ctl->x.condition = ds_fix(args->val);
		pstack->ctl_stack = list_add(pstack->ctl_stack, ctl);
		ctl->cur_branch = current->x.s.ctl->branches;
		eval_funcallx(i, current, ctl->cur_branch->condition);
	}
	else
	{
		if(!ds_isempty(args->val))
		{
			pstack->nexttop = pstack->ctl_stack->cur_branch->loc;
		}
		else
		{
			pstack->ctl_stack->cur_branch = list_next(pstack->ctl_stack->cur_branch);
			if(!pstack->ctl_stack->cur_branch)
			{
				pstack->nexttop = current->x.s.ctl->ctl_end;
				pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
			}
			else eval_funcallx(i, current, pstack->ctl_stack->cur_branch->condition);
		}
	}
}


static void bi_case(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error");
}

static void bi_caserange(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error");
}

static void bi_when(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error");
}


static void bi_breaksw(chunk *current, interpreter *i, value *args)
{
	while(pstack->ctl_stack && pstack->ctl_stack->master->x.s.statement != s_switch)
		pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
	if(!pstack->ctl_stack)
		recover_error("breaksw outside switch...endsw");
	pstack->nexttop = pstack->ctl_stack->master->x.s.ctl->ctl_end;
	pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
}

static void bi_endsw(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error");
}

///
// while [arg] ... continue ... break ... loop[(arg2)]
// executes the loop while \em[arg] and \em[arg2] are both not null
// If \em[arg] or \em[arg2] missing, there are considered always true
// \em3['][break]['] exits the loop immediately
// \em3['][loop]['] skips until the next pass
// \use[tlo]
static void bi_while(chunk *current, interpreter *i, value *args)
{
	recover_error("internal error");
}

static void stmt_while(chunk *current, interpreter *i, value *args)
{
	if(!pstack->ctl_stack || pstack->ctl_stack->master != current)
	{
		ctlstack *ctl = ALLOC(ctlstack);
		ctl->master = current;
		pstack->ctl_stack = list_add(pstack->ctl_stack, ctl);
	}
	if(args && ds_isempty(args->val))
	{
		pstack->nexttop = list_next(current->x.s.ctl->ctl_end);
		pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
	}
}

static void bi_loop(chunk *current, interpreter *i, value *args)
{
	if(!pstack->ctl_stack || 
		(pstack->ctl_stack->master->x.s.statement != s_foreach &&
		 pstack->ctl_stack->master->x.s.statement != s_while))
		recover_error("'loop' without 'while/foreach'");
	if(!args || !ds_isempty(args->val))
		pstack->nexttop = pstack->ctl_stack->master;
	else
		pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
}

static void find_loop_ctlstack(char *func)
{
	while(pstack->ctl_stack &&
		(pstack->ctl_stack->master->x.s.statement != s_foreach &&
		 pstack->ctl_stack->master->x.s.statement != s_while))
		pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
	if(!pstack->ctl_stack)
		recover_error("'%s' without 'while/foreach'", func);
}

static void bi_break(chunk *current, interpreter *i, value *args)
{
	find_loop_ctlstack("break");
	if(pstack->ctl_stack->master->x.s.ctl->on_break)
		pstack->nexttop = pstack->ctl_stack->master->x.s.ctl->on_break;
	else
	{
		pstack->nexttop = list_next(pstack->ctl_stack->master->x.s.ctl->ctl_end);
		pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
	}
}

static void bi_after_break(chunk *current, interpreter *i, value *args)
{
	pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
	pstack->nexttop = list_next(pstack->nexttop);
}

static void bi_continue(chunk *current, interpreter *i, value *args)
{
	find_loop_ctlstack("continue");
	pstack->nexttop = pstack->ctl_stack->master->x.s.ctl->ctl_end;
}

///
// foreach var storage ... loop
// Repeats the loop for every record in \em[storage]
// Every time \em[var] is assigned the record key
// \em3['][loop]['], \em3['][continue]['], \em3['][break]['] are used 
// just like with \em3['][while][']
// \use[tlo]
static void bi_foreach(chunk *current, interpreter *i, value *args)
{
	char *sname = DS_BODY(next_value(args)->val);
	ctlstack *nls = ALLOC(ctlstack);
	nls->x.loop.hangon = find_storage(i, sname);
	nls->x.loop.var = safe_get_var(i, DS_BODY(args->val), 1);
	nls->x.loop.first_time = 1;
	nls->master = list_next(current);
	pstack->ctl_stack = list_add(pstack->ctl_stack, nls);
}

static void stmt_foreach(chunk *current, interpreter *i, value *args)
{
	dstring *key;
	if(!pstack->ctl_stack)
		recover_error("internal error");
	key = (pstack->ctl_stack->x.loop.first_time ? pstack->ctl_stack->x.loop.hangon->class->first :
		   pstack->ctl_stack->x.loop.hangon->class->next)
		(i, pstack->ctl_stack->x.loop.hangon, &pstack->ctl_stack->x.loop.param);
	pstack->ctl_stack->x.loop.first_time = 0;
	if(!key)
	{
		pstack->nexttop = list_next(current->x.s.ctl->ctl_end);
		pstack->ctl_stack = list_pop(pstack->ctl_stack, (list_destroy_t)destroy_ctl, sizeof(ctlstack));
	}
	else
		ds_assign(pstack->ctl_stack->x.loop.var->vals->val, key);
}

///
// return[(arg)]
// Finished the current procedure returning \em2[arg][t]o a parent.
// If \em[arg] is not present, a null string is returned
static void bi_return(chunk *current, interpreter *i, value *args)
{
	if(pstack->trace_flag)
	{
		fprintf(stderr, "return ");
		if(args)
			fputs(DS_BODY(args->val), stderr);
		fputc('\n', stderr);
	}
	pstack = list_pop(pstack, (list_destroy_t)destroy_procframe, sizeof(procstack));
	if(!pstack || pstack->boundary)
	{
		if(args)
			ret_val = ds_fix(args->val);
	}
	else if(pstack->state && args)
		ds_append(pstack->state->buffer, args->val);
}

///
// .return(arg)
// Finishes the current procedure joining \em[arg] to the parent buffer
static void bi_preturn(chunk *current, interpreter *i, value *args)
{
	if(pstack->trace_flag)
	{
		fprintf(stderr, ".return ");
		if(args)
			fputs(DS_BODY(args->val), stderr);
		fputc('\n', stderr);
	}
	pstack = list_pop(pstack, (list_destroy_t)destroy_procframe, sizeof(procstack));
	if(!pstack || pstack->boundary)
		ret_val = ds_fix(args->val);
	else if(pstack->state && args)
		ds_assign(pstack->state->buffer, args->val);
}

///
// once(arg)
// \returns \em[arg] and replaces itself with it.
// This allows to create self-modifying programs
// \error['once' called in an inappropriate context]
void bi_once(chunk *current, interpreter *i, value *args)
{
	if(!current)
		recover_error("'once' called in an inappropriate context");
	list_delete(current->x.func.args, (list_destroy_t)destroy_argument, sizeof(argument));
	current->type = arg_literal;
	current->x.literal = ds_fix(args->val);
	ds_append(pstack->state->buffer, args->val);
}

/// 
// Exit
// ends the main execution loop, forcing next input chunk to be read
// and processed
// \use[tlo]
static void bi_exit(chunk *current, interpreter *i, value *args)
{
	while(pstack && !pstack->boundary)
		pstack = list_pop(pstack, (list_destroy_t)destroy_procframe, sizeof(procstack));
}

/// exit[(code)]
// unconditionally ends the execution of the program
// \em[code] may be either \em[success] or \em[failure]
// or an integer. If it is missing, \em[success] is assumed.
// \use[tlo]
static void bi_realexit(chunk *current, interpreter *i, value *args)
{
	exit_code = EXIT_SUCCESS;
	if(args)
	{
		if(!ds_comparestr(args->val, "success", NULL, NULL))
			exit_code = EXIT_SUCCESS;
		else if(!ds_comparestr(args->val, "failure", NULL, NULL))
			exit_code = EXIT_FAILURE;
		else exit_code = convert_to_int(DS_BODY(args->val));
	}
	longjmp(exit_jmp, 1);
}

/// export kind int-name ext-name
// exports an object of a \em[kind] with a name \em[int-name] as
// \em[ext-name].
// Kind may be one of \em2[storage][,] \em2[stream][,] \em2[autoseq][,]
// \em2[proc][,] \em[subint][.]
// As a special case, \em[export subint self ...] exports the current
// interpreter
// \error[kind 'int-name' undefined]
static void bi_export(chunk *current, interpreter *i, value *args)
{
	shared_int *si;
	if(!strcmp(DS_BODY(args->val), "subint") && !strcmp(DS_BODY(next_value(args)->val), "self"))
		export_shared(&i->sh, DS_BODY(third_value(args)->val));
	else
	{
		si = find_si_itself(i, DS_BODY(args->val), DS_BODY(next_value(args)->val));
		if(!si)
			recover_error("%s '%s' undefined", DS_BODY(args->val), DS_BODY(next_value(args)->val));
		export_shared(si->so, DS_BODY(third_value(args)->val));
	}
}

///
// embed subint kind int-name remote-name
// like \em[export], but instead of global exporting directly embeds the
// object \em[int-name] into the interpreter \em[subint] as \em[remote-name]
// \error[kind 'int-name' undefined]
// \error[kind 'int-name' already defined]
static void bi_embed(chunk *current, interpreter *i, value *args)
{
	shared_int *si;
	interpreter *target = find_subint(i, DS_BODY(args->val));
	
	if(!strcmp(DS_BODY(next_value(args)->val), "subint") && !strcmp(DS_BODY(third_value(args)->val), "self"))
		si = embed_shared(i, "subint", DS_BODY(next_value(third_value(args))->val));
	else
	{
		si = find_si_itself(i, DS_BODY(next_value(args)->val), DS_BODY(third_value(args)->val));
		if(!si)
			recover_error("%s '%s' undefined", DS_BODY(next_value(args)->val), DS_BODY(third_value(args)->val));
		if(find_si_itself(target, DS_BODY(next_value(args)->val), DS_BODY(next_value(third_value(args))->val)))
			recover_error("%s '%s' already defined", DS_BODY(next_value(args)->val), DS_BODY(next_value(third_value(args))->val));
		si = embed_shared(si->obj, DS_BODY(next_value(args)->val), DS_BODY(next_value(third_value(args))->val));
	}
	add_shared_int(target, si);
}

/// isexport(kind ext-name)
// \returns true if an object of the \em[kind] is exported as
// \em2[ext-name][.]
static void bi_isexport(chunk *current, interpreter *i, value *args)
{
	if(find_exported(DS_BODY(args->val), DS_BODY(next_value(args)->val)))
		ds_appendstr(pstack->state->buffer, "true");
}

/// import(kind int-name ext-name)
// import an object of the \em[kind] with external name \em[ext-name] as
// \em2[int-name][.]
static void bi_import(chunk *current, interpreter *i, value *args)
{
	shared_int *si;
	si = find_si_itself(i, DS_BODY(args->val), DS_BODY(next_value(args)->val));
	if(si)
		recover_error("%s '%s' already defined", DS_BODY(args->val), DS_BODY(next_value(args)->val));
	si = import_shared(DS_BODY(next_value(args)->val), DS_BODY(third_value(args)->val), DS_BODY(args->val));
	add_shared_int(i, si);
}

/// allow(kind int-name subint-names ...) and deny(kind int-name subint-names ...)
// allows or denies access to an object of the \em[kind] named \em[int-name] for environments
// listed under \em2[subint-names][.] If a \em[subint-name] is \em2[*][,] access is granted to or
// revoked from any environment. If a \em[subint-name] starts with \em2[::][,] the rest of it is a
// local name of a environment, otherwise -- an exported one.
// \note Only the object owner may change its access lists.
// \note Access list takes precedence over deny list; if both are empty, access is granted to
// everybody
// \note If you allow or deny access to an object to an environment and the environment is already
// denied or allowed access, the previous access rule is discarded. Consequently, when you
// deny/allow access to everybody, the opposite access list is cleared. 
static void bi_allow(chunk *current, interpreter *i, value *args)
{
	shared_int *si;
	interpreter *subi;
	char *name;
	si = find_si_itself(i, DS_BODY(args->val), DS_BODY(next_value(args)->val));
	if(!si)
		recover_error("%s '%s' undefined", DS_BODY(args->val), DS_BODY(next_value(args)->val));
	for(args = third_value(args); args; args = next_value(args))
	{
		name = DS_BODY(args->val);
		if(*name == '*' && !name[1])
			allow_shared_int(i, si, NULL);
		else
		{
			if(*name == ':' && name[1] == ':')
				 subi = find_subint(i, name + 2);
			else subi = find_exported("subint", name);
			allow_shared_int(i, si, subi);
		}
	}
}

static void bi_deny(chunk *current, interpreter *i, value *args)
{
	shared_int *si;
	interpreter *subi;
	si = find_si_itself(i, DS_BODY(args->val), DS_BODY(next_value(args)->val));
	if(!si)
		recover_error("%s '%s' undefined", DS_BODY(args->val), DS_BODY(next_value(args)->val));
	for(args = third_value(args); args; args = next_value(args))
	{
		if(*DS_BODY(args->val) == '*')
			deny_shared_int(i, si, NULL);
		else
		{
			subi = find_subint(i, DS_BODY(args->val));
			deny_shared_int(i, si, subi);
		}
	}
}

/// rename(kind old-name new-name [force])
// renames an object \em[old-name] of the \em[kind] to \em2[new-name][.]
// If the object doesn't exist an error occurs unless \em[force] is present
// and not empty.

static void bi_rename(chunk *current, interpreter *i, value *args)
{
	rename_shared_int(i, DS_BODY(args->val), DS_BODY(next_value(args)->val), 
					  DS_BODY(third_value(args)->val), 
					  next_value(third_value(args)) ? !!*DS_BODY(next_value(third_value(args))->val) : 0);
}

typedef void (*enum_t)(interpreter *i, list_t *iter, dstring *buffer);

static void enum_globals(interpreter *i, list_t *iter, dstring *buffer)
{
	if(!*iter)
	{
		*iter = i->globvars;
		if(!*iter)
			return;
	}
	ds_appendstr(buffer, ((variable *)*iter)->name);
}

static void enum_locals(interpreter *i, list_t *iter, dstring *buffer)
{
	if(!*iter)
	{
		*iter = pstack->vars;
		if(!*iter)
			return;
	}
	ds_appendstr(buffer, ((variable *)*iter)->name);
}

static void enum_shareds(interpreter *i, list_t *iter, dstring *buffer)
{
	if(!*iter)
	{
		if(pstack->cur_pool)
			*iter = pstack->cur_pool->vars;
		if(!*iter)
			return;
	}
	ds_appendstr(buffer, ((variable *)*iter)->name);
}

static void enum_pools(interpreter *i, list_t *iter, dstring *buffer)
{
	if(!*iter)
	{
		*iter = all_pools;
		if(!*iter)
			return;
	}
	ds_appendstr(buffer, ((shared_pool *)*iter)->name);
}

static void enum_streams(interpreter *i, list_t *iter, dstring *buffer)
{
	if(!*iter)
	{
		*iter = i->streams;
		if(!*iter)
			return;
	}
	ds_appendstr(buffer, ((shared_int *)*iter)->intname);
}

static void enum_storages(interpreter *i, list_t *iter, dstring *buffer)
{
	if(!*iter)
	{
		*iter = i->stos;
		if(!*iter)
			return;
	}
	ds_appendstr(buffer, ((shared_int *)*iter)->intname);
}

static void enum_procs(interpreter *i, list_t *iter, dstring *buffer)
{
	if(!*iter)
	{
		*iter = i->procs;
		if(!*iter)
			return;
	}
	ds_appendstr(buffer, ((shared_int *)*iter)->intname);
}

static void enum_autoseqs(interpreter *i, list_t *iter, dstring *buffer)
{
	if(!*iter)
	{
		*iter = i->aseqs;
		if(!*iter)
			return;
	}
	ds_appendstr(buffer, ((shared_int *)*iter)->intname);
}

/// listof(kind [arg])
// \returns the list of entities of the \em2[kind][.] It is a plain list like
// those created by \em2[list][.] It contains names only no values whatsoever.
// The following \em[kinds] are supported by the core:
// \item globals
// \item locals
// \item shareds
// \item pools
// \item streams
// \item storages
// \item procs
// \item autoseqs
// \endlist
//
// \note \em[arg] is not used by the core, but the extensions provided by modules
// might use it.
static void bi_listof(chunk *current, interpreter *i, value *args)
{
	static table_t types[] = {
								{"globals", enum_globals},
								{"locals",  enum_locals},
								{"shareds",	enum_shareds},
								{"pools",	enum_pools},
								{"streams", enum_streams},
								{"storages", enum_storages},
								{"procs", enum_procs}, 
								{"autoseqs", enum_autoseqs}, 
								{NULL, NULL}
							};
	enum_t enumerator = lookup_table(types, DS_BODY(args->val));
	list_t iter = NULL;
	dstring *buffer = ds_create(NULL);
	dstring *null = ds_create(NULL);
	storage *s = create_storage(i, &plist_storage_class, NULL, NULL);
	
	if(!enumerator)
		recover_error("listof: unknown enumeration '%s'", DS_BODY(args->val));
	enumerator(i, &iter, buffer);
	if(iter)
	{
		for(;;)
		{
			s->class->add(i, s, buffer, null);
			iter = list_next(iter);
			if(!iter)
				break;
			ds_assign(buffer, NULL);
			enumerator(i, &iter, buffer);
		}
	}
}

/// declare(name declspec)
// declares a procedure \em[name] according to \em[declspec]
//
// if \em[name] is already declared, nothing happens

static void bi_declare(chunk *current, interpreter *i, value *args)
{
	parse_declaration(i, DS_BODY(args->val), 0, DS_BODY(next_value(args)->val), NULL);
}

/// override(name declspec)
// declares a procedure \em[name] according to \em[declspec] overriding the previous 
// declaration if present
static void bi_override(chunk *current, interpreter *i, value *args)
{
	parse_declaration(i, DS_BODY(args->val), 1, DS_BODY(next_value(args)->val), NULL);
}

/// declof(name)
// \returns the declaration specification for \em[name]
static char *unique_name(void);
static void bi_declof(chunk *current, interpreter *i, value *args)
{
	ds_append(pstack->state->buffer, stringize_declaration(lookup_declaration(i, DS_BODY(args->val))));
}

/// undeclare(name)
// removes the most recent declaration of \em[name]. If there was an overriden declaration 
// for \em2[name][,] it becomes visible
static void bi_undeclare(chunk *current, interpreter *i, value *args)
{
	i->decls = list_remove(i->decls, DS_BODY(args->val), (list_predicate_t)decl_find, 
							(list_destroy_t)destroy_decl, sizeof(*i->decls));
}

/// release(kind int-name)
// decrements the reference counter of an object \em[int-name] of the
// \em2[kind][.] If the counter reaches zero, the object is destroyed.
static void bi_release(chunk *current, interpreter *i, value *args)
{
	shared_int *si = find_si_itself(i, DS_BODY(args->val), DS_BODY(next_value(args)->val));
	shared_int **base;
	if(!si)
		recover_error("%s '%s' undefined", DS_BODY(args->val), DS_BODY(next_value(args)->val));
	base = (shared_int **)((char *)i + si->so->class->offs2);
	*base = list_remove(*base, si, (list_predicate_t)find_byref, (list_destroy_t)destroy_shared_int, sizeof(shared_int));
}

/// gensym()
// \returns an unique identifier
static void bi_gensym(chunk *current, interpreter *i, value *args)
{
	ds_appendstr(pstack->state->buffer, unique_name());
}

/// load(script-file)
// loads the \em[script-file] in the same way as if being included by
// \em[.include] directive. 
// \note This command affects the \em[current] execution environment;
// to create a new environment, use \em2[subload][.]
static void parse_script(interpreter *i, char *start);
static void bi_load(chunk *current, interpreter *i, value *args)
{
	int lno = set_line_no(1);
	char *body;
	body = load_file(DS_BODY(args->val), getenv(SCRIPT_PATH_ENV), "sc");
	parse_script(i, body);
	free(body);
	set_line_no(lno);
}

/// subload(int-name script-file)
// Creates a new execution environment with the name \em[int-name] and loads
// a script from \em[script-file] to it. 
// \note An execution environment may be restricted only \em[after]
// creating, so the \em[script-file] will have access to \em[all] builtins
// of the NSL. 
// \tip You can load a script within a restricted environment with
// following trick:
//
// subload some-name empty 
//
// restrict <all-necessary-functions> load
// 
// callin some-name load untrusted-script
//
// Using load cannot break security, so it is safe to allow it even in
// untrusted scripts.
static void bi_subload(chunk *current, interpreter *i, value *args)
{
	char *name = DS_BODY(args->val);
	interpreter *subint;
	if(!*name || *name == '*')
		name = unique_name();
	if(list_find(i->subints, name, (list_predicate_t)si_find_exact))
		recover_error("interpreter '%s' already loaded", name);
	subint = load_interpreterx(DS_BODY(next_value(args)->val), i, NULL, NULL);
	add_shared_int(i, init_shared_subint(i, subint, name));
	ds_appendstr(pstack->state->buffer, name);
}

/// module(modulename)
// Loads a dynamically pluggable module \em[modulename] (if not yet loaded)
// and attach it to the current environment (if not yet attached)
// \note All module's builtins will be available only thru subsequent
// \em[call] and \em[eval] calls
static void bi_module(chunk *current, interpreter *i, value *args)
{
#ifdef USE_LTDL	
	attach_module(i, DS_BODY(args->val));
#endif		
}

/// pragma(module arg)
// Sends a pragma \em[arg] to a \em[module]
static void bi_pragma(chunk *current, interpreter *i, value *args)
{
#ifdef USE_LTDL	
	module_pragma(i, DS_BODY(args->val), DS_BODY(next_value(args)->val));
#endif		
}

/// callin(subint proc-name [arg ...])
// Calls a \em[proc-name] with \em2[arg]s within the execution environment
// \em[subint] where \em[proc-name] may be either a user-defined procedure
// or a builtin. Restrictions of the \em[subint] do apply.
// \returns the result of executing \em[proc-name]
static void bi_call(chunk *current, interpreter *i, value *args);
static void bi_callin(chunk *current, interpreter *i, value *args)
{
	interpreter *subint = find_subint(i, DS_BODY(args->val));
	bi_call(NULL, subint, next_value(args));
}

instruction_src *copy_builtins(instruction_src *src);
/// restrict(subint [!]bi-pattern ...)
// removes from the \em[subint] all builtins which do not match any of
// \em2[bi-pattern][s.] Patterns are the same as in \em[match] etc. If a
// pattern is prepended by !, it is inverted
static void bi_restrict(chunk *current, interpreter *i, value *args)
{
	interpreter *subint = ds_isempty(args->val) ? i : find_subint(i, DS_BODY(args->val));
	instruction_src *iter;
	value *aiter;
	char *pattern;
	int not_flag;
	int deny = 0;
	
	for(iter = subint->builtins; iter->name; iter++)
	{
		deny = 1;
		for(aiter = next_value(args); aiter; aiter = next_value(aiter))
		{
			not_flag = 0;
			pattern = DS_BODY(aiter->val);
			if(*pattern == '!')
			{
				not_flag = 1;
				pattern++;
			}
			if(not_flag)
			{
				if(!fnmatch(pattern, iter->name, 0))
				{
					deny = 1;
					break;
				}
			}
			else if(!fnmatch(pattern, iter->name, 0))
				deny = 0;
		}
		if(deny)
			iter->name = "";
	}
}

/// revoke(subint bi-pattern ...)
// removes from the \em[subint] all builtins which do match any of
// \em2[bi-pattern][s.] Patterns are the same as in \em[match] etc. 
static void bi_revoke(chunk *current, interpreter *i, value *args)
{
	interpreter *subint = ds_isempty(args->val) ? i : find_subint(i, DS_BODY(args->val));
	instruction_src *iter;
	value *aiter;
	
	for(iter = subint->builtins; iter->name; iter++)
	{
		for(aiter = next_value(args); aiter; aiter = next_value(aiter))
		{
			if(!fnmatch(DS_BODY(aiter->val), iter->name, 0))
			{
				iter->name = "";
				break;
			}
		}
	}
}


///
// trace() or trace(arg ...)
// enables tracing procedures whose names match one of \em2[arg][s]; 
// an argument \em[#] enables line tracing;
// no arguments is equivalent to '*'
static void bi_trace(chunk *current, interpreter *i, value *args)
{
	tracelist *nt;
	if(!args)
	{
		nt = ALLOC(tracelist);
		nt->pattern = xstrdup("*");
		i->to_trace = list_add(i->to_trace, nt);
	}
	else
	{
		for(; args; args = list_next(args))
		{
			if(*DS_BODY(args->val) == '#')
				i->line_tracing = 1;
			else
			{
				nt = ALLOC(tracelist);
				nt->pattern = xstrdup(DS_BODY(args->val));
				i->to_trace = list_add(i->to_trace, nt);
			}
		}
	}
}

static void trace_destroy(tracelist *t)
{
	free(t->pattern);
}

static int trace_find1(tracelist *t, char *key)
{
	return !strcmp(t->pattern, key);
}

///
// notrace() or notrace(arg1 ...)
// disables tracing procedures with names matching \em2[arg1][...]
// without arguments, disables tracing for all procedures
// \note procedure name patterns should coincide with those in 'trace'
static void bi_notrace(chunk *current, interpreter *i, value *args)
{
	if(!args)
	{
		list_delete(i->to_trace, (list_destroy_t)trace_destroy, sizeof(tracelist));
		i->to_trace = NULL;
	}
	else
	{
		for(; args; args = list_next(args))
		{
			if(*DS_BODY(args->val) == '#')
				i->line_tracing = 0;
			else
			{
				i->to_trace = list_remove(i->to_trace, DS_BODY(args->val), 
									(list_predicate_t)trace_find1, 
									(list_destroy_t)trace_destroy,
									sizeof(tracelist)
									);
			}
		}
	}
}

///
// dispatch
// starts the handler for the current chunk
// \error[nothing to dispatch]
static void bi_dispatch(chunk *current, interpreter *i, value *args)
{
	shared_int *si_proc;
	procedure *proc;
	static shared_int faked_proc;
	dstring *ds = args ? args->val : ds_create(i->prefix);
	if(!i->this_chunk)
		recover_error("nothing to dispatch");
	if(HTML_CHUNK_TYPE(i->this_chunk) == hct_char)
	{
		int ch;
		ds_appendch(ds, ch = get_html_char(i->this_chunk));
		if(!(si_proc = i->char_cache[ch]))
 		{
			si_proc = list_find(i->filters, DS_BODY(ds), (list_predicate_t)si_find_match);
			i->char_cache[ch] = si_proc ? si_proc : &faked_proc;
		}
	}
	else
	{
		ds_appendstr(ds, get_tag_name(i->this_chunk));
		si_proc = list_find(i->tags, DS_BODY(ds), (list_predicate_t)si_find_match);
	}
//	bi_exit(current, i, args);
	i->already_read = 0;
	if(si_proc && si_proc != &faked_proc)
	{
		proc = si_proc->obj;
		if(proc->body)
			eval_proc(proc->sh.owner, proc->body, si_proc->intname, proc->pool, NULL);
	}
}

/// 
// prefix() or prefix(arg)
// if \em[arg] is present, sets the current dispatching prefix to \em[arg]
// \returns previous DP
static void bi_prefix(chunk *current, interpreter *i, value *args)
{
	if(i->prefix)
		ds_appendstr(pstack->state->buffer, i->prefix);
	if(args)
	{
		char *oldprefix = i->prefix;
		int k;
		for(k = 0; k <= UCHAR_MAX; k++)
			i->char_cache[k] = NULL;
		i->prefix = xstrdup(DS_BODY(args->val));
		if(oldprefix)
			free(oldprefix);
	}
}

/// echo(arg1 ...)
// prints \em2[arg1][...] to stderr. If the last argument ends with \\r, 
// no carriage return occurs after printing the message.
// \note top level only
static void bi_echo(chunk *current, interpreter *i, value *args)
{
	for(; list_next(args); args = list_next(args))
		fputs(DS_BODY(args->val), stderr);
	if(DS_BODY(args->val)[ds_length(args->val) - 1] == '\r')
	{
		dstring *tmp = ds_substr(args->val, 0, ds_length(args->val) - 1);
		fputs(DS_BODY(tmp), stderr);
	}
	else
	{
		fputs(DS_BODY(args->val), stderr);
		fputc('\n', stderr);
	}
}

///
// error(arg1 ...)
// generates a recoverable error 
// error message is concatenated from \em2[arg1][...] 
// \note top level only
static void bi_error(chunk *current, interpreter *i, value *args)
{
	dstring *temp = ds_create(NULL);
	for(; args; args = list_next(args))
		ds_append(temp, args->val);
	recover_error("%s", DS_BODY(temp));
}

///
// fatal-error(arg1 ...)
// generates a fatal error
// error message is concatenated from \em2[arg1][...]
// \note top level only
static void bi_fatal_error(chunk *current, interpreter *i, value *args)
{
	dstring *temp = ds_create(NULL);
	for(; args; args = list_next(args))
		ds_append(temp, args->val);
	fatal_error("%s", DS_BODY(temp));
}

/// 
// pass-error arg
// passes an error \em3['][arg]['] to the built-in error handler
// useful within a user error handler
// \note top level only
static void bi_pass_error(chunk *current, interpreter *i, value *args)
{
	unwind_error();
	while(pstack && !pstack->boundary)
		pstack = list_pop(pstack, (list_destroy_t)destroy_procframe, sizeof(procstack));
	if(pstack)
		pstack->boundary = 0;
	recover_error("%s", DS_BODY(args->val));
}

/// 
// onerror() or onerror(arg)
// if \em[arg] is present, makes the procedure \em[arg] a user error handler
// \returns the name of the previous error handler
static void bi_onerror(chunk *current, interpreter *i, value *args)
{
	shared_int *si = i->error_handler ? list_find(i->procs, i->error_handler, (list_predicate_t)si_find_byso) : NULL;
	char *oldname = si ? si->intname : "";
	if(args)
	{
		shared_int *ehsi;
		if(ds_isempty(args->val))
			i->error_handler = NULL;
		else
		{
			ehsi = list_find(i->procs, DS_BODY(args->val), (list_predicate_t)si_find_exact);
			if(!ehsi)
				recover_error("undefined error handler: '%s'", DS_BODY(args->val));
			i->error_handler = ehsi->obj;
		}
	}
	ds_appendstr(pstack->state->buffer, oldname);
}

/// 
// cancel and cancel(proc)
// passes the control to the next top-level instruction of the calling
// function or to the caller of the calling function, if `proc' is present. 
// \note Since errors may occur both at parse- and run-time, you should
// think a lot before using this function in an error handler which may be
// called from within \em[eval] and the like, since the point where
// execution goes depends on the kind of error. You'd better use \em[catch]
// ... \em2[throw][.]
// \use[tlo]
static void bi_cancel(chunk *current, interpreter *i, value *args)
{
	pstack = list_pop(pstack, (list_destroy_t)destroy_procframe, sizeof(procstack));
	if(pstack && !pstack->boundary)
	{
		if(!args || strcmp(DS_BODY(args->val), "proc"))
		{
			list_delete(pstack->state, (list_destroy_t)destroy_execstate, sizeof(execstack));
			pstack->state = NULL;
		}
		else
			pstack = list_pop(pstack, (list_destroy_t)destroy_procframe, sizeof(procstack));
	}
}

/// catch(arg) and throw(throw-arg)
// catch simply returns its \em[arg] if no \em[throw] was called during
// \em[arg] evaluation. Otherwise, \em[throw-arg] is returned.
//
// \em[throw] sets the executiomn point to the innermost \em[catch] in the
// execution chain, making it return \em2[throw-arg][.]
// \error[uncaught throw 'throw-arg']

static void bi_catch(chunk *current, interpreter *i, value *args)
{
	ds_append(pstack->state->buffer, args->val);
}

static chunk *throw_chunk;

static void bi_throw(chunk *current, interpreter *i, value *args)
{
	for(;;)
	{
		if(!pstack)
			recover_error("uncaught throw '%s'", DS_BODY(args->val));
		if(pstack->boundary)
		{
			list_delete(throw_chunk, (list_destroy_t)destroy_chunk, sizeof(chunk));
			throw_chunk = ALLOC(chunk);
			throw_chunk->type = arg_builtin;
			throw_chunk->x.func.f.bi = bi_throw;
			throw_chunk->x.func.args = ALLOC(argument);
			throw_chunk->x.func.args->arg = ALLOC(chunk);
			throw_chunk->x.func.args->arg->type = arg_literal;
			throw_chunk->x.func.args->arg->x.literal = ds_fix(args->val);
			eval_funcall(i, throw_chunk);
			return;
		}
		while(pstack->state)
		{
			if(pstack->state->curfunc && pstack->state->curfunc->type == arg_builtin 
					&& pstack->state->curfunc->x.func.f.bi == bi_catch)
			{
				pstack->state->curop = NULL;
				ds_assign(pstack->state->buffer, args->val);
				return;
			}
			pstack->state = list_pop(pstack->state, (list_destroy_t)destroy_execstate, sizeof(execstack)); 
		}
		pstack = list_pop(pstack, (list_destroy_t)destroy_procframe, sizeof(procstack));
	}
}

/// trap(signame handler)
// sets the handler of a signal \em[signame] to handler. 
// If \em[handler] is empty, the previous handler is restored.
// If \em[handler] is \em3[`][ignore]['], the signal will be ignored;
// if \em[handler] is \em3[`][default]['], the default behaviour is set.
static void bi_trap(chunk *ch, interpreter *i, value *args)
{
	procedure *p;
	trap_handler *h;
	int behaviour;
	if(ds_isempty(next_value(args)->val))
	{
		all_traps = list_remove(all_traps, DS_BODY(args->val), (list_predicate_t)trap_find,
								(list_destroy_t)trap_destroy, sizeof(trap_handler));
		h = list_find(all_traps, DS_BODY(args->val), (list_predicate_t)trap_find);
		if(!h || h->proc == (procedure *)SIG_DFL)
			set_trap(DS_BODY(args->val), trap_default);
		else if(h->proc == (procedure *)SIG_IGN)
			set_trap(DS_BODY(args->val), trap_ignore);
		else
			set_trap(DS_BODY(args->val), trap_catch);
	}
	else 
	{
		h = ALLOC(trap_handler);
		if(!strcmp(DS_BODY(next_value(args)->val), "ignore"))
		{
			p = (procedure *)SIG_IGN;
			behaviour = trap_ignore;
		}
		else if(!strcmp(DS_BODY(next_value(args)->val), "default"))
		{
			p = (procedure *)SIG_DFL;
			behaviour = trap_default;
		}
		else
		{
			p = find_proc(i, DS_BODY(next_value(args)->val));
			behaviour = trap_catch;
		}
		h->trapname = xstrdup(DS_BODY(args->val));
		h->proc = p;
		all_traps = list_add(all_traps, h);
		set_trap(DS_BODY(args->val), behaviour);
	}
}


/// stop
// stops the current process (i. e the interpreter itself)
// \note the behaviour of this builtin depends on how the system handles
// SIGSTOP. On non-Unix systems this may do nothing or terminate the
// process.
static void bi_stop(chunk *current, interpreter *i, value *args)
{
	raise(SIGSTOP);
}

/// 
// write(arg1 ...)
// writes \em2[arg1][...] to the current output stream
// \error[output stream not defined]
// \note top level only
static void write_to(interpreter *i, stream *s, value *args)
{
	for(; args; args = list_next(args))
		s->class->writestr(i, s, DS_BODY(args->val));
}

static void bi_write(chunk *current, interpreter *i, value *args)
{
	if(!i->outstream)
		recover_error("output stream not defined");
	write_to(i, i->outstream, args);
}

/// writeto(stream arg1 ...)
// like \em2[write][,] but uses \em[stream] instead of the current output.
static void bi_writeto(chunk *current, interpreter *i, value *args)
{
	write_to(i, find_stream(i, DS_BODY(args->val)), next_value(args));
}

///
// writetag() or writetag(tag-name [param-name1 param-val1 ...])
// with no arguments, echoes the current input chunk to the output stream;
// with arguments, writes a tag named \em[tag-name] with parameters
// \em[param-name1] = \em[param-val1] to the output stream
// \error[output stream not defined]
// \error[no chunk to echo]

static void write_tag_to(interpreter *i, stream *s, value *args)
{
	static html_chunk hc; 
	if(!args)
	{
		if(!i->this_chunk)
			recover_error("no chunk to echo");
		s->class->writetag(i, s, i->this_chunk);
		return;
	}
	set_tag_name(&hc, args->val);
	for(args = list_next(args); args; args = next_value(args) ? third_value(args) : NULL)
		set_html_param(&hc, args->val, next_value(args) ? next_value(args)->val : NULL);
	s->class->writetag(i, s, &hc);
	release_html_chunk(&hc);
}

static void bi_writetag(chunk *current, interpreter *i, value *args)
{
	if(!i->outstream)
		recover_error("output stream not defined");
	write_tag_to(i, i->outstream, args);
}

/// writetagto(stream tag-name [param-name1 param-val1 ...])
// like \em2[writetag][2], but uses \em[stream] instead of the current
// output.
static void bi_writetagto(chunk *current, interpreter *i, value *args)
{
	write_tag_to(i, find_stream(i, DS_BODY(args->val)), next_value(args));
}

///
// ioctl(stream ctl-str)
// sends a control string \em[ctl-str] to the \em2[stream][.]
// The exact effect is determined by the stream type
// \returns a stream-specific value
// \error[don't know how to process request 'ctl-str']
static void bi_ioctlin(chunk *current, interpreter *i, value *args)
{
	stream *s;
	char *val;
	
	s = find_stream(i, DS_BODY(args->val));
	val = s->class->ioctl(i, s, DS_BODY(next_value(args)->val));
	if(!val)
		recover_error("don't know how to process request '%s'", DS_BODY(next_value(args)->val));
	ds_appendstr(pstack->state->buffer, val);
}

///
// getpos(stream)
// \returns a current input position in the \em[stream]
// \error[stream doesn't support repositioning]
// \note the value returned by the function can only be used in 
// a subsequent 'setpos' call
static void bi_getpos(chunk *current, interpreter *i, value *args)
{
	stream *s = find_stream(i, DS_BODY(args->val));
	int p;
	p = s->class->getpos(i, s);
	if(p < 0)
		recover_error("%s stream doesn't support repositioning", DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(p, 10, 0));
}

///
// setpos(stream pos)
// sets the current input position of the \em[stream] to \em[pos]
// \note only values returned by 'getpos' can safely be used here
// \note top level only
// \error[input stream not defined]
static void bi_setpos(chunk *current, interpreter *i, value *args)
{
	stream *s = find_stream(i, DS_BODY(args->val));
	int p;
	p = convert_to_int(DS_BODY(next_value(args)->val));
	s->class->setpos(i, s, p);
}

///
// rewind(stream)
// resets the current input position of the \em[stream] to the beginning
// of the stream
// \uses[tlo]
static void bi_rewind(chunk *current, interpreter *i, value *args)
{
	stream *s = find_stream(i, DS_BODY(args->val));
	
	s->class->rewind(i, s);
}

// Functions


void define_var(interpreter *i, int status, char *name, dstring *val)
{
	variable *var = NULL;
	
	if(pstack)
		var = list_find(pstack->vars, name, (list_predicate_t)var_find);
	if(var)
		recover_error("redefinition of local '%s'", name);
	if(status == var_global)
	{
		var = list_find(i->globvars, name, (list_predicate_t)var_find);
		if(var)
			recover_error("redefinition of global '%s'", name);
	}
	else if(status == var_shared)
	{
		if(!pstack->cur_pool)
			recover_error("no current pool");
		if(list_find(pstack->cur_pool->vars, name, (list_predicate_t)var_find))
			return;
	}
	var = ALLOC(variable);
	var->name = xstrdup(name);
	var->vals = ALLOC(value);
	var->vals->val  = val ? ds_fix(val) : ds_fix(ds_create(NULL));
	if(status == var_global)
		i->globvars = list_add(i->globvars, var);
	else if(status == var_shared)
		pstack->cur_pool->vars = list_add(pstack->cur_pool->vars, var);
	else pstack->vars = list_add(pstack->vars, var);
}

static void define_vars(interpreter *i, int global, value *args)
{
	value *temp;
	dstring *inival;
	for(; args; args = list_next(temp))
	{
		inival = NULL;
		temp = args;
		if(list_next(args) && !strcmp(DS_BODY(next_value(args)->val), "="))
		{
			temp = third_value(args);
			if(!temp)
				recover_error("missing right side of '='");
			inival = temp->val;
		}
		define_var(i, global, DS_BODY(args->val), inival);
	}
}

/// 
// def(var-name1 [= var-value1] ...)
// defines local variables \em2[var-name1][...] ands sets the value to 
// \em2[var-value1][...]
// \uses[tlo]
// \error[redefinition of local 'varname']
static void bi_def(chunk *current, interpreter *i, value *args)
{
	define_vars(i, var_local, args);
}

/// 
// global(var-name1 [= var-value1] ...)
// defines global variables \em2[var-name1][...] ands sets the value 
// to \em2[var-value1][...]
// \error[redefinition of local 'varname']
// \error[redefinition of global 'varname']

static void bi_global(chunk *current, interpreter *i, value *args)
{
	define_vars(i, var_global, args);
}

/// globalin(subint var-name1 [= var-value1] ...)
// defines global variables with an execution environment \em[subint] no
// matter whether it allows the 'global' builtin. In other respects is
// analogous to \em2[global][.]
static void bi_globalin(chunk *current, interpreter *i, value *args)
{
	interpreter *sub = find_subint(i, DS_BODY(args->val));
	define_vars(sub, var_global, list_next(args));
}

/// 
// shared(var-name1 [= var-value1] ...)
// defines shared variables \em2[var-name1][...] within the current pool 
// and sets the value to \em2[var-value1][...]
// \note unlike \em[def] and \em2[global][,] \em[shared] is idempotent, i.\ e. 
// if \em[var-name] is already defined, nothing happens.
// \uses[tlo]
// \error[redefinition of local 'varname']
// \error[no current pool]

static void bi_shared(chunk *current, interpreter *i, value *args)
{
	define_vars(i, var_shared, args);
}

///
// undef(var-name1 ...)
// undefines \em[local] variables with names \em[var-name1] ...
// \note \em[undef] says nothing about variables already undefined
// \note top level only
static void bi_undef(chunk *current, interpreter *i, value *args)
{
	for(; args; args = list_next(args))
		pstack->vars = list_remove(pstack->vars, DS_BODY(args->val), (list_predicate_t)var_find, (list_destroy_t)destroy_variable, sizeof(variable));
}

///
// defined(var-name)
// \returns true if \em[var-name] is defined and available at the point of a call
static void bi_defined(chunk *current, interpreter *i, value *args)
{
	variable *v = get_var(i, DS_BODY(args->val));
	if(v)
		ds_appendstr(pstack->state->buffer, "true");
}

///
// isglobal(var-name)
// \returns true if \em[var-name] is a global variable
static void bi_isglobal(chunk *current, interpreter *i, value *args)
{
	variable *v = list_find(i->globvars, DS_BODY(args->val), (list_predicate_t)var_find);
	if(v)
		ds_appendstr(pstack->state->buffer, "true");
}

///
// readonly(var1 ... varn)
// makes variables \em[var1 ... varn] read-only
static void bi_readonly(chunk *current, interpreter *i, value *args)
{
	variable *var;
	for(; args; args = next_value(args))
	{
		var = safe_get_var(i, DS_BODY(args->val), 0);
		var->readonly = 1;
	}
}

static dstring *get_value(value *args)
{
	if(!args)
		return NULL;
	if(list_next(args) && !strcmp(DS_BODY(args->val), "="))
		args = list_next(args);
	return args->val;
}

///
// set(var-name [=] var-value)
// sets the variable \em[var-name] to \em[var-value]
// \returns \em[var-value]
// \error[variable 'var-name' not defined]
// \error[missing 'set' value ]
static void bi_set(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(args->val), 1);
	dstring *nv;
	nv = get_value(list_next(args));
	if(!nv)
		recover_error("missing 'set' value");
	ds_assign(v->vals->val, nv);
	ds_append(pstack->state->buffer, nv);
}

///
// push(var-name [[=] var-value])
// pushes the current value of the variable \em[var-name] and 
// sets it to \em[var-value] or null if var-value is missing
// \returns var-value
// \error[variable 'var-name' not defined]
static void bi_push(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(args->val), 1);
	dstring *nv;
	value *nval;

	nv = get_value(list_next(args));
	nval = ALLOC(value);
	nval->val = !nv ? ds_fix(ds_create(NULL)) : ds_fix(nv);
	v->vals = list_add(v->vals, nval);
	if(nv)
		ds_append(pstack->state->buffer, nv);
}

///
// pop(var-name)
// replaces the curent value of \em[var-name] with the last pushed value
// or null if the stack is empty
// \returns the new value of \em[var-name]
// \error[variable 'var-name' not defined]
static void bi_pop(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(args->val), 1);
	
	if(list_next(v->vals))
		v->vals = list_pop(v->vals, (list_destroy_t)destroy_value, sizeof(value));
	else ds_assign(v->vals->val, ds_create(NULL));
	ds_append(pstack->state->buffer, v->vals->val);
}

///
// clear(var-name)
// clears the internal stack of \em[var-name]
// \note top level only
// \error[variable 'var-name' not defined]
static void bi_clear(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(args->val), 1);
	
	list_delete(v->vals, (list_destroy_t)destroy_value, sizeof(value));
	v->vals = ALLOC(value);
	v->vals->val = ds_fix(ds_create(NULL));
}

/// depth(var-name)
// \returns the depth of the internal stack of \em2[var-name][,] including the
// current visible value
// \error[variable 'var-name' not defined]
static void bi_depth(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(args->val), 0);
	
	ds_append(pstack->state->buffer, ds_fromint(list_length(v->vals), 10, 0));
}

/// copy(dest-var src-var)
// copies the whole internal stack of \em[src-var] to \em2[dest-var][.]
// the stack of \em[src-var] is completely lost
// \note top level only
// \error[variable 'dest-var or src-var' not defined]
static void bi_copy(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(next_value(args)->val), 0);
	variable *dest = safe_get_var(i, DS_BODY(args->val), 1);
	value *nvs = NULL, *nv;
	value *iter;
	
	for(iter = v->vals; iter; iter = list_next(iter))
	{
		nv = ALLOC(value);
		nv->val = ds_fix(ds_copy(iter->val));
		nvs = list_add(nvs, nv);
	}
	nvs = list_reversip(nvs);
	list_delete(dest->vals, (list_destroy_t)destroy_value, sizeof(value));
	dest->vals = nvs;
}

/// revstack(var-name)
//  reverses the internal stack of \em[var-name]
//  \use[tlo]
// \error[variable 'var-name' not defined]
static void bi_revstack(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(args->val), 1);
	
	v->vals = list_reversip(v->vals);
}

/// exch(var-name)
// exchanges two topmost items in the stack of \em[var-name]
// \error[variable 'var-name' not defined]
static void bi_exch(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(args->val), 1);
	
	if(next_value(v->vals))
	{
		dstring *temp = v->vals->val;
		v->vals->val = next_value(v->vals)->val;
		next_value(v->vals)->val = temp;
	}
}

/// iseof()
// \returns true, if there is no current chunk
static void bi_iseof(chunk *current, interpreter *i, value *args)
{
	if(!i->this_chunk)
		ds_appendstr(pstack->state->buffer, "true");
}

static char *my_readline(char *prompt)
{
	char buffer[256];
	char *buf_pos = buffer;
	int ch;
	
	fputs(prompt, stdout);
	fflush(stdout);
	for(;;)
	{
		ch = getchar();
		if(ch == EOF || ch == '\n')
		{
			*buf_pos = '\0';
			return xstrdup(buffer);
		}
		if(ch == '\b')
		{
			if(buf_pos != buffer)
				buf_pos--;
		}
		else if(buf_pos == buffer + sizeof(buffer) - 2)
			fputc('\a', stderr);
		else *buf_pos++ = ch;
	}
}

/// readline(prompt)
//  shows a prompt and waits for a user to type in a string
//  \returns a user string
//  \note This builtin only simulates the true readline. 
//  If you have the GNU readline, you may access it via \em[readline]
//  module.
static void bi_readline(chunk *current, interpreter *i, value *args)
{
	char *prompt = args ? DS_BODY(args->val) : "";
	char *in = my_readline(prompt);
	if(in)
	{
		ds_appendstr(pstack->state->buffer, in);
		free(in);
	}
}

/// read[(autoseq)]
//  reads the next token from the input stream and translates it with the
//  \em2[autoseq][.] If \em[autoseq] is not given, the identical
//  translations is accepted. The read process stops when a tag is
//  encountered.
//  \returns the translated sequence
//  \error[no input stream]
//  \error[translation undefined: 'autoseq']

static int translate_input(dstring *ns, autoseq *as, int ch);
static int read_from(interpreter *i, stream *s, value *args)
{
	autoseqdef *asd;
	dstring *ns;
	int result;
	int kept = 0;
	int first = 1;

	asd = find_asd(i, args ? DS_BODY(args->val) : "");
	ns = ds_create(NULL);
	for(;;)
	{
		release_html_chunk(i->this_chunk);
		i->this_chunk = copy_html_chunk(s->class->read_chunk(i, s));
		if(!i->this_chunk || HTML_CHUNK_TYPE(i->this_chunk) != hct_char)
		{
			translate_input(ns, asd->seq, 0);
			kept = !first;
			break;
		}
		result = translate_input(ns, asd->seq, get_html_char(i->this_chunk));
		if(result == ar_rejected)
		{
			kept = 1;
			break;
		}
		if(!asd->seq || is_in_final(&asd->seq->state))
			break;
		first = 0;
	}
	ds_append(pstack->state->buffer, ns);
	return kept;
}

static void bi_read(chunk *current, interpreter *i, value *args)
{
	if(!i->instream)
		recover_error("input stream not defined");
	if(read_from(i, i->instream, args))
		i->already_read = 1;
}

/// readfrom(stream [autoseq])
// like \em[read] but uses \em[stream] instead of the current input
// \note there is a single current chunk for all the streams, so a call to
// \em[readfrom] will replace the chunk read by e. g. main reading loop.
static void bi_readfrom(chunk *current, interpreter *i, value *args)
{
	read_from(i, find_stream(i, DS_BODY(args->val)), next_value(args));
}

static char *unique_name(void)
{
	static char buffer[64];
	static int counter = 0;
	sprintf(buffer, "#%8.8x%8.8x%8.8x", getpid(), (int)time(NULL), counter++);
	return buffer;
}

storage *create_storagex(interpreter *i, storage_class *class, char **name, char *arg)
{
	storage *s;
	if(!*name || !**name || (**name == '*' && !(*name)[1]))
		*name = unique_name();
	if(list_find(i->stos, *name, (list_predicate_t)si_find_exact))
		recover_error("storage '%s' already defined", *name);
	s = class->create(i, *name, arg);
	add_shared_int(i, init_shared_storage(i, s, *name));
	return s;
}

storage *create_storage(interpreter *i, storage_class *class, char *name, char *arg)
{
	storage *s = create_storagex(i, class, &name, arg);
	ds_appendstr(pstack->state->buffer, name);
	return s;
}

/// array[(name [val1 ...])]
//  creates an array named \em2[name][.] If \em[name] is missing, is empty, 
//  or is equal to '*', a unique name is generated
//  \returns the name of the array
//  \error[storage 'name' already defined]
static void bi_array(chunk *current, interpreter *i, value *args)
{
	storage *s = create_storage(i, &array_storage_class, args ? DS_BODY(args->val) : NULL, NULL);
	int j = 0;
	for(args = list_next(args); args; args = list_next(args))
		s->class->add(i, s, ds_fromint(j++, 10, 0), args->val);
}

/// table(name extname)
// creates a table named \em2[name][.] See \em2[array][.] and links it
// with a Berkley DB file named \em[extname] (Btree format is used)
// If \em[extname] has no extenstion, \em[.t] is assumed.
// If \em[extname] is not absolute, the file is search in the current 
// directory and in the directory given by SL_TABLE_PATH (if present).
// If the first character of \em[extname] is '!', the new file is created
// even if it existed before.
// If the first character of \em[extname] is '+', the table file is opened
// read-only
// 
// If \em[extname] is null, the table exists solely in memory
//
// Tables support two special items: \em[__readonly__] which disables
// modifying a table (note that setting __readonly__ cannot be undone); 
// and \em[__reversed__] which controls the traversal order: ascending, when
// null, and descending otherwise. This parameter may be changed at any
// time, even in the middle of enumerating.
// \note In versions prior to 0.8 GDBM format was used
static void bi_table(chunk *current, interpreter *i, value *args)
{
	create_storage(i, &table_storage_class, DS_BODY(args->val), DS_BODY(next_value(args)->val));
}

/// list([name1 [= value1] ...])
// creates a property list with a unique name. Lists allow any data as keys,
// like tables, but they are not persistent; in addition, unlike other
// storages, property lists may store multiple values of the same key though
// only the last stored one is accessible.
// \returns the name of the storage created
static void bi_list(chunk *current, interpreter *i, value *args)
{
	storage *s = create_storage(i, &plist_storage_class, NULL, NULL);
	value *temp;
	dstring *inival;
	
	for(; args; args = list_next(temp))
	{
		inival = NULL;
		temp = args;
		if(list_next(args) && !strcmp(DS_BODY(next_value(args)->val), "="))
		{
			temp = third_value(args);
			if(!temp)
				recover_error("missing right side of '='");
			inival = temp->val;
		}
		s->class->add(i, s, args->val, inival ? inival : ds_create(NULL));
	}
}

/// sequence(seq-def)
// generates a sequence of integers which fits a range definition given in
// \em2[seq-def][.] The syntax of \em[seq-def] is that of ranges in automata
// defintions (see \ref[sla] ).
// \note This function is constant, unlike other storage constructors, so
// you should not use it as an argument to \em[local] if it have a constant
// argument.
static void bi_sequence(chunk *current, interpreter *i, value *args)
{
	create_storage(i, &seq_storage_class, NULL, DS_BODY(args->val));
}

/// iota(int-num)
// this is equivalent to \em[sequence('\[1-int-num\]')]
static void bi_iota(chunk *current, interpreter *i, value *args)
{
	dstring *temp = ds_create("[1-");
	ds_append(temp, args->val);
	ds_appendch(temp, ']');
	create_storage(i, &seq_storage_class, NULL, DS_BODY(temp));
}

/// new(subint [arg ...])
// creates a new instance of a user-defined storage whose handlers are in
// the execution environment \em[subint]. \em2[arg][s] are passed to the
// \em[init] function of the storage.
static void bi_new(chunk *current, interpreter *i, value *args)
{
	storage *s = create_storage(i, &uds_storage_class, NULL, DS_BODY(args->val));	
	uds_init(i, s, next_value(args));
}

/// local(storage)
// make the \em[storage] local, that is it automatically released when
// leaving the current block.
static void bi_local(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	s->local_to = pstack;
	ds_append(pstack->state->buffer, args->val);
}

/// get(storage key)
//  \returns a \em[storage] record with the \em[key]
//  \error[depends on a storage type]
static void bi_get(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	ds_append(pstack->state->buffer, s->class->get(i, s, next_value(args)->val));
}

/// check(storage key)
// \returns true if a record \em[key] is in \em[storage]
static void bi_check(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	if(s->class->check(i, s, next_value(args)->val))
		ds_appendstr(pstack->state->buffer, "true");
}

/// put(storage key val)
// puts a pair \em[key - val] into \em2[storage][.]
// For most types of storages, if a \em[key] is not already in
// \em[storage], an error occurs.
static void bi_put(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	s->class->put(i, s, next_value(args)->val, 
				  third_value(args)->val
				 );
}

/// insert(storage key val)
/// adds a pair \em[key - val] into \em2[storage][.]
// For most types of storages, if a \em[key] is \em[already] in
// \em[storage], an error occurs.
static void bi_insert(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	s->class->add(i, s, next_value(args)->val, 
				  third_value(args)->val
				 );
}

/// remove(storage key)
// removes a pair with the \em[key] from the \em[storage].
// For property list, this may give access to a previously stored key.
static void bi_remove(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	s->class->remove(i, s, next_value(args)->val);
}

/// apply(storage func [args ...])
// applies a user-defined function \em[func] to all the items of the
// \em2[storage][.] The function is passed \em[storage], a current item and
// all the \em2[arg][s.] 
// \returns concatenation of returned values for each item
// \note a builtin may not be used in place of \em[func][.]
static void bi_apply(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	dstring *key, *res;
	void *data;
	value *nargs = NULL, *arg1;
	
	key = s->class->first(i, s, &data);
	while(key)
	{
		nargs = ALLOC(value);
		nargs->val = ds_fix(ds_copy(args->val));
		arg1 = ALLOC(value);
		arg1->val = ds_fix(ds_copy(key));
		nargs = list_append(nargs, list_append(arg1, copy_values(third_value(args))));
		res = eval_procedure(i, DS_BODY(next_value(args)->val), nargs);
		if(res)
		{
			ds_append(pstack->state->buffer, res);
			ds_unfix(res);
		}
		key = s->class->next(i, s, &data);
	}
}

/// mark-remove(storage key)
// marks the item \em[key] of the \em[storage] as removed.
// This function was introduced since it is generally not safe to delete
// items while inside an enumeration loop. Note that items marked for
// removal do not hide from following operations
// \note Removal marking is done by maintaining a linear list of items attached to
// a storage, so marking a large amount of items may slow down your system
// and even exceed the system's memory.
// \note if you need to zap the whole storage, the better way in many cases
// would be releasing the storage altogether and recreating it (This won't
// work if your storage is shared among several environments; if it has a
// generated name you need to recreate it with the same name not with a
// newly generated one).
static void bi_mark_remove(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	to_remove_list *trl;
	trl = ALLOC(to_remove_list);
	trl->key = ds_fix(next_value(args)->val);
	s->to_remove = list_add(s->to_remove, trl);
}

/// unmark-remove(storage key)
// if the item \em[key] of the \em[storage] has been previously marked as
// removed, clear that mark. Otherwise, nothing happens.
static void bi_unmark_remove(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	s->to_remove = list_remove(s->to_remove, DS_BODY(next_value(args)->val), (list_predicate_t)to_remove_find, (list_destroy_t)destroy_to_remove, sizeof(to_remove_list));
}

/// ismark-remove(storage key)
// \returns true of the item \em[key] of the \em[storage] is marked for
// removal
static void bi_ismark_remove(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	if(list_find(s->to_remove, DS_BODY(next_value(args)->val), (list_predicate_t)to_remove_find))
		ds_appendstr(pstack->state->buffer, "true");
}

/// cancel-remove(storage)
// clears all removal marks
static void bi_cancel_remove(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	list_delete(s->to_remove, (list_destroy_t)destroy_to_remove, sizeof(to_remove_list));
	s->to_remove = NULL;
}

/// sto.enum(storage handle-var)
// \returns if the value of \em[handle-var] is empty, the first key in the
// \em[storage] and puts a handler to an enumeration into \em[handle-var].
// Otherwise, the next key from the enumeration and adjusts it resp. When
// there are no more items, \em[handle-var] is cleared.
// \note value held in \em[handle-var] is opaque; you shouldn't try use it
// but with \em[sto.enum] and \em[sto.break] functions
static void bi_sto_enum(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(next_value(args)->val), 1);
	storage *s = find_storage(i, DS_BODY(args->val));
	void *data;
	dstring *key;
	if(ds_isempty(v->vals->val))
	{
		if(!third_value(args))
			key = s->class->first(i, s, &data);
		else
		{
			if(!s->class->seek)
				recover_error("seek not implemented");
			key = s->class->seek(i, s, &data, third_value(args)->val);
		}
	}
	else
	{
		char *ref = DS_BODY(v->vals->val);
		if(*ref != '#' || ref[1] != '#' || ref[2] != '?')
			recover_error("invalid enumeration handle");
		data = (void *)strtoul(ref + 3, NULL, 16);
		key = s->class->next(i, s, &data);
	}
	if(!key)
		ds_assign(v->vals->val, ds_create(NULL));
	else
	{
		ds_append(pstack->state->buffer, key);
		ds_assign(v->vals->val, ds_create("##?"));
		ds_append(v->vals->val, ds_fromint((int)data, 16, 0));
	}
}

/// sto.break(storage handle-var)
// cleans up the enumeration of \em[storage] whose handler is in
// \em[handle-var]. The latter is cleared.
static void bi_sto_break(chunk *current, interpreter *i, value *args)
{
	variable *v = safe_get_var(i, DS_BODY(next_value(args)->val), 1);
	storage *s = find_storage(i, DS_BODY(args->val));
	void *data;
	char *ref = DS_BODY(v->vals->val);
	if(*ref != '#' || ref[1] != '#' || ref[2] != '?')
		recover_error("invalid enumeration handle");
	data = (void *)strtoul(ref + 3, NULL, 16);
	s->class->enditer(i, s, &data);
	ds_assign(v->vals->val, ds_create(NULL));
}

static interpreter *trl_interpreter;
static storage *trl_storage;

static void do_remove(to_remove_list *trl)
{
	trl_storage->class->remove(trl_interpreter, trl_storage, trl->key);
	destroy_to_remove(trl);
}

/// 
// do-remove(storage)
// physically removes all the marked items from the \em[storage]
static void bi_do_remove(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	trl_interpreter = i;
	trl_storage = s;
	list_delete(s->to_remove, (list_destroy_t)do_remove, sizeof(to_remove_list));
	s->to_remove = NULL;
}

/// 
// count(storage)
//  \returns the number of items in \em[storage].
//  \note It is not guaranteed that this is a total number of accessible
//  items nor that this is the number of iterations of a \em[foreach] loop,
//  though in most cases both statements holds true
static void bi_count(chunk *current, interpreter *i, value *args)
{
	storage *s = find_storage(i, DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(s->class->numofitems(i, s), 10, 0));
}

/// defautoseq(as-name autoseq)
// defines a sequence of automata \em[autoseq] under the name \em[as-name].
// For the syntax of \em[autoseq] see \ref[sla(1)]
static void bi_defautoseq(chunk *current, interpreter *i, value *args)
{
	char *name = DS_BODY(args->val);
	autoseqdef *asd;
	autoseq *as;
	if(!*name || (*name == '*' && !name[1]))
		name = unique_name();
	if(list_find(i->aseqs, DS_BODY(args->val), (list_predicate_t)si_find_exact))
		recover_error("automata sequence '%s' is already defined", DS_BODY(args->val));
	as = parse_autoseq(DS_BODY(next_value(args)->val));
	asd = ALLOC(autoseqdef);
	asd->seq = as;
	autoseq_init(asd->seq);
	add_shared_int(i, init_shared_autoseq(i, asd, name));
	ds_appendstr(pstack->state->buffer, name);
}

/// rmautoseq(as-name)
// equivalent to \em[release autoseq as-name]
static void bi_rmautoseq(chunk *current, interpreter *i, value *args)
{
	i->aseqs = list_remove(i->aseqs, DS_BODY(args->val), (list_predicate_t)si_find_exact,
							(list_destroy_t)destroy_shared_int, sizeof(shared_int));
}

/// au.load(au-name)
// loads an automaton \em[au-name]
static void bi_au_load(chunk *current, interpreter *i, value *args)
{
	load_automaton(DS_BODY(args->val));
}

/// au.new(au-name)
// creates an empty automaton named \em[au-name]
// \error[automaton 'au-name' already defined]
static void bi_au_new(chunk *current, interpreter *i, value *args)
{
	create_automaton(DS_BODY(args->val));
}

/// au.build(au-name au-definition)
// parses \em[au-definition] as described in \ref[sla(1)] and
// adds it to an automaton \em2[au-name][.] Several \em2[au.build][s] may be
// applied to the same automaton.
// \note it is recommended to use this function only with empty or
// explicitly loaded automata with no automata sequences using them.
static void bi_au_build(chunk *current, interpreter *i, value *args)
{
	int old_lno = set_line_no(1);
	automaton *au = get_au_byname(DS_BODY(args->val));
	if(!au)
		recover_error("automaton '%s' undefined", DS_BODY(args->val));
	parse_automaton(au, DS_BODY(next_value(args)->val), NULL);
	set_line_no(old_lno);
}

static int translate_input(dstring *ns, autoseq *as, int ch)
{
	int result;
	int och = -1;
	for(;;)
	{
		result = autoseq_step(as, &ch, &och);
		if(och > 0 && ns)
		{
			if(och > UCHAR_MAX)
				recover_error("translated character U+%8.8X doesn't fit into 8 bit", och);
			ds_appendch(ns, och);
		}
		if(result == ar_rejected || (ch < 0 && need_feeding(as)))
			break;
	}
	return result;
}

static int itranslate_input(dstring *ns, autoseq *as, int ch)
{
	int result;
	int och = -1;
	char buf[16];
	for(;;)
	{
		result = autoseq_step(as, &ch, &och);
		if(och >= 0 && ns)
		{
			sprintf(buf, "%#x ", och);
			ds_appendstr(ns, buf);
		}
		if(result == ar_rejected || (ch < 0 && need_feeding(as)))
			break;
	}
	return result;
}

/// itranslatez(as-name string)
// passes the \em[string] thru the automata sequence \em[as-name]. The
// leading zero is passed too.
// \returns the output of the last automaton in the sequence as a list
// of hexadecimal numbers
static void bi_itranslatez(chunk *current, interpreter *i, value  *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));
	autoseq *as;
	char *str;
	dstring *ns;
	as = asd->seq;
	ns = ds_create(NULL);
	for(str = DS_BODY(next_value(args)->val); *str; str++)
	{
		if(itranslate_input(ns, as, *str) == ar_rejected)
			return;
	}
	if(itranslate_input(ns, as, 0) == ar_accepted)
		ds_append(pstack->state->buffer, ns);
}

/// itranslate(as-name string)
// same as \em[itranslate] but the leading zero is not passed.
static void bi_itranslate(chunk *current, interpreter *i, value  *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));
	autoseq *as;
	char *str;
	dstring *ns;
	as = asd->seq;
	ns = ds_create(NULL);
	for(str = DS_BODY(next_value(args)->val); *str; str++)
	{
		if(itranslate_input(ns, as, *str) == ar_rejected)
			return;
	}
	ds_append(pstack->state->buffer, ns);
}

/// translatez(as-name string)
// same as \em[itranslatez], but the result is an ASCII string.
// \note If an output value is outside the ASCII range, an error occurs.
static void bi_translatez(chunk *current, interpreter *i, value  *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));
	autoseq *as;
	char *str;
	dstring *ns;
	as = asd->seq;
	ns = ds_create(NULL);
	for(str = DS_BODY(next_value(args)->val); *str; str++)
	{
		if(translate_input(ns, as, *str) == ar_rejected)
			return;
	}
	if(translate_input(ns, as, 0) == ar_accepted)
		ds_append(pstack->state->buffer, ns);
}

/// translate(as-name string)
// same as \em[translate], but the leading zero is not passed
static void bi_translate(chunk *current, interpreter *i, value  *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));
	autoseq *as;
	char *str;
	dstring *ns;
	as = asd->seq;
	ns = ds_create(NULL);
	for(str = DS_BODY(next_value(args)->val); *str; str++)
	{
		if(translate_input(ns, as, *str) == ar_rejected)
			return;
	}
	ds_append(pstack->state->buffer, ns);
}

/// ctranslate(as-name int-val)
// passes a character whose code is \em[int-val] (possibly, outside the
// ASCII range) thru the automata sequence \em[as-name]
// \returns an empty string if the input is rejected; otherwise the decimal
// code of the output character prepended by '+' if there are no more output
// characters for this input and '-' otherwise. If the output is empty a
// single + or - is returned.
static void bi_ctranslate(chunk *current, interpreter *i, value  *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));
	int in = convert_to_int(DS_BODY(next_value(args)->val));
	int out, result;
	result = autoseq_step(asd->seq, &in, &out);
	if(result != ar_rejected)
	{
		ds_appendch(pstack->state->buffer, (in < 0 && need_feeding(asd->seq)) ? '+' : '-');
		if(out >= 0)
		{
			dstring *n = ds_fromint(out, 10, 0);
			ds_append(pstack->state->buffer, n);
		}
	}
}

/// reset(as-name)
// resets the automata sequence \em[as-name] to its initial state
// \returns \em[as-name] so that it is possible to use this function where
// an automata sequence name is required
static void bi_reset(chunk *current, interpreter *i, value  *args)
{
	shared_int *si = list_find(i->aseqs, DS_BODY(args->val), (list_predicate_t)si_find_exact);
	autoseqdef *asd;
	
	if(!si)
		recover_error("autoseq '%s' undefined", si->intname);
	asd = si->obj;
	autoseq_init(asd->seq);
	ds_appendstr(pstack->state->buffer, si->intname);
}

/// state(as-name)
// \returns the state of the last automata from the automata sequence
// \em[as-name]
static void bi_state(chunk *current, interpreter *i, value *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));
	
	ds_append(pstack->state->buffer, ds_fromint(as_get_reg(asd->seq, nv_state), 10, 0));
} 

/// topstack(as-name)
// \returns the topmost item of the internal stack of the last automata from
// the sequence \em[as-name] or null if the stack is empty
static void bi_topstack(chunk *current, interpreter *i, value *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));

	if(as_get_reg(asd->seq, nv_depth))
		ds_append(pstack->state->buffer, ds_fromint(as_get_reg(asd->seq, nv_topstack), 10, 0));
} 

/// mapstate(as-name state-name)
// \returns the value of the macro \em[state-name] in the last automaton
// from the sequence \em[as-name] or null if no such macro. See \ref[sla(1)]
// for more details.
static void bi_mapstate(chunk *current, interpreter *i, value *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));
	int state;
	
	state = as_map_name(asd->seq, DS_BODY(next_value(args)->val));
	if(state >= 0)
		ds_append(pstack->state->buffer, ds_fromint(state, 10, 0));
}

/// statename(as-name prefix value)
// \returns a name of a macro in the last automaton from the sequence
// \em[as-name] whose value is \em[value] and name starts with \em[prefix]
static void bi_statename(chunk *current, interpreter *i, value *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));
	int state;
	char *name;
	
	if(asd->seq)
	{
		state = third_value(args) ? convert_to_int(DS_BODY(third_value(args)->val)) : as_get_reg(asd->seq, nv_state);
		name = as_getname_byid(asd->seq, state, DS_BODY(next_value(args)->val));
		if(name)
			ds_appendstr(pstack->state->buffer, name);
	}
}

/// isfinal(as-name)
// \returns true if the last automaton from the sequence \em[as-name] is in
// one of its final states.
static void bi_isfinal(chunk *current, interpreter *i, value *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(args->val));

	if(as_is_in_final(asd->seq))
		ds_appendstr(pstack->state->buffer, "true");
}

#define INT_LENGTH 0

int convert_to_int(char *str)
{
	char *check = "";
	int nv = *str ? strtol(str + (*str == '#'), &check, *str == '#' ? 16 : 0) : 0;
	if(*check)
		recover_error("can't convert value '%s'", str);
	return nv;
}

/// isint(arg)
// \returns true if \em[arg] is a valid representation of an integer
static void bi_isint(chunk *current, interpreter *i, value *args)
{
	char *str = DS_BODY(args->val);
	char *check = "";
	*str ? strtol(str + (*str == '#'), &check, *str == '#' ? 16 : 0) : 0;
	if(!*check)
		ds_appendstr(pstack->state->buffer, "true");
}

/// add(val1 ...)
// \returns the sum of \em2[val][s.]
static void bi_add(chunk *current, interpreter *i, value *args)
{
	int val = 0;

	for(; args; args = list_next(args))
		val += convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(val, 10, INT_LENGTH));
}

/// postincr(var val)
// increments \em[var] to \em[val]
// \returns the old value of \em[var]
static void bi_postincr(chunk *current, interpreter *i, value *args)
{
	int val = 0;
   	int ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ds_append(pstack->state->buffer, var->vals->val);
	val = convert_to_int(DS_BODY(next_value(args)->val));
	ovalue = convert_to_int(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromint(val + ovalue, 10, 0));
}

/// incr(var val)
// increments \em[var] to \em[val]
// \returns the new value of \em[var]
static void bi_incr(chunk *current, interpreter *i, value *args)
{
	int val = 0; 
	int ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	val = convert_to_int(DS_BODY(next_value(args)->val));
	ovalue = convert_to_int(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromint(val + ovalue, 10, 0));
	ds_append(pstack->state->buffer, var->vals->val);
}

/// postdecr(var val)
// decrements \em[var] to \em[val]
// \returns the old value of \em[var]
static void bi_postdecr(chunk *current, interpreter *i, value *args)
{
	int val = 0;
   	int ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ds_append(pstack->state->buffer, var->vals->val);
	val = convert_to_int(DS_BODY(next_value(args)->val));
	ovalue = convert_to_int(DS_BODY(var->vals->val));
	ds_assign(var->vals->val, ds_fromint(ovalue - val, 10, 0));
}

/// decr(var val)
// decrements \em[var] to \em[val]
// \returns the new value of \em[var]
static void bi_decr(chunk *current, interpreter *i, value *args)
{
	int val = 0;
	int ovalue;
	variable *var = safe_get_var(i, DS_BODY(args->val), 1);

	ovalue = convert_to_int(DS_BODY(var->vals->val));
	val = convert_to_int(DS_BODY(next_value(args)->val));
	ds_assign(var->vals->val, ds_fromint(ovalue - val, 10, 0));
	ds_append(pstack->state->buffer, var->vals->val);
}

/// mul(val1 ...)
// \returns the production of \em2[val][s]
static void bi_mul(chunk *current, interpreter *i, value *args)
{
	int val = 1;

	for(; args; args = list_next(args))
		val *= convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(val, 10, INT_LENGTH));
}

/// sub(val1 val2)
// \returns the result of subtracting \em[val2] from \em[val1]
static void bi_sub(chunk *current, interpreter *i, value *args)
{
	int v1 = convert_to_int(DS_BODY(args->val));
	int v2 = convert_to_int(DS_BODY(((value	*)list_next(args))->val));
	ds_append(pstack->state->buffer, ds_fromint(v1 - v2, 10, INT_LENGTH));
}

/// div(val1 val2)
// \returns the integral quotient of dividing \em[val1] to \em[val2]
// \error[zero division]
static void bi_div(chunk *current, interpreter *i, value *args)
{
	int v1 = convert_to_int(DS_BODY(args->val));
	int v2 = convert_to_int(DS_BODY(((value	*)list_next(args))->val));
	if(!v2)
		recover_error("zero division");
	ds_append(pstack->state->buffer, ds_fromint(v1 / v2, 10, INT_LENGTH));
}

/// rem(val1 val2)
// \returns the remainder of dividing \em[val1] to \em[val2]
// \note if \em[val1] is negative, the result is system-dependent. Consult
// your C compiler reference or experiment.
// \error[zero division]
static void bi_rem(chunk *current, interpreter *i, value *args)
{
	int v1 = convert_to_int(DS_BODY(args->val));
	int v2 = convert_to_int(DS_BODY(((value	*)list_next(args))->val));
	if(!v2)
		recover_error("zero division");
	ds_append(pstack->state->buffer, ds_fromint(v1 % v2, 10, INT_LENGTH));
}

/// neg(val)
// \returns the negation of \em[val]
static void bi_neg(chunk *current, interpreter *i, value *args)
{
	int v = convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(-v, 10, INT_LENGTH));
}

/// abs(val)
// \returns the absolute value of \em[val]
static void bi_abs(chunk *current, interpreter *i, value *args)
{
	int v = convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(abs(v), 10, INT_LENGTH));
}

/// bitand(val1 ...)
// \returns a bitwise AND of all \em[val][s] regarded as integers
static void bi_bitand(chunk *current, interpreter *i, value *args)
{
	int val;

	val = convert_to_int(DS_BODY(args->val));
	for(args = list_next(args); args; args = list_next(args))
		val &= convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(val, 10, INT_LENGTH));
}

/// bitor(val1 ...)
// \returns a bitwise OR of all \em[val][s] regarded as integers
static void bi_bitor(chunk *current, interpreter *i, value *args)
{
	int val = 0;

	for(; args; args = list_next(args))
		val |= convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(val, 10, INT_LENGTH));
}

/// bitxor(val1 ...)
// \returns a bitwise XOR of all \em[val][s] regarded as integers
static void bi_bitxor(chunk *current, interpreter *i, value *args)
{
	int val = 0;

	for(; args; args = list_next(args))
		val ^= convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(val, 10, INT_LENGTH));
}

/// bitinv(val)
// \returns a bitwise complement of \em[val]
static void bi_bitinv(chunk *current, interpreter *i, value *args)
{
	int val = ~convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(val, 10, INT_LENGTH));
}

/// bitshift(val1 shift)
// \returns shifts \em[val] to \em[shift] bits left if the \em[shift] is
// positive or right if negative
static void bi_bitshift(chunk *current, interpreter *i, value *args)
{
	int val = convert_to_int(DS_BODY(args->val));
	int shift = convert_to_int(DS_BODY(next_value(args)->val));

	if(shift < 0)
		val >>= -shift;
	else 
		val <<= shift;
	ds_append(pstack->state->buffer, ds_fromint(val, 10, INT_LENGTH));
}

/// int(val)
// \returns An internal representation of integers, suitable for 
// numeric comparisons.
// \note This function is deprecated, use \em[int] modifier of comparison
// functions instead.
static void bi_int(chunk *current, interpreter *i, value *args)
{
	int v = convert_to_int(DS_BODY(args->val));
	int shift = CHAR_BIT * sizeof(int) - CHAR_BIT;
	static char hd[] = "0123456789ABCDEF";
	ds_appendch(pstack->state->buffer, '#');
	if(v >= 0)
		ds_appendch(pstack->state->buffer, '0');
	else
	{
		ds_appendch(pstack->state->buffer, '-');
		v = -v;
	}
	for(; shift >=0; shift -= 4) 
		ds_appendch(pstack->state->buffer, hd[(v >> shift) & 0x0F]);
}

/// dec(val)
// \returns a decimal representation of \em[val]
static void bi_dec(chunk *current, interpreter *i, value *args)
{
	int v = convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(v, 10, INT_LENGTH));
}

/// hex(val)
// \returns a hexadecimal representation of \em[val]
static void bi_hex(chunk *current, interpreter *i, value *args)
{
	int v = convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(v, 16, INT_LENGTH));
}

/// oct(val)
// \returns an octal representation of \em[val]
static void bi_oct(chunk *current, interpreter *i, value *args)
{
	int v = convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(v, 8, INT_LENGTH));
}

/// uint8(val)
// converts an integer as it's been converted from unsigned char to signed
static void bi_uint8(chunk *current, interpreter *i, value *args)
{
	unsigned char v = (signed char)convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(v, 10, INT_LENGTH));
}

/// uint16(val)
// converts an integer as it's been converted from unsigned short to signed
static void bi_uint16(chunk *current, interpreter *i, value *args)
{
	unsigned short v = (signed short)convert_to_int(DS_BODY(args->val));
	ds_append(pstack->state->buffer, ds_fromint(v, 10, INT_LENGTH));
}

typedef int (*comparison_t)(dstring *val1, dstring *val2);

static int comparison_dispatch(value *args, comparison_t strf, comparison_t numf, 
								comparison_t verf)
{
	if(third_value(args))
	{
		if(!strcmp(DS_BODY(args->val), "int"))
			return numf(next_value(args)->val, third_value(args)->val);
		else if(!strcmp(DS_BODY(args->val), "ver"))
			return verf(next_value(args)->val, third_value(args)->val);
		else if(!strcmp(DS_BODY(args->val), "string"))
			return strf(args->val, next_value(args)->val);
		else recover_error("unknown type specifier: %s", DS_BODY(args->val));
	}
	return strf(args->val, next_value(args)->val);
}

static int c_eq_s(dstring *val, dstring *val2)
{
	return !ds_compare(val, val2, NULL, NULL);
}

static int c_eq_n(dstring *val, dstring *val2)
{
	return convert_to_int(DS_BODY(val)) == convert_to_int(DS_BODY(val2));
}

static int c_eq_v(dstring *val, dstring *val2)
{
	return !versioncmp(DS_BODY(val), DS_BODY(val2));
}

static int c_less_s(dstring *val, dstring *val2)
{
	return ds_compare(val, val2, NULL, NULL) < 0;
}

static int c_less_n(dstring *val, dstring *val2)
{
	return convert_to_int(DS_BODY(val)) < convert_to_int(DS_BODY(val2));
}

static int c_less_v(dstring *val, dstring *val2)
{
	return versioncmp(DS_BODY(val), DS_BODY(val2)) < 0;
}

static int c_greater_s(dstring *val, dstring *val2)
{
	return ds_compare(val, val2, NULL, NULL) > 0;
}

static int c_greater_n(dstring *val, dstring *val2)
{
	return convert_to_int(DS_BODY(val)) > convert_to_int(DS_BODY(val2));
}

static int c_greater_v(dstring *val, dstring *val2)
{
	return versioncmp(DS_BODY(val), DS_BODY(val2)) > 0;
}

/// Comparison functions
// All the comparison functions may take two or three arguments (three or
// fours resp. in case of \em[between]. In case of three arguments, the
// first one indicates the type of comparison. The core defines two types:
// \em[int] for comparing integers and \em[string] for comparing strings.
// Pluggable modules may define other types. If there are two arguments, the
// type \em[string] is assumed.
// \error[unknown type specifier: type]

/// eq([type] arg1 arg2)
// \returns true if \em[arg1] is equal to \em[arg2]
static void bi_eq(chunk *current, interpreter *i, value *args)
{
	if(comparison_dispatch(args, c_eq_s, c_eq_n, c_eq_v))
		ds_appendstr(pstack->state->buffer, "true");
}

/// neq([type] arg1 arg2)
// \returns true if \em[arg1] is not equal to \em[arg2]
static void bi_neq(chunk *current, interpreter *i, value *args)
{
	if(!comparison_dispatch(args, c_eq_s, c_eq_n, c_eq_v))
		ds_appendstr(pstack->state->buffer, "true");
}

/// eqi(arg1 arg2)
// \returns true if \em[arg1] is equal to \em[arg2] ignoring the case
// \note this function cannot be used with a type specifier
static void bi_eqi(chunk *current, interpreter *i, value *args)
{
	int c = ds_compare(args->val, next_value(args)->val, ds_p_casefold, NULL);
	if(!c)
		ds_appendstr(pstack->state->buffer, "true");
}

/// less([type] arg1 arg2)
// \returns true if \em[arg1] is less than \em[arg2]
static void bi_less(chunk *current, interpreter *i, value *args)
{
	if(comparison_dispatch(args, c_less_s, c_less_n, c_less_v))
		ds_appendstr(pstack->state->buffer, "true");
}

/// greater([type] arg1 arg2)
// \returns true if \em[arg1] is greater than \em[arg2]
static void bi_greater(chunk *current, interpreter *i, value *args)
{
	if(comparison_dispatch(args, c_greater_s, c_greater_n, c_greater_v))
		ds_appendstr(pstack->state->buffer, "true");
}

/// le([type] arg1 arg2)
// \returns true if \em[arg1] is less than or equal to \em[arg2]
static void bi_le(chunk *current, interpreter *i, value *args)
{
	if(!comparison_dispatch(args, c_greater_s, c_greater_n, c_greater_v))
		ds_appendstr(pstack->state->buffer, "true");
}

/// ge([type] arg1 arg2)
// \returns true if \em[arg1] is greater than or equal to \em[arg2]
static void bi_ge(chunk *current, interpreter *i, value *args)
{
	if(!comparison_dispatch(args, c_less_s, c_less_n, c_less_v))
		ds_appendstr(pstack->state->buffer, "true");
}

/// expr(arg)
// \returns the result of evaluating \em[arg] as arithmetic expression
// \note the syntax is the same as for expressions in automata definitions
// (see \ref[sla(1)]). This means, in particular, that no variable expansion
// is done
static void bi_expr(chunk *current, interpreter *i, value *args)
{
	int val = eval_expr(DS_BODY(args->val), NULL, NULL, NULL, NULL);
	ds_append(pstack->state->buffer, ds_fromint(val, 10, 0));
}

/// match(pattern arg)
// \returns true if \em[arg] matches a \em2[pattern][.] Patterns have the
// same syntax as shell globbing patterns.
static void bi_match(chunk *current, interpreter *i, value *args)
{
	int c = fnmatch(DS_BODY(args->val), DS_BODY(next_value(args)->val), 0);
	if(!c)
		ds_appendstr(pstack->state->buffer, "true");
}

/// matchi(pattern1 arg)
// \returns true if \em[arg] matches a \em[pattern] ignoring the case.
static void bi_matchi(chunk *current, interpreter *i, value *args)
{
	int c = fnmatch(DS_BODY(args->val), DS_BODY(next_value(args)->val), FNM_CASEFOLD);
	if(!c)
		ds_appendstr(pstack->state->buffer, "true");
}

/// between([type] arg1 arg2 arg3)
// \returns true if \em[arg2] lies between \em[arg2] and \em2[arg3][.]
static void bi_between(chunk *current, interpreter *i, value *args)
{
	if(!next_value(third_value(args)) || !strcmp(DS_BODY(args->val), "string"))
	{
		if(next_value(third_value(args)))
			args = list_next(args);
		// block
		{
			int c1 = ds_compare(args->val, next_value(args)->val, NULL, NULL);
			int c2 = ds_compare(next_value(args)->val, third_value(args)->val, NULL, NULL);
			if(c1 <= 0 && c2 <= 0)
				ds_appendstr(pstack->state->buffer, "true");
		}
	}
	else if(!strcmp(DS_BODY(args->val), "int"))
	{
		int i1 = convert_to_int(DS_BODY(next_value(args)->val));
		int i2 = convert_to_int(DS_BODY(third_value(args)->val));
		int i3 = convert_to_int(DS_BODY(next_value(third_value(args))->val));
		if(i1 <= i2 && i2 <= i3)
			ds_appendstr(pstack->state->buffer, "true");
	}
	else if(!strcmp(DS_BODY(args->val), "ver"))
	{
		int v1 = versioncmp(DS_BODY(next_value(args)->val), DS_BODY(third_value(args)->val));
		int v2 = versioncmp(DS_BODY(third_value(args)->val), DS_BODY(next_value(third_value(args))->val)); 
		if(v1 <= 0 && v2 <= 0)
			ds_appendstr(pstack->state->buffer, "true");
	}
}

/// and(arg1 ...)
// \returns true if all of \em2[arg][s] are not null
// \note this function is not short-circuit

static void bi_and(chunk *current, interpreter *i, value *args)
{
	if(args)
		ds_appendstr(pstack->state->buffer, "true");
}
// static void bi_and(chunk *current, interpreter *i, value *args)
// {
// 	for(; args; args = list_next(args))
// 	{
// 		if(ds_isempty(args->val))
// 			return;
// 	}
// 	ds_appendstr(pstack->state->buffer, "true");
// }

/// or(arg1 ...)
// \returns true if any of \em2[arg][s] is not null
// \note this function is not short-circuit

static void bi_or(chunk *current, interpreter *i, value *args)
{
	if(!args)
		ds_appendstr(pstack->state->buffer, "true");
}
// static void bi_or(chunk *current, interpreter *i, value *args)
// {
// 	for(; args; args = list_next(args))
// 	{
// 		if(!ds_isempty(args->val))
// 			ds_appendstr(pstack->state->buffer, "true");
// 	}
// }

/// not(arg)
// \returns true if \em[arg] is null
static void bi_not(chunk *current, interpreter *i, value *args)
{
	if(ds_isempty(args->val))
		ds_appendstr(pstack->state->buffer, "true");
}

/// cond(condition if-arg [else-arg])
// \returns \em[if-arg] if \em[condition] is not null, otherwise
// \em[else-arg] or null if the latter is not present
// \note this function is not short-circuit. Alas!

static void bi_cond(chunk *current, interpreter *i, value *args)
{
	if(next_value(args))
		ds_append(pstack->state->buffer, next_value(args)->val);
}

// static void bi_cond(chunk *current, interpreter *i, value *args)
// {
// 	if(!ds_isempty(args->val))
// 		ds_append(pstack->state->buffer, next_value(args)->val);
// 	else if(third_value(args))
// 		ds_append(pstack->state->buffer, third_value(args)->val);
// }

/// length(arg)
// \returns the length of the string \em[arg]
static void bi_length(chunk *current, interpreter *i, value *args)
{
	int v = ds_length(args->val);
	ds_append(pstack->state->buffer, ds_fromint(v, 10, INT_LENGTH));
}

/// ascii(arg)
// \returns the ASCII code of the first character of \em[arg]
// (zero if \em[arg] is null)
static void bi_ascii(chunk *current, interpreter *i, value *args)
{
	ds_append(pstack->state->buffer, ds_fromint(*DS_BODY(args->val), 10, INT_LENGTH));
}

/// char(arg)
// \returns a character whose ASCII code is equal to \em[arg]. Null is
// returned if \em[arg] is zero. If \em[arg] doesn't fit the ASCII range,
// the result is system-dependent.
static void bi_char(chunk *current, interpreter *i, value *args)
{
	int v = convert_to_int(DS_BODY(args->val));
	ds_appendch(pstack->state->buffer, v);
}

/// trim(arg)
// \returns \em[arg] with leading and trailing spaces stripped away
static void bi_trim(chunk *current, interpreter *i, value *args)
{
	dstring *out = pstack->state->buffer;
	dstring *temp = ds_create(NULL);
	dstring *null = ds_create(NULL);
	char *ptr;
	for(ptr = DS_BODY(args->val); isspace(*ptr); ptr++)
		;
	for(; *ptr; ptr++)
	{
		if(isspace(*ptr))
			ds_appendch(temp, *ptr);
		else
		{
			if(!ds_isempty(temp))
			{
				ds_append(out, temp);
				ds_assign(temp, null);
			}
			ds_appendch(out, *ptr);
		}
	}
}

/// substr(arg init-pos [len])
// \returns a substring of \em[arg] staring from \em[init-pos] with the
// length \em[len]. If \em[len] is not present, the substring is from
// \em[init-pos] till the end of \em[arg].
// \note the first character is number 0
static void bi_substr(chunk *current, interpreter *i, value *args)
{
	int start = convert_to_int(DS_BODY(next_value(args)->val));
	int len = third_value(args) ? convert_to_int(DS_BODY(third_value(args)->val)) :
									ds_length(args->val) - start;
	ds_append(pstack->state->buffer, ds_substr(args->val, start, len));
}

/// .substr(init-pos [len])
// same as \em[substr], but it operates on the contents of the current
// execution buffer
static void bi_psubstr(chunk *current, interpreter *i, value *args)
{
	int start = convert_to_int(DS_BODY(args->val));
	int len = next_value(args) ? convert_to_int(DS_BODY(next_value(args)->val)) : 
								 ds_length(pstack->state->buffer) - start;
	ds_assign(pstack->state->buffer, ds_substr(pstack->state->buffer, start, len));
}

/// index(arg substr [startpos])
// \returns a position at which \em[substr] occurs in \em[arg]. If
// \em[substr] is not contained in \em[arg], null is returned. If
// \em[startpos] is present, it specifies the initial position for seeking.
static void bi_index(chunk *current, interpreter *i, value *args)
{
	int startpos = third_value(args) ? convert_to_int(DS_BODY(third_value(args)->val)) : 0;
	int pos = ds_find(args->val, startpos, next_value(args)->val, NULL, NULL);

	if(pos >= 0)
		ds_append(pstack->state->buffer, ds_fromint(pos, 10, 0));
}

/// rindex(arg substr)
// \returns a position of \em[last] occurence of \em[substr] within \em[arg]
static void bi_rindex(chunk *current, interpreter *i, value *args)
{
	int pos = ds_rfind(args->val, next_value(args)->val, NULL, NULL);

	if(pos >= 0)
		ds_append(pstack->state->buffer, ds_fromint(pos, 10, 0));
}

/// isprefix(arg prefix)
// \returns true if \em[arg] starts with a \em[prefix]
static void bi_isprefix(chunk *current, interpreter *i, value *args)
{
	if(ds_isprefix(args->val, next_value(args)->val))
		ds_appendstr(pstack->state->buffer, "true");
}

/// issuffix(arg suffix)
// \returns true if \em[arg] ends in a \em[suffix]
static void bi_issuffix(chunk *current, interpreter *i, value *args)
{
	if(ds_issuffix(args->val, next_value(args)->val))
		ds_appendstr(pstack->state->buffer, "true");
}

/// commonprefix(arg1 arg2)
// \returns the most common prefix of \em[arg1] and \em[arg2]
static void bi_commonprefix(chunk *current, interpreter *i, value *args)
{
	ds_append(pstack->state->buffer, ds_commonprefix(args->val, next_value(args)->val));
}
	
static int au_match_input(autoseq *seq, int *val)
{
	int och = -1;
	int res;
	int state;

	for(;;)
	{
		state = as_get_reg(seq, nv_state);
		res = autoseq_step(seq, val, &och);
		if(res == ar_rejected || (*val < 0 && need_feeding(seq)) || (as_get_reg(seq, nv_state) != state))
			break;
	}
	return res;
}

/// auindex(arg aseq [len-var])
// \returns a position of the first substring in \em[arg] which is accepted
// by the automata sequence \em[aseq]. If \em[len-var] is present, it is
// stored the length of the substring.
static void bi_auindex(chunk *current, interpreter *i, value *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(next_value(args)->val));
	char *base, *pos;
	variable *v = NULL;
	int res;
	int inch = -1;
	int delta = 1;
	
	if(third_value(args))
		v = safe_get_var(i, DS_BODY(third_value(args)->val), 1);
	for(base = DS_BODY(args->val); *base; base++)
	{
		autoseq_init(asd->seq);
		delta = 1;
		for(pos = base; *pos; pos += delta)
		{
			inch = *pos;
			if((res = au_match_input(asd->seq, &inch)) == ar_rejected)
				break;
			delta = inch < 0 && need_feeding(asd->seq);
			if(as_is_in_final(asd->seq))
			{
				ds_append(pstack->state->buffer, ds_fromint(base - DS_BODY(args->val), 10, 0));
				if(v)
					ds_assign(v->vals->val, ds_fromint(pos - base + delta, 10, 0));
				return;
			}
		}
	}
	if(v)
		ds_assign(v->vals->val, ds_create(NULL));
}

/// auindexmax(arg aseq [len-var])
// like \em[auindex] but seeks for the maximal length substring
static void bi_auindexmax(chunk *current, interpreter *i, value *args)
{
	autoseqdef *asd = find_asd(i, DS_BODY(next_value(args)->val));
	char *base, *pos;
	variable *v = NULL;
	int res;
	int delta;
	int inch;
	
	if(third_value(args))
		v = safe_get_var(i, DS_BODY(third_value(args)->val), 1);
	for(base = DS_BODY(args->val); *base; base++)
	{
		autoseq_init(asd->seq);
		delta = 1;
		for(pos = base; *pos; pos += delta)
		{
			inch = *pos;
			if((res = au_match_input(asd->seq, &inch)) == ar_rejected)
				break;
			delta = inch < 0 && need_feeding(asd->seq);
		}
		if(as_is_in_final(asd->seq))
		{
			ds_append(pstack->state->buffer, ds_fromint(base - DS_BODY(args->val), 10, 0));
			if(v)
				ds_assign(v->vals->val, ds_fromint(pos - base + delta, 10, 0));
			return;
		}
	}
	if(v)
		ds_assign(v->vals->val, ds_create(NULL));
}

/// indexi(arg substr [start-pos])
// like \em[index] but ignoring case
static void bi_indexi(chunk *current, interpreter *i, value *args)
{
	int startpos = third_value(args) ? convert_to_int(DS_BODY(third_value(args)->val)) : 0;
	int pos = ds_find(args->val, startpos, next_value(args)->val, ds_p_casefold, NULL);

	if(pos >= 0)
		ds_append(pstack->state->buffer, ds_fromint(pos, 10, 0));
}

/// rindexi(arg substr [start-pos])
// like \em[rindex] but ignoring case
static void bi_rindexi(chunk *current, interpreter *i, value *args)
{
	int pos = ds_rfind(args->val, next_value(args)->val, ds_p_casefold, NULL);

	if(pos >= 0)
		ds_append(pstack->state->buffer, ds_fromint(pos, 10, 0));
}

/// replace(arg start len [subst])
// replaces the substring of \em[arg] which starts at \em[start] having the length
// of \em[len] with \em[subst] or null if the latter is omitted 
static void bi_replace(chunk *current, interpreter *i, value *args)
{
	dstring *rep = next_value(third_value(args)) ? next_value(third_value(args))->val : ds_create(NULL);
	ds_setsubstr(args->val, convert_to_int(DS_BODY(next_value(args)->val)), 
							convert_to_int(DS_BODY(third_value(args)->val)),
							rep
				);
	ds_append(pstack->state->buffer, args->val);
}

/// .replace(start len [subst])
// same as \em[replace] but operates on the current execution buffer
static void bi_preplace(chunk *current, interpreter *i, value *args)
{
	dstring *rep = third_value(args) ? third_value(args)->val : ds_create(NULL);
	ds_setsubstr(pstack->state->buffer, 
							convert_to_int(DS_BODY(args->val)), 
							convert_to_int(DS_BODY(next_value(args)->val)),
							rep
				);
}

/// repeat(arg count)
// \returns \em[arg] repeated \em[count] times; null if \em[count] is zero
// \error[negative counter (count) for 'repeat']
static void bi_repeat(chunk *current, interpreter *i, value *args)
{
	int iter = convert_to_int(DS_BODY(next_value(args)->val));

	if(iter < 0)
		recover_error("negative counter (%d) for 'repeat'", iter);
	while(iter--)
		ds_append(pstack->state->buffer, args->val);
}

static int str_bound_f(char **str, void *sep)
{
	return strchr(sep, *((*str)++)) != NULL;
}

static void *never_end = NULL;

static int au_bound_f(char **str, void *au)
{
	int res;
	int inch;
	if(au == never_end)
	{
		if(**str)
		{
			(*str)++;
			return 0;
		}
		return 1;
	}
	inch = **str;
	if((res = au_match_input(au, &inch)) == ar_rejected)
		return 1; 
	if(inch < 0 && need_feeding(au) && **str)
		(*str)++;
	return (!**str || as_is_in_final(au));
}

static char *g_split(char *buffer, void *fs, void *rs, 
					int (*bound_f)(char **iter, void *sep), 
					int (*add_f)(dstring *fld, interpreter *i, void *data),
					interpreter *i,
					void *data, int skipfld)
{
	dstring *fld = ds_create(NULL);
	dstring *null = ds_create(NULL);
	int flag = 0;
	char *iter = buffer, *new_iter = iter;

	for(;;)
	{
		flag = bound_f(&new_iter, rs);
		new_iter = iter;
		if(!flag && !bound_f(&new_iter, fs))
		{
			if(iter != new_iter)
				ds_appendch(fld, *iter);
		}
		else
		{
			if(!skipfld)
			{
				if(iter != new_iter)
					ds_appendch(fld, *iter);
			}
			if(!add_f(fld, i, data))
			{
				if(!flag)
					break;
			}
			ds_assign(fld, null);
			if(flag)
			{
				if(*new_iter)
					new_iter++;
				break;
			}
		}
		iter = new_iter;
	}
	return new_iter;
}

static int var_add_f(dstring *fld, interpreter *i, void *arg)
{
	value **varnames = arg;
	variable *var;
	
	if(!*varnames)
		return 0;
	if(!ds_isempty((*varnames)->val))
	{
		var = safe_get_var(i, DS_BODY((*varnames)->val), 1);
		ds_assign(var->vals->val, fld); 
	}
	*varnames = list_next(*varnames);
	return *varnames != NULL;
}

typedef struct sto_add_data
{
	storage *sto;
	int cnt;
} sto_add_data;

static int sto_add_f(dstring *fld, interpreter *i, void *arg)
{
	sto_add_data *sad = arg;
	dstring *key = ds_fromint(sad->cnt++, 10, 0);
	if(sad->sto->class->check(i, sad->sto, key))
		sad->sto->class->put(i, sad->sto, key, fld);
	else
		sad->sto->class->add(i, sad->sto, key, fld);
	return 1;
}

static void fill_variables_with_null(interpreter *i, value *varnames)
{
	dstring *null = ds_create(NULL);
	variable *var;
	
	for(; varnames; varnames = list_next(varnames))
	{
		if(!ds_isempty(varnames->val))
		{
			var = safe_get_var(i, DS_BODY(varnames->val), 1);
			ds_assign(var->vals->val, null);
		}
	}
}

/// split(arg fs rs var1 ...)
// splits \em[args] into fields delimited by one of characters in \em[fs].
// Splitting ends when the list of variables is over or when a character
// from \em[rs] is met. If a \em[var] is null, when a field is processed but
// not stored.
// \note This function does not strip leading separators, so two consecutive 
// characters will result in an empty field.
// \returns the rest of \em[arg]
static void bi_split(chunk *current, interpreter *i, value *args)
{
	char *fs = DS_BODY(next_value(args)->val);
	char *rs = DS_BODY(third_value(args)->val);
	char *iter = DS_BODY(args->val);
	value *varnames = next_value(third_value(args));

	iter = g_split(iter, fs, rs, str_bound_f, var_add_f, i, &varnames, 1);
	fill_variables_with_null(i, varnames);
	ds_appendstr(pstack->state->buffer, iter);
}

/// stosplit(args fs rs storage)
// like \em[split] but stores fields into \em[storage] with field numbers
// as keys. If a key is found in \em[storage], its value is overwritten
static void bi_stosplit(chunk *current, interpreter *i, value *args)
{
	sto_add_data sad = {NULL, 0};
	char *fs = DS_BODY(next_value(args)->val);
	char *rs = DS_BODY(third_value(args)->val);
	char *iter = DS_BODY(args->val);

	if(third_value(third_value(args)))
		sad.cnt = convert_to_int(DS_BODY(third_value(third_value(args))->val));
	
	sad.sto = find_storage(i, DS_BODY(next_value(third_value(args))->val));
	iter = g_split(iter, fs, rs, str_bound_f, sto_add_f, i, &sad, 1);
	ds_appendstr(pstack->state->buffer, iter);
}

static autoseq *get_aseq_for_ausplit(interpreter *i, char *name)
{
	autoseqdef *asd;
	if(!strcmp(name, "<none>"))
		return never_end;
	asd = find_asd(i, name);
	return asd->seq;
}

/// ausplit(limit arg auf aur var ...)
// like \em[split] but uses automata sequence to detect field/record
// boundaries. Note that automata should recognize not the boundaries, but
// fields/records itself. If \em[limit] is null, then a delimiting character
// is included into a field.
static void bi_ausplit(chunk *current, interpreter *i, value *args)
{
	autoseq *fs = get_aseq_for_ausplit(i, DS_BODY(third_value(args)->val));
	autoseq *rs = get_aseq_for_ausplit(i, DS_BODY(next_value(third_value(args))->val));
	char *iter = DS_BODY(next_value(args)->val);
	value *varnames = third_value(third_value(args));

	iter = g_split(iter, fs, rs, au_bound_f, var_add_f, i, &varnames, !ds_isempty(args->val));
	fill_variables_with_null(i, varnames);
	ds_appendstr(pstack->state->buffer, iter);
}

/// austosplit(limit arg auf aus storage)
// like \em[ausplit] but uses \em[storage] to keep parsed fields
static void bi_austosplit(chunk *current, interpreter *i, value *args)
{
	sto_add_data sad = {NULL, 0};
	autoseq *fs = get_aseq_for_ausplit(i, DS_BODY(third_value(args)->val));
	autoseq *rs = get_aseq_for_ausplit(i, DS_BODY(next_value(third_value(args))->val));
	char *iter = DS_BODY(next_value(args)->val);

	if(next_value(third_value(third_value(args))))
		sad.cnt = convert_to_int(DS_BODY(next_value(third_value(third_value(args)))->val));
	
	sad.sto = find_storage(i, DS_BODY(third_value(third_value(args))->val));
	iter = g_split(iter, fs, rs, au_bound_f, sto_add_f, i, &sad, !ds_isempty(args->val));
	ds_appendstr(pstack->state->buffer, iter);
}

/// toksplit(fmt val varcnt var1 ...)
// splits \em[val] into tokens like those used in NSL itself: integral
// values, characters, names and strings. 
// The types of fields are given in \em[fmt] in C scanf manner, via
// %-sequences. A space is treated as an arbitrary sequence of spaces
// comments. All other characters in \em[fmt] should be present in
// \em2[val][.] 
// 
// Resulting tokens are put into \em2[vari][...] if it is not empty;
// otherwise eaten (as with \em[split] ). The number of successfully
// processed fields (including those not really assigned) 
// is put into \em[varcnt] if it is not empty
// 
// The valid %-codes are:
// \item \em[%i] -- an integral value, i. e. an integer or a character
// enclosed in apostrophes
// \item \em[%c] -- a single character without any delimiters, may contain
// escape sequences.
// \item \em[%s] -- a string. A delimiter is the first character of the
// string.
// \item \em[%n] -- a name 
// \item \em[%N] -- a plain name (not containing \em[:] )
// \item \em[%#] -- a sequence of spaces and '#' comments (not put
// into \em[vars] )
// \item \em[%%] -- % character
// \returns[the rest of val that cannot be parsed or if the format line is ended]
static void bi_toksplit(chunk *current, interpreter *i, value *args)
{
	char *fmt_line = DS_BODY(args->val);
	char *dataline = DS_BODY(next_value(args)->val);
	char *val, *next;
	int ival;
	value *vars = next_value(third_value(args));
	while(*fmt_line && *dataline)
	{
		if(isspace(*fmt_line))
		{
			skip_spaces_nc(dataline, &dataline);
			fmt_line++;
		}
		else if(*fmt_line == '%')
		{
			switch(*++fmt_line)
			{
				case 'i':
				{
					int sign = 1;
					if(*dataline == '-')
					{
						sign = -sign;
						dataline++;
					}
					ival = sign * parse_num_token(dataline, &next);
					if(dataline != next)
						var_add_f(ds_fromint(ival, 10, 0), i, &vars);
					break;
				}
				case 'c':
					var_add_f(ds_createch(parse_single_char(dataline, &next)), i, &vars);
					break;
				case 'n':
					val = parse_id(dataline, &next);
					if(dataline != next)
						var_add_f(ds_create(val), i, &vars);
					break;
				case 'N':
					val = parse_plain_id(dataline, &next);
					if(dataline != next)
						var_add_f(ds_create(val), i, &vars);
					break;
				case 's':
					val = parse_string(dataline, &next);
					if(dataline != next)
						var_add_f(ds_create(val), i, &vars);
					break;
				case '#':
					skip_spaces(dataline, &dataline);
					fmt_line++;
					continue;
				case '%':
					next = dataline + (*dataline == '%');
					break;
				default:
					recover_error("invalid specifier: %%%c", *fmt_line);
			}
			fmt_line++;
			if(dataline != next)
				dataline = next;
			else
				break;
		}
		else if(*fmt_line == *dataline)
		{
			fmt_line++;
			dataline++;
		}
		else break;
	}
	if(!ds_isempty(third_value(args)->val))
	{
		variable *var = safe_get_var(i, DS_BODY(third_value(args)->val), 1);
		ds_assign(var->vals->val, ds_fromint(list_length(next_value(third_value(args))) - list_length(vars), 10, 0));
	}
	ds_appendstr(pstack->state->buffer, dataline);
}


/// span(arg delims)
// skips a prefix of \em[arg] which contains solely of characters from
// \em[delims]
// \returns the rest of \em[arg]
static void bi_span(chunk *current, interpreter *i, value *args)
{
	int n = strspn(DS_BODY(args->val), DS_BODY(next_value(args)->val));
	ds_appendstr(pstack->state->buffer, DS_BODY(args->val) + n);
}
	
/// reverse(arg)
// \returns reversed \em[arg]
static void bi_reverse(chunk *current, interpreter *i, value *args)
{
	ds_append(pstack->state->buffer, ds_reversip(args->val));
}

/// pad(arg length [pad-char])
// pads an \em[arg] to a specific \em[length] using \em[pad-char] as a
// filler (space by default). No truncation occurs if \em[arg] already
// exceeds \em[length]
//
// If \em[lenggth] is positive, \em[arg] is right-justified, otherwise 
// -- left-justified.
// \returns padded string
static void bi_pad(chunk *current, interpreter *i, value *args)
{
	int l = convert_to_int(DS_BODY(next_value(args)->val));
	int ch = third_value(args) ? *DS_BODY(third_value(args)->val) : ' ';
	int src_len = ds_length(args->val);
	if(l < 0)
	{
		l = -l;
		ds_append(pstack->state->buffer, args->val);
		for(; src_len < l; src_len++)
			ds_appendch(pstack->state->buffer, ch);
	}
	else
	{
		for(; src_len < l; src_len++)
			ds_appendch(pstack->state->buffer, ch);
		ds_append(pstack->state->buffer, args->val);
	}
}

static int pattern_find(int basech, int pattch, void *arg)
{
	int case_fold = arg != NULL;
	switch(pattch)
	{
		case '\a':
			return 0;
		default:
			return case_fold ? tolower(pattch) != tolower(basech) : pattch != basech;
	}
}

/// substit(arg substr repl [flags])
// Substitutes \rm[repl] for occurences of \em[substr] in \em[arg]. Flags
// may be used to control the behaviour:
// \item \em[nocase] case is ignored when looking for a substring
// \item \em[first] only the first occurence of \em[substr] is replaced
// \item \em[pattern] uses a simple pattern technique when looking for and
// substituting, namely \em[\\a] in \em[substr] means 'any character', 
// \em[\\a] in \em[repl] -- 'the character found at the corresponding
// position'; \em[\\v] -- the whole found substring.
// \endlist
// \note Pattern metacharacters should not be escaped twice -- they are
// really BEL and VT characters. No escaping mechanism is present, since
// literal usage of those characters should be rare enough.
// \returns \em[arg] with substitutions made
static void bi_substit(chunk *current, interpreter *i, value *args)
{
	int case_fold = 0;
	int use_pattern = 0;
	ds_predicate_t pred = NULL;
	int offset = 0, prev_offset = 0;
	int first_only = 0;
	char *src_ptr, *dest_ptr;
	
	if(next_value(third_value(args)))
	{
		case_fold = strstr(DS_BODY(next_value(third_value(args))->val), "nocase") != NULL;
		use_pattern = strstr(DS_BODY(next_value(third_value(args))->val), "pattern") != NULL;
		first_only = strstr(DS_BODY(next_value(third_value(args))->val), "first") != NULL;
	}
	if(use_pattern)
		pred = pattern_find;
	else if(case_fold)
		pred = ds_p_casefold;
	for(;;)
	{
		offset = ds_find(args->val, offset, next_value(args)->val, pred, case_fold ? &case_fold : NULL);
		if(offset < 0)
			break;
		ds_append(pstack->state->buffer, ds_substr(args->val, prev_offset, offset - prev_offset));
		src_ptr = DS_BODY(args->val) + offset;
		dest_ptr = DS_BODY(third_value(args)->val);
		for(; *dest_ptr; dest_ptr++, src_ptr++)
		{
			if(use_pattern)
			{
				if(*dest_ptr == '\a')
					ds_appendch(pstack->state->buffer, *src_ptr);
				else if(*dest_ptr == '\v')
					ds_append(pstack->state->buffer, ds_substr(args->val, offset, ds_length(next_value(args)->val)));
				else ds_appendch(pstack->state->buffer, *dest_ptr);
			}
			else
				ds_appendch(pstack->state->buffer, *dest_ptr);
		}
		offset += ds_length(next_value(args)->val);
		prev_offset = offset;
		if(first_only)
			break;
	}
	ds_append(pstack->state->buffer, ds_substr(args->val, prev_offset, ds_length(args->val) - prev_offset));
}
	
/// expand(arg [storage [delims]])
// expands variables in \em2[arg][.] The syntax here is slightly different
// from the standard: namely, '$$' means literal $; computed variable names
// aren't allowed; tag references aren't allowed too. If \em[storage] is
// present and not empty, not the real variables are used for expansion, but items from
// \em[storage]
// \returns an expanded \em[arg]
static void bi_expand(chunk *current, interpreter *i, value *args)
{
	char *iter;
	char *name;
	enum delim_types { del_startvar, del_openpar, del_closepar, del_delim, del_escape };
	const char *delims = next_value(args) && third_value(args) ? DS_BODY(third_value(args)->val)
															   : "$()$$";
	dstring *v;
	storage *s = NULL;
	
	if(strlen(delims) < 4)
		recover_error("delimiter string too short");
	if(next_value(args) && !ds_isempty(next_value(args)->val))
		s = find_storage(i, DS_BODY(next_value(args)->val));
	for(iter = DS_BODY(args->val); *iter; iter++)
	{
		if(*iter != delims[del_startvar])
		{
			if(delims[del_escape] != delims[del_startvar] && *iter == delims[del_escape])
				ds_appendch(pstack->state->buffer, *++iter);
			else
				ds_appendch(pstack->state->buffer, *iter);
		}
		else
		{
			iter++;
			if(delims[del_escape] == delims[del_startvar] && *iter == delims[del_startvar])
			{
				ds_appendch(pstack->state->buffer, delims[del_startvar]);
				continue;
			}
			else if(delims[del_openpar] != delims[del_startvar] && *iter == delims[del_openpar])
			{
				dstring *tmp = ds_create(NULL);
				for(iter++; *iter && *iter != delims[del_closepar]; iter++)
					ds_appendch(tmp, *iter);
				if(!*iter)
					recover_error("'%c' expected", delims[del_closepar]);
				name = DS_BODY(tmp);
			}
			else
			{
				name = parse_id(iter, &iter);
				if((delims[del_delim] == delims[del_startvar]) || (*iter != delims[del_delim]))
				   iter--;
			}
			if(s)
				ds_append(pstack->state->buffer, s->class->get(i, s, ds_create(name)));
			else
			{
				v = get_variable(i, 0, name);
				if(!v)
					recover_error("variable '%s' undefined", name);
				ds_append(pstack->state->buffer, v);
			}
		}
	}
}

instruction_src *lookup_instruction(interpreter *i, char *name)
{
	instruction_src *iter;

	for(iter = i->builtins; iter->name; iter++)
	{
		if(!strcmp(iter->name, name))
			return iter;
	}
	return NULL;
}

declaration *lookup_declaration(interpreter *i, char *name)
{
	declaration *d = NULL;
	static declaration std_declaration = { NULL, "", 1, 0, INT_MAX, -1, 0};

	if(name)
		d = list_find(i->decls, name, (list_predicate_t)decl_find);
	return d ? d : &std_declaration;
}

/// call(proc-name [arg ...])
// calls a procedure \em[proc-name] with \em2[arg][s] in the current
// execution environment. \em[proc-name] may be either a user-defined
// procedure or a builtin
// \returns the result of calling \em[proc-name]; null if \em[proc-name]
// doesn't normally return a value
static void bi_call(chunk *current, interpreter *i, value *args)
{
	instruction_src *in = NULL;
	declaration *dcl;
	char *pname = DS_BODY(args->val);

	if(*pname != ':' || pname[1] != ':')
		in = lookup_instruction(i, pname);
	else
		pname += 2;
	dcl = lookup_declaration(i, pname);
	if(dcl->minarg > list_length(args) - 1)
		recover_error("too few arguments for '%s'", pname);
	if(dcl->maxarg < list_length(args) - 1)
		recover_error("too many arguments for '%s'", pname);
	if(in)
		in->action(current, i, list_next(args));
	else
	{
		shared_int *si = list_find(i->procs, pname, (list_predicate_t)si_find_exact);
		procedure *proc;
		if(!si)
			recover_error("undefined procedure name: '%s'", pname);
		check_si_perm(i, si);
		proc = si->obj;
		eval_proc(proc->callback ? i : proc->sh.owner, proc->body, si->intname, proc->pool, copy_values(next_value(args)));
	}
}

static void bi_joiner(chunk *current, interpreter *i, value *args)
{
	ds_unfix(pstack->state->buffer);
	pstack->state->buffer = ds_fix(args->val);
}

/// .(proc [arg...])
// like \em[call] but the result replaces the contents of the current
// execution buffer
static void bi_pcall(chunk *current, interpreter *i, value *args)
{
	execstack *es = ALLOC(execstack);
	value *nv, *nv1;
	static chunk joiner;

	joiner.type = arg_builtin;
	joiner.x.func.f.bi = bi_joiner;

	es->buffer = ds_fix(ds_create(NULL));
	es->curfunc = &joiner;
	nv = ALLOC(value);
	nv1 = ALLOC(value);
	nv->val = ds_fix(ds_copy(pstack->state->buffer));
	nv1->val = ds_fix(args->val);
	pstack->state = list_add(pstack->state, es);
	nv = list_add(list_next(args), nv);
	bi_call(current, i, list_add(nv, nv1));
	ds_unfix(nv->val);
	ds_unfix(nv1->val);
	FREE(nv);
	FREE(nv1);
}

static table_t stream_classes[] = {{"basic", &basic_stream_class}, 
									{"script", &script_stream_class},
									{"html", &html_stream_class},
									{"xml", &html_stream_class},
									{"sgml", &html_stream_class},
									{"pipe", &pipe_stream_class},
									{"incgi", &incgi_stream_class},
									{"outcgi", &outcgi_stream_class},
									{NULL}
								}; 
								

/// defstream(name extname type [writeable])
// defines a stream with \em[name] of \em[type] and links it to an external
// entity \em[extname] (what depends on the type of the stream). 
// If \em[name] is null or '*', a stream is assigned a unique name.
// If \em[writeable] is present and true, the stream can be written to.
// \note Whether it is possible to read from a writeable stream depends on
// the kind of stream
// \returns the name of the stream
static int mod_stream_find(mod_stream *ms, char *key)
{
	return !strcmp(ms->cls->name, key);
}

void register_mod_stream(interpreter *i, stream_class *cls)
{
	mod_stream *nms;
	if(list_find(i->mod_streams, cls->name, (list_predicate_t)mod_stream_find))
		recover_error("stream class \'%s\' already defined", cls->name);
	nms = ALLOC(mod_stream);
	nms->cls = cls;
	i->mod_streams = list_add(i->mod_streams, nms);
}

static void bi_defstream(chunk *current, interpreter *i, value *args)
{
	stream_class *cls = lookup_table(stream_classes, DS_BODY(third_value(args)->val));
	stream *s;
	int writeable = next_value(third_value(args)) && !ds_isempty(next_value(third_value(args))->val);
	char *name = DS_BODY(args->val);

	if(list_find(i->streams, DS_BODY(args->val), (list_predicate_t)si_find_exact))
		recover_error("stream '%s' already defined", DS_BODY(args->val));
	if(!cls)
	{
		mod_stream *ms = list_find(i->mod_streams, DS_BODY(third_value(args)->val), (list_predicate_t)mod_stream_find);
		if(!ms)
			recover_error("unknown stream class: %s", DS_BODY(third_value(args)->val));
		cls = ms->cls;
	}
	if(!*name || *name == '*')
		name = unique_name();
	s = cls->open(i, name, DS_BODY(next_value(args)->val), writeable);
	add_shared_int(i, init_shared_stream(i, s, name));
	ds_appendstr(pstack->state->buffer, name);
}

/// rmstream(stream)
// This function is not completely equivalent to \em[release stream],
// since it checks whether the \em[stream] is the current input or output
// stream, and if so, clears it, so that there were not invalid pointers
static void bi_rmstream(chunk *current, interpreter *i, value *args)
{
	stream *s = find_stream(i, DS_BODY(args->val));
	if(i->instream && i->instream == s)
		i->instream = NULL;
	if(i->outstream && i->outstream == s)
		i->outstream = NULL;
	i->streams = list_remove(i->streams, DS_BODY(args->val), (list_predicate_t)si_find_exact,
							(list_destroy_t)destroy_shared_int, sizeof(shared_int));
}

/// instream([stream])
// if \em[stream] is present, sets the current input stream to \em[stream];
// if null, sets the current input stream to none
// \returns the name of the current input stream, or null if none
static void bi_instream(chunk *current, interpreter *i, value *args)
{
	shared_int *osi = i->instream ? list_find(i->streams, i->instream, (list_predicate_t)si_find_byso) : NULL;
	char *old = osi ? osi->intname : "";

	if(args)
	{
		if(ds_isempty(args->val))
			i->instream = NULL;
		else
			i->instream = find_stream(i, DS_BODY(args->val));
	}
	ds_appendstr(pstack->state->buffer, old);
}

/// outstream([stream])
// if \em[stream] is present, sets the current output stream to \em[stream]
// if null, sets the current input stream to none
// \returns the name of the current output stream, or null if none
static void bi_outstream(chunk *current, interpreter *i, value *args)
{
	shared_int *osi = i->outstream ? list_find(i->streams, i->outstream, (list_predicate_t)si_find_byso) : NULL;
	char *old = osi ? osi->intname : "";

	if(args)
	{
		if(ds_isempty(args->val))
			i->outstream = NULL;
		else
			i->outstream = find_stream(i, DS_BODY(args->val));
	}
	ds_appendstr(pstack->state->buffer, old);
}

/// defproc(name body)
// defines a procedure in the current environment with \em[name] and
// \em[body]. Note there is no way to specify named parameter aliases here
// (see SYNTAX OVERVIEW below)
static chunk *parse_toplevel(interpreter *i, char *start, char **end);
static void bi_defproc(chunk *current, interpreter *i, value *args)
{
	procedure *np;

	np = ALLOC(procedure);
	if(list_find(i->procs, DS_BODY(args->val), (list_predicate_t)si_find_exact))
		recover_error("procedure '%s' already defined", DS_BODY(args->val));
	ds_appendstr(next_value(args)->val, "}");
	np->body = parse_toplevel(i, DS_BODY(next_value(args)->val), NULL);
	add_shared_int(i, init_shared_proc(i, np, DS_BODY(args->val)));
}

/// defprocin(subint name body)
// like \em[defproc] but operates on the environment \em[subint] no matter
// whether it has \em[defproc] allowed
static void bi_defprocin(chunk *current, interpreter *i, value *args)
{
	interpreter *subint = find_subint(i, DS_BODY(args->val));
	bi_defproc(current, subint, list_next(args));
}

/// isproc(name)
// \returns true if a procedure \em[name] is defined in the current
// environment
static void bi_isproc(chunk *current, interpreter *i, value *args)
{
	if(quiet_find_shared_int(i, "proc", DS_BODY(args->val)))
		ds_appendstr(pstack->state->buffer, "true");
}

/// iscallback(subint name)
// \returns true if a procedure \em[name] within \em[subint] is a callback
// \note this function is provided primarily for security reasons -- 
// to ensure that a called function wouldn't hurt the caller. It seems to be
// no use to check the environment's own procedures.
static void bi_iscallback(chunk *current, interpreter *i, value *args)
{
	interpreter *sub = ds_isempty(args->val) ? i : find_subint(i, DS_BODY(args->val));
	procedure *p = quiet_find_shared_int(sub, "proc", DS_BODY(next_value(args)->val));
	if(p && p->callback)
		ds_appendstr(pstack->state->buffer, "true");
}

/// getenv(name)
// \returns the value of an environment variable \em[name] or null if it is
// not defined
static void bi_getenv(chunk *current, interpreter *i, value *args)
{
	char *env = getenv(DS_BODY(args->val));
	if(env)
		ds_appendstr(pstack->state->buffer, env);
}

/// setenv(name value)
// sets an environment variable \em[name] to \em2[value][,] probably
// overwriting the existing one
// \error[setenv 'name' to 'value' failed]
static void bi_setenv(chunk *current, interpreter *i, value *args)
{
	if(setenv(DS_BODY(args->val), DS_BODY(next_value(args)->val), 1))
		recover_error("setenv '%s' to '%s' failed!", DS_BODY(args->val), DS_BODY(next_value(args)->val));
}

#ifndef USE_SQL

/// sql(subcommand [arg])
// An interface to SQL. Refer to \ref[sl_sql(1)] for more details
static void bi_nosql(chunk *current, interpreter *i, value *args)
{
	if(!strcmp(DS_BODY(args->val), "feature") ||
	   !strcmp(DS_BODY(args->val), "engine"))
		return;
	recover_error("This build has no support for SQL");
}

#endif

/// feature(name)
// \returns true if a feature \em[name] is supported by the interpreter.
// \note Now it just checks for the presence and availability of builtins,
// but other types of features may be added in future
static void bi_feature(chunk *current, interpreter *i, value *args)
{
	if(lookup_instruction(i, DS_BODY(args->val)))
		ds_appendstr(pstack->state->buffer, "true");
}

/// version([module])
// \returns the version of the interpreter or of the \em[module]
static void bi_version(chunk *current, interpreter *i, value *args)
{
	if(!args || ds_isempty(args->val))
		ds_appendstr(pstack->state->buffer, VERSION);
	else if(!strcmp(DS_BODY(args->val), "libutils"))
		ds_appendstr(pstack->state->buffer, (char *)libutils_version());
	else
		ds_appendstr(pstack->state->buffer, (char *)module_version(DS_BODY(args->val)));
}

/// builddate([module])
// \returns the build date of the interpreter or of the \em[module] in the
// format of "MMM DD YYYY" like __DATE__ macro in C
static void bi_builddate(chunk *current, interpreter *i, value *args)
{
	if(!args || ds_isempty(args->val))
		ds_appendstr(pstack->state->buffer, __DATE__);
	else
		ds_appendstr(pstack->state->buffer, (char *)module_date(DS_BODY(args->val)));
}

/// pid()
// \returns a PID of the interpreter process
static void bi_pid(chunk *current, interpreter *i, value *args)
{
	ds_append(pstack->state->buffer, ds_fromint(getpid(), 10, 0));
}

/// id([subint])
// \returns a unique id of the environment \em[subint] or of the current
// environment if \em[subint] is omitted
static void bi_id(chunk *current, interpreter *i, value *args)
{
	int id = args ? (int)find_subint(i, DS_BODY(args->val)) : (int)i;
	ds_append(pstack->state->buffer, ds_fromint(id, 10, 0));
}

/// caller([type])
// \returns an identification of the environment that called the current
// procedure. If \em[type] is omitted or is 'subint', the internal name of
// the caller in the current environment is returned (or null if the caller
// is not subordinate to the current environment); if \em[type] is 'id', the
// unique id of the called is returned; if \em[type] is 'extname', the
// exported name of the caller is returned, or null if it is not exported.
static void bi_caller(chunk *current, interpreter *i, value *args)
{
	interpreter *where;
	if(!list_next(pstack))
		return;
	where = ((procstack *)list_next(pstack))->where;
	if(!args || !strcmp(DS_BODY(args->val), "subint"))
	{
		shared_int *caller;
		caller = list_find(i->subints, where, (list_predicate_t)si_find_byso);
		if(caller)
			ds_appendstr(pstack->state->buffer, caller->intname);
	}
	else if(!strcmp(DS_BODY(args->val), "id"))
		ds_append(pstack->state->buffer, ds_fromint((int)where, 10, 0));
	else if(!strcmp(DS_BODY(args->val), "extname"))
	{
		interpreter *caller = list_find(all_ints, where, (list_predicate_t)find_byref);
		if(caller && caller->sh.extnames)
			ds_appendstr(pstack->state->buffer, caller->sh.extnames->extname);
	}
}

/// sysvar(name [value]
// gets or sets an internal interpreter variable
//
// The following variables are defined:
// \item mem:squee_interval
// \item mem:squee_long
// \item mem:maxreserved
// \item mem:maxkeep
// \endlist
// \note those variables are global so you'd better restrict \em[sysvar] for
// unsecure environments
typedef dstring *(*sysvar_handler_t)(dstring* val);
 
#define SYSCARDVAR(name, varname, fullname) \
static dstring *name(dstring *val) \
{ \
	if(val) \
	{ \
		int nval = convert_to_int(DS_BODY(val)); \
		if(nval <= 0) \
			recover_error("invalid value for " fullname ": %d", nval); \
		varname = nval; \
	} \
	return ds_fromint(varname, 10, 0); \
}

SYSCARDVAR(sv_msi, squee_interval, "mem:squee_interval")
SYSCARDVAR(sv_msl, squee_long_string, "mem:squee_long")
SYSCARDVAR(sv_mmr, ds_reserved_mem_limit, "mem:maxreserved")
SYSCARDVAR(sv_mmk, ds_keep_limit, "mem:maxkeep") 

static dstring *sv_location(dstring *val)
{
	if(val)
		recover_error("env:location is read-only");
	return ds_create(nsl_location);
}

static void bi_sysvar(chunk *current, interpreter *i, value *args)
{
	static table_t sysvars[] = { {"mem:squee_interval", sv_msi},
								 {"mem:squee_long", sv_msl},
								 {"mem:maxreserved", sv_mmr},
								 {"mem:maxkeep", sv_mmk},
								 {"env:location", sv_location},
								 {NULL, NULL}
	};
	sysvar_handler_t sv = lookup_table(sysvars, DS_BODY(args->val));
	if(!sv)
		recover_error("no such sysvar: %s", DS_BODY(args->val));
	ds_append(pstack->state->buffer, sv(next_value(args) ? next_value(args)->val : NULL));
}

/// \section[SYNTAX OVERVIEW]

/// Source file
//
// A source file consists of a series of procedure definitions and 
// directives.
// A directive starts with '.' if it occurs outside any procedure
// definition, and with '..' elsewhere
// \note this is ugly. I am thinking on changing '.' to '!'
//

/// Procedure
// 
// Procedures may be of three kinds: plain procedures, tag handlers and 
// character filter. The latter two are used when processing input stream
// and cannot be called directly.
// The definition of a plain proc starts with its header, followed by its
// body. The header consists of an optional qualifier, the procedure name 
// and an optional parameter declaration. 
//
// An optional qualifier may be either \em[:void:] to indicate that a function returns 
// no meaningful value and can therefore be used only at the top level; or \em[:pure:] indicating
// that a function produces no side effects and that it may and will be executed at parse-time if
// all its arguments are constant.
//
// The procedure name may end in \em[*] indicating a \em[callback] function. A callback function 
// is executed within the execution environment of a caller, not in its native environment.
// 
// The parameter declaration has the following syntax:
//
// (\em[argname1] \em[argname2] ... \em[optargnamei=value] ... // \em3[[][...][\]])
//
// The declaration may be empty as well, meaning no arguments allowed. If
// the declaration is present, during parse-time \em2[argname]s are
// translated into 'def \em[argnamei] = arg\em3[(][i][)]', so that reading
// access might be available by name, not by index. The number of arguments
// is also checked when entering the procedure. Arguments followed by \em[=]
// are \em[optional] -- if they are not present, the default \em[value] is
// assigned. No required arguments may follow optional ones. A \em[...]
// means unlimited maximal number of arguments (all unnamed actual
// parameters are accessible thru their indices only)
// \note parameter names behave exactly like local variable, i. e. they
// cannot be overriden by following \em2[def][s]; when setting the
// parameter, the actual parameter accessible thru \em[arg] does not change,
// only its named alias. If a parameter is optional and is not present, its
// default value is assigned to the alias only, not to the parameter itself
//
// A \em[;] may also be present in the argument list. It marks the argument
// which is considered the last one if a second form of the procedure call
// is used (see below)
//
// The definitions of tag handlers/filters start with 'tag' or 'char' resp.
// followed by a glob-like pattern followed by the body
//
// The body is a sequence of procedure calls surrounded by {}

/// Top-level procedure call
// 
// Procedure calls at top-level may look either as normal procedure calls
// (see below) or as following:
// 
// \em[procedure-name arg1 ... argn ; ]
//
// \note ';' may be missing; then the \em[default] number of arguments for a
// given procedure is assumed. For some procedures such number is unlimited, 
// so you'd better always end such calls with ';'. 
// \note a newline is \em[not] a terminator!
//
// If a \em[procedure-name] starts with \em2[::][,] the rest of it is
// interpreted as a user-defined procedure even if a builtin with the same
// name exists. This holds
// true for inner procedure calls and for indirect calls via \em[call] as well.

/// Arguments & chunks
// 
// Each argument is a series of \em[chunks] ended with a space. 
// The value of an argument is concatenation of values of all its chunks.
// Chunks may be of ... kinds (yielded values are in parentheses):
// \item
// \em[single character] (the character itself). A character may be an
// escape sequence too.
// \item
// \em[single-quoted string] (the same). Escape sequences are allowed.
// \item
// \em[text] (the same). It has the following syntax:
// <<\em[delim-id] \em[spaces] text \em2[delim-id][>>.] Spaces in the beginning
// of the text are eaten. Escape sequences are not recognized.
// \item
// \em[parameter] It starts with '$' following by a digit or '#'. 
// When '#' is given, the value is the number of parameters of the current
// procedure. If '0' is given, the value is the contents of the working
// buffer. Elsewhere, the value is the \em2[n][th] parameter of the current
// procedure.
// \item
// \em[variable] It starts with '$' followed by variable name. If the name
// is not an identifier or may produce an identifier with the following
// text, parentheses are used to delimit the name. If the variable is not
// defined, an error occurs.
// \em3[$((][expr][))] is equivalent to \em2[var(expr)][.] 
// \item 
// \em[tag] Its syntax is \em3[$\[][tagname][\]] where \em[tagname] is the
// same as an argument for \em2[tag][.] If the current chunk is not defined or
// has no such parameter, an error occurs.
// \em3[$\[(][expr][)\]] is equivalent to \em[tag(expr)]
// \item
// \em[double-quoted string.] It is considered as a sequence of chunks ended
// with a closing quote.
// \item
// \em[bracketed string.] It is a sequence of chunks ended with a closing
// bracket. It is much like a dq-string, but since opening and closing
// characters differ, bracketed strings can be nested.
// \item
// \em[bracketed quoted string.] It is a \em[single-quoted string] enclosed
// in brackets. It is equivalent to \em[expand(single-quoted-str)]
// \item
// \em[backquoted string.] `string` is equivalent to \em[eval('string')]
// \item
// \em[block] is in all respects similar to a procedure body. In fact, blocks
// may be regarded as unnamed procedures. Though local variables of an outer
// block/procedure are accessible within an inner one.
// \endlist

static chunk *parse_variable(interpreter *i, char *start, char **end)
{
	int basic = *start == '(' ? ')' : ']';
	chunk *nc, *temp;
	start++;
	if(*start == '(')
	{
		start++;
		temp = parse_argument(i, paren_fc, start, &start);
		start++;
		nc = ALLOC(chunk);
		nc->type = arg_builtin;
		if(!lookup_instruction(i, basic == ']' ? "tag" : "var"))
		{
			recover_error("$%c()%c is not allowed, since %s is prohibited", 
							basic == ']' ? '[' : '(', 
							basic == ']' ? ']' : ')', 
							basic == ']' ? "tag" : "var"
						);
		}
		nc->x.func.f.bi = basic == ']' ? bi_tag : bi_var;
		nc->x.func.args = ALLOC(argument);
		nc->x.func.args->arg = temp;
	}
	else
	{
		char *s = strchr(start, basic);
		if(!s)
			recover_error("missing '%c'", basic);
		nc = ALLOC(chunk);
		nc->type = arg_variable;
		nc->x.var.is_tag = basic == ']';
		nc->x.var.name = xcalloc(s - start + 1, 1);
		memcpy(nc->x.var.name, start, s - start);
		start = s;
	}
	if(*start != basic)
		recover_error("missing '%c'", basic);
	if(end)
		*end = start + 1;
	return nc;
}

chunk *optimize_funcall(interpreter *i, chunk *src)
{
	argument *arg;
	chunk *ch;
	dstring *res;
	int lno = src->lno;
	char *file = src->file;
	
	for(arg = src->x.func.args; arg; arg = list_next(arg))
	{
		for(ch = arg->arg; ch; ch = list_next(ch))
		{
			if(ch->type != arg_literal)
				return src;
		}
	}
	ch = ALLOC(chunk);
	ch->lno = src->lno;
	ch->file = src->file;
	ch->type = arg_builtin;
	ch->x.func.f.bi = bi_return;
	ch->x.func.args = ALLOC(argument);
	ch->x.func.args->arg = src;
	src = ch;
	res = eval_script(i, src, NULL, NULL, NULL);
	ch = ALLOC(chunk);
	ch->lno = lno;
	ch->file = file;
	ch->type = arg_literal;
	ch->x.literal = res ? res : ds_fix(ds_create(NULL));
	list_delete(src, (list_destroy_t)destroy_chunk, sizeof(chunk));
	return ch;
}

argument *parse_arguments(interpreter *i, char *start, char **end);
chunk *parse_argument(interpreter *i, fence_condition fc, char *start, char **end)
{
	chunk *arg = NULL;
	chunk *nc = NULL, *temp;
	dstring *lit = NULL;
	instruction_src *op;
	declaration *decl;
	
	for(;;)
	{
		if(fc(*start))
		{
			if(lit)
			{
				temp = ALLOC(chunk);
				bind_chunk(temp);
				temp->type = arg_literal;
				temp->x.literal = lit;
				arg = list_append(arg, temp);
			}
			if(end)
				*end = start;
			return arg;
		}
		if(!*start)
			recover_error("unexpected EOF");
		switch(*start)
		{
			case '{':
				nc = ALLOC(chunk);
				bind_chunk(nc);
				nc->type = arg_embedded;
				nc->x.func.f.emb = parse_toplevel(i, start + 1, &start);
				nc->x.func.args = NULL;
				break;
			case '\'':
				nc = ALLOC(chunk);
				bind_chunk(nc);
				nc->type = arg_literal;
				nc->x.literal = ds_fix(ds_create(parse_string(start, &start)));
				break;
			case '\"':
				nc = parse_argument(i, quote_fc, start + 1, &start);
				if(!nc)
				{
					nc = ALLOC(chunk);
					bind_chunk(nc);
					nc->type = arg_literal;
					nc->x.literal = ds_fix(ds_create(NULL));
				}
				start++;
				break;
			case '[':
				if(start[1] == '\'')
				{
					if(!lookup_instruction(i, "expand"))
						recover_error("['...'] is now allowed, since expand is prohibited");
					nc = create_bi_call(bi_expand, create_literal(ds_create(parse_string(start + 1, &start))), NULL);
					if(*start != ']')
						recover_error("']' expected");
					start++;
					break;
				}
				nc = parse_argument(i, bracket_fc, start + 1, &start);
				if(!nc)
				{
					nc = ALLOC(chunk);
					bind_chunk(nc);
					nc->type = arg_literal;
					nc->x.literal = ds_fix(ds_create(NULL));
				}
				else
				{
					temp = nc;
					nc = ALLOC(chunk);
					bind_chunk(nc);
					nc->type = arg_builtin;
					nc->x.func.f.bi = bi_ident;
					nc->x.func.args = ALLOC(argument);
					nc->x.func.args->arg = temp;
				}
				start++;
				break;
			case '`':
				if(!lookup_instruction(i, "eval"))
					recover_error("`...` is now allowed, since eval is prohibited");
				temp = parse_argument(i, bquote_fc, start + 1, &start);
				start++;
				nc = ALLOC(chunk);
				bind_chunk(nc);
				nc->type = arg_builtin;
				nc->x.func.f.bi = bi_eval;
				nc->x.func.args = ALLOC(argument);
				nc->x.func.args->arg = temp;
				break;
			case '$':
				start++;
				nc = ALLOC(chunk);
				bind_chunk(nc);
				if(isdigit(*start) || *start == '#' || *start == '^')
				{
					nc->type = arg_parameter;
					nc->x.paramno = *start == '#' ? param_count : 
									(*start == '^' ? switch_header : *start - '0');
					start++;
				}
				else
				{
					if(*start == '(' || *start == '[')
						nc = parse_variable(i, start, &start);
					else
					{
						nc->type = arg_variable;
						nc->x.var.is_tag = 0;
						nc->x.var.name = xstrdup(parse_id(start, &start));
					}
				}
				break;
			case '<':
				if(start[1] == '<')
				{
					dstring *delim = ds_create(parse_id(start + 2, &start));
					char *fin;
					skip_spaces_nc(start, &start);
					ds_appendstr(delim, ">>");
					fin = strstr(start, DS_BODY(delim));
					if(!fin)
						recover_error("missing %s", DS_BODY(delim));
					nc = ALLOC(chunk);
					bind_chunk(nc);
					nc->type = arg_literal;
					nc->x.literal = ds_fix(ds_create(NULL));
					if(*start == '\\')
						start++;
					while(start != fin)
						ds_appendch(nc->x.literal, *start++);
					start += ds_length(delim);
					break;
				}
			default:
			{
				if(isnamesym(*start))
				{
					char *fwd;
					char *id = parse_id(start, &fwd);
					if(*fwd == '(')
					{
						int l;
						nc = ALLOC(chunk);
						bind_chunk(nc);
						if(*id != ':' || id[1] != ':')
							op = lookup_instruction(i, id);
						else
						{
							op = NULL;
							id += 2;
						}
						if(!op)
							id = xstrdup(id);
						decl = lookup_declaration(i, id);
						if(!decl->returns_value)
							recover_error("'%s' does not return a value", decl->name);
						nc->x.func.args = parse_arguments(i, fwd + 1, &start);
						l = list_length(nc->x.func.args);
						if(l < decl->minarg)
							recover_error("missing arguments for '%s' (min. %d)", decl->name, decl->minarg);
						if(l > decl->maxarg)
							recover_error("too manu arguments for '%s' (max. %d)", decl->name, decl->maxarg);
						if(op)
						{
							nc->type = arg_builtin;
							nc->x.func.f.bi = op->action;
						}
						else
						{
							nc->type = arg_funcall;
							nc->x.func.f.name = id;
						}
						if(decl->optimizable)
							nc = optimize_funcall(i, nc);
						break;
					}
				}
				if(!lit)
					lit = ds_fix(ds_create(NULL));
				ds_appendch(lit, parse_single_char(start, &start));
			}
		}
		if(nc)
		{
			if(lit)
			{
				temp = ALLOC(chunk);
				bind_chunk(temp);
				temp->type = arg_literal;
				temp->x.literal = lit;
				lit = NULL;
				arg = list_append(arg, temp);
			}
			arg = list_append(arg, nc);
			nc = NULL;
		}
	}
}

argument *parse_arguments(interpreter *i, char *start, char **end)
{
	argument *args = NULL;
	argument *newarg;
	chunk *c;

	for(;;)
	{
		skip_spaces(start, &start);
		c = parse_argument(i, normal_fc, start, &start);
		if(c)
		{
			newarg = ALLOC(argument);
			newarg->arg = c;
			args = list_add(args, newarg);
		}
		if(*start == ')')
			break;
	}
	if(end)
		*end = start + 1;
	return list_reversip(args);
}

static void parse_directive(interpreter *i, char *start, char **end, int safe);

static chunk *simple_parse_toplevel(interpreter *i, char *start, char **end)
{
	chunk *toplevel = NULL;
	chunk *nc = NULL;
	char *id;
	instruction_src *op;
	declaration *decl;
	argument *as, *na;
	int l;

	for(;;)
	{
		skip_spaces(start, &start);
		if(!*start)
			recover_error("missing }");
		if(*start == '}')
			break;
		if(*start == ';')
		{
			start++;
			continue;
		}
		if(*start == '.' && start[1] == '.')
		{
			parse_directive(i, start + 1, &start, 1);
			continue;
		}
		if(*start == '{')
		{
			nc = ALLOC(chunk);
			bind_chunk(nc);
			nc->type = arg_embedded;
			nc->x.func.f.emb = parse_toplevel(i, start + 1, &start);
			nc->x.func.args = NULL;
			toplevel = list_add(toplevel, nc);
			continue;
		}
		id = parse_id(start, &start);
		if(*id != ':' || id[1] != ':')
			op = lookup_instruction(i, id);
		else
		{
			op = NULL;
			id += 2;
		}
		decl = lookup_declaration(i, id);
		if(!op)
			id = xstrdup(id);
		if(*start == '(')
			as = parse_arguments(i, start + 1, &start);
		else
		{
			int j;
			if(decl->defarg < 0)
				recover_error("expecting '(' after procedure name '%s'", id);
			for(j = decl->defarg, as = NULL; j; j--)
			{
				skip_spaces(start, &start);
				if(*start == ';' || *start == '}')
					break;
				nc = parse_argument(i, toplevel_fc, start, &start);
				na = ALLOC(argument);
				na->arg = nc;
				as = list_add(as, na);
			}
			as = list_reversip(as);
		}
		l = list_length(as);
		if(l < decl->minarg)
			recover_error("missing arguments for '%s' (min. %d)", decl->name, decl->minarg);
		if(l > decl->maxarg)
			recover_error("too many arguments for '%s' (max. %d)", decl->name, decl->maxarg);
		nc = ALLOC(chunk);
		bind_chunk(nc);
		if(op)
		{
			nc->type = arg_builtin;
			nc->x.func.f.bi = op->action;
			nc->x.func.args = as;
			if(decl->returns_value)
			{
				chunk *temp = nc;
				nc = ALLOC(chunk);
				bind_chunk(nc);
				nc->type = arg_builtin;
				nc->x.func.f.bi = bi_dummy;
				nc->x.func.args = ALLOC(argument);
				nc->x.func.args->arg = temp;
			}
		}
		else
		{
			nc->type = arg_funcall;
			nc->x.func.f.name = id;
			nc->x.func.args = as;
		}
		toplevel = list_add(toplevel, nc);
	}
	if(end)
		*end = start + 1;
	if(!toplevel)
	{
		toplevel = ALLOC(chunk);
		bind_chunk(toplevel);
		toplevel->file = set_current_filename(NULL);
		toplevel->type = arg_builtin;
		toplevel->x.func.f.bi = bi_return;
	}
	return list_reversip(toplevel);
}

static void add_branch(chunk *dest, argument *condition, chunk *loc)
{
	branch *br = ALLOC(branch);
	br->condition = condition;
	br->loc = loc;
	dest->x.s.ctl->branches = list_append(dest->x.s.ctl->branches, br);
}

static chunk *parse_toplevel(interpreter *i, char *start, char **end)
{
	chunk *src = simple_parse_toplevel(i, start, end);
	argument *nesting = NULL, *newtop;
	argument *temparg;
	int push, pop;
	chunk *base = src;

	for(; src; src = list_next(src))
	{
		push = pop = 0;
		if(src->type == arg_builtin)
		{
			temparg = src->x.func.args;
			if(src->x.func.f.bi == bi_if)
			{
				src->x.s.statement = s_if;
				src->type =  arg_statement;
				src->x.s.ctl = ALLOC(ctl_data);
				newtop = ALLOC(argument);
				newtop->arg = src;
				nesting = list_add(nesting, newtop);
				add_branch(src, temparg, list_next(src));
			}
			else if(src->x.func.f.bi == bi_else || src->x.func.f.bi == bi_elsif)
			{
				if(!nesting)
					recover_error("else/elsif without if");
				if(nesting->arg->x.s.statement == s_if)
				{
					add_branch(nesting->arg, temparg, list_next(src));
					src->x.func.f.bi = bi_jump_to_endif;
				}
				else if(nesting->arg->x.s.statement == s_foreach || nesting->arg->x.s.statement == s_while)
				{
					if(nesting->arg->x.s.ctl->on_break)
						recover_error("too many else after foreach/while");
					nesting->arg->x.s.ctl->on_break = list_next(src);
					src->x.func.f.bi = bi_continue;
				}
				else recover_error("else/elsif without if");
				src->x.func.args = NULL;
			}
			else if(src->x.func.f.bi == bi_endif)
			{
				if(!nesting || nesting->arg->x.s.statement != s_if)
					recover_error("endif without if");
				src->x.func.f.bi = bi_popdown_ctl;
				pop = 1;
			}
			else if(src->x.func.f.bi == bi_while)
			{
				src->x.s.statement = s_while;
				push = 1;
			}
			else if(src->x.func.f.bi == bi_foreach)
			{
				chunk *ch = ALLOC(chunk);
				ch = list_add(list_next(src), ch);
				list_add(ch, src);
				src = ch;
				src->x.s.statement = s_foreach;
				temparg = NULL;
				push = 1;
			}
			else if(src->x.func.f.bi == bi_loop)
			{
				if(!nesting || (nesting->arg->x.s.statement != s_while && nesting->arg->x.s.statement != s_foreach))
					recover_error("loop without while/foreach");
				if(nesting->arg->x.s.ctl->on_break)
				{
					chunk *temp = ALLOC(chunk);
					*temp = *src;
					src->type = arg_builtin;
					src->x.func.f.bi = bi_after_break;
					src->x.func.args = NULL;
					src->next = temp;
					src = temp;
				}
				nesting->arg->x.s.ctl->ctl_end = src;
				nesting = list_pop(nesting, NULL, sizeof(argument));
			}
			else if(src->x.func.f.bi == bi_switch)
			{
				src->x.s.statement = s_switch;
				push = 1;
			}
			else if(src->x.func.f.bi == bi_case)
			{
				argument *arg = ALLOC(argument), *arg2 = ALLOC(argument);
				if(!nesting || nesting->arg->x.s.statement != s_switch)
					recover_error("case/caserange without switch");
				arg->arg = ALLOC(chunk);
				arg->arg->type = arg_builtin;
				arg->arg->x.func.f.bi = bi_match;
				arg2 = ALLOC(argument);
				arg2->arg = ALLOC(chunk);
				arg2->arg->type = arg_parameter;
				arg2->arg->x.paramno = switch_header;
				arg->arg->x.func.args = list_add(arg2, src->x.func.args); 
				add_branch(nesting->arg, arg, list_next(src));
				src->x.func.f.bi = bi_dummy;
				src->x.func.args = NULL;
			}
			else if(src->x.func.f.bi == bi_caserange)
			{
				argument *arg = ALLOC(argument), *arg2 = ALLOC(argument);
				if(!nesting || nesting->arg->x.s.statement != s_switch)
					recover_error("case/caserange without switch");
				arg->arg = ALLOC(chunk);
				arg->arg->type = arg_builtin;
				arg->arg->x.func.f.bi = bi_between;
				arg2 = ALLOC(argument);
				arg2->arg = ALLOC(chunk);
				arg2->arg->type = arg_parameter;
				arg2->arg->x.paramno = switch_header;
				arg2 = list_add(list_next(src->x.func.args), arg2);
				arg->arg->x.func.args = list_add(arg2, src->x.func.args); 
				add_branch(nesting->arg, arg, list_next(src));
				src->x.func.f.bi = bi_dummy;
				src->x.func.args = NULL;
			}
			else if(src->x.func.f.bi == bi_when)
			{
				if(!nesting || nesting->arg->x.s.statement != s_switch)
					recover_error("when without switch");
				add_branch(nesting->arg, src->x.func.args, list_next(src));
				src->x.func.f.bi = bi_dummy;
				src->x.func.args = NULL;
			}
			else if(src->x.func.f.bi == bi_endsw)
			{
				if(!nesting || nesting->arg->x.s.statement != s_switch)
					recover_error("endsw without switch");
				src->x.func.f.bi = bi_popdown_ctl;
				pop = 1;
			}
			if(push)
			{
				src->type =  arg_statement;
				src->x.s.ctl = ALLOC(ctl_data);
				src->x.s.ctl->header = temparg;
				newtop = ALLOC(argument);
				newtop->arg = src;
				nesting = list_add(nesting, newtop);
			}
			else if(pop)
			{
				nesting->arg->x.s.ctl->ctl_end = list_next(src);
				nesting = list_pop(nesting, NULL, sizeof(argument));
			}
		}
	}
	if(nesting)
	{
		switch(nesting->arg->type)
		{
			case s_if:
				recover_error("endif without if");
				break;
			case s_switch:
				recover_error("switch without endsw");
				break;
			case s_while:
				recover_error("while without loop");
				break;
			case s_foreach:
				recover_error("foreach without loop");
				break;
		}
	}
	return base;
}

static chunk *parse_arg_list(interpreter *i, char *start, char **end, declaration *decl)
{
	char *id;
	chunk *arg_list = NULL;
	chunk *new_chunk;
	int req_optional = 0;
 	int min_arg = 0, max_arg = 0, def_arg = -1;
	
	for(;;)
	{
		skip_spaces(start, &start);
		if(*start == ';')
		{
			def_arg = max_arg;
			skip_spaces(start + 1, &start);
		}
		if(*start == ')')
			break;
		id = parse_id(start, &start);
		if(!*id)
			recover_error("syntax error");
		if(!strcmp(id, "..."))
		{
			max_arg = INT_MAX;
			if(*start == ';')
			{
				def_arg = max_arg;
				start++;
			}
			skip_spaces(start, &start);
			if(*start != ')')
				recover_error("')' expected");
			break;
		}
		skip_spaces(start, &start);
		if(*start != '=')
		{
			new_chunk = create_bi_call(bi_def, create_literal(ds_create(id)), 
										create_literal(ds_createch('=')), 
										create_parameter(max_arg + 1), NULL
										);
			if(req_optional)
				recover_error("'=' expected");
		}
		else
		{
			dstring *name = ds_create(id);
			skip_spaces(start + 1, &start);
			new_chunk = create_bi_call(bi_def, create_literal(name), 
										create_literal(ds_createch('=')),
										create_bi_call(bi_cond, 
											create_bi_call(bi_less,
												create_literal(ds_create("int")),
												create_parameter(param_count),
												create_literal(ds_fromint(max_arg + 1, 10, 0)),
												NULL
											),
											parse_argument(i, normal_fc, start, &start),
											create_parameter(max_arg + 1),
											NULL
										),
										NULL
									);
			req_optional = 1;
		}
		arg_list = list_append(arg_list, new_chunk);
		max_arg++;
		if(!req_optional)
			min_arg++;
	}
	decl->minarg = min_arg; 
	decl->maxarg = max_arg; 
	decl->defarg = def_arg; 
	*end = start + 1;
	return arg_list;
}

static shared_pool *cur_pool;

void parse_procedure(interpreter *i, char *start, char **end)
{
	char *id;
	shared_int **dest = &i->procs;
	procedure *np;
	chunk *arg_check = NULL;
	int callback = 0;
	declaration new_decl = *lookup_declaration(i, NULL);
	
	skip_spaces(start, &start);
	id = parse_id(start, &start);
	if(!strcmp(id, "tag"))
	{
		dest = &i->tags;
		skip_spaces(start, &start);
		id = parse_string(start, &start);
	}
	else if(!strcmp(id, "char"))
	{
		dest = &i->filters;
		skip_spaces(start, &start);
		id = parse_string(start, &start);
	}
	else if(!strcmp(id, ":void:"))
	{
		new_decl.returns_value = 0;
		skip_spaces(start, &start);
		id = parse_id(start, &start);
	}
	else if(!strcmp(id, ":pure:"))
	{
		new_decl.optimizable = 1;
		skip_spaces(start, &start);
		id = parse_id(start, &start);
	}
	id = xstrdup(id);
	new_decl.name = id;
	if(*start == '*')
	{
		callback = 1;
		start++;
	}
	skip_spaces(start, &start);
	if(*start == '(')
	{
		arg_check = parse_arg_list(i, start + 1, &start, &new_decl);
		skip_spaces(start, &start);
	}
	if(*start != '{')
		recover_error("'{' expected");
	add_declaration(i, &new_decl);
	np = ALLOC(procedure);
	np->body = list_append(arg_check, parse_toplevel(i, start + 1, end));
	np->pool = cur_pool;
	np->callback = callback;
	*dest = list_append(*dest, init_shared_proc(i, np, id));
}

/// \section[PREPROCESSING DIRECTIVES]
// \note Only .if.../.else/.endif and .env directives are allowed within
// procedure body


typedef void (*directive_t)(interpreter *i, char *start, char **end);

/// .include <name> or .include "name"
// switches to processing file \em[name]; after the end, the current file
// continues. The maximal depth of nesting is not limited.
// 
// If angle brackets are used, the script is searched for in SCRIPT_PATH, 
// otherwise in the current directory

void i_include(interpreter *i, char *start, char **end)
{
	int lno = set_line_no(1);
	char *ofn, *fname;
	char *path;
	char *body;
	if(*start != '<' && *start != '\"')
		recover_error("syntax error after '.include'");
	path = *start == '<' ? getenv(SCRIPT_PATH_ENV) : NULL;
	if(*start == '<')
		*start = '>';
	fname = parse_string(start, end);
	ofn = set_current_filename(xstrdup(fname));
	body = load_file(parse_string(start, end), path, "sc");
	parse_script(i, body);
	free(body);
	set_line_no(lno);
	set_current_filename(ofn);
}

/// .if expr ... .else ... .endif
// evaluates a NSL expression \em[expr] and if it is null skips until 
// matching .else/.endif. \em[expr] is limited by the end-of-line
void i_if(interpreter *i, char *start, char **end)
{
	chunk *ch = parse_argument(i, eol_fc, start, end);
	chunk *retch;
	dstring *val;
	retch = ALLOC(chunk);
	retch->type = arg_builtin;
	retch->x.func.f.bi = bi_return;
	retch->x.func.args = ALLOC(argument);
	retch->x.func.args->arg = ch;
	val = eval_script(i, retch, NULL, cur_pool, NULL);
	if(val)
	{
		ds_unfix(val);
		if(!ds_isempty(val))
			return;
	}
	skip_ifs(start, end, 1);
}

/// .ifdef proc-name
// is equivalent to .if isproc(proc-name)
void i_ifdef(interpreter *i, char *start, char **end)
{
	char *id = parse_id(start, end);
	if(!list_find(i->procs, id, (list_predicate_t)si_find_exact))
		skip_ifs(start, end, 1);
}

/// .ifndef proc-name
// is equivalent to .if not(isproc(proc-name))
void i_ifndef(interpreter *i, char *start, char **end)
{
	char *id = parse_id(start, end);
	if(list_find(i->procs, id, (list_predicate_t)si_find_exact))
		skip_ifs(start, end, 1);
}

/// .ifhas feature-name
// is equivalent to .if feature(feature-name)
void i_ifhas(interpreter *i, char *start, char **end)
{
	char *id = parse_id(start, end);
	if(!lookup_instruction(i, id))
		skip_ifs(start, end, 1);
}

/// .ifhasnt feature-name
// is equivalent to .if not(feature(feature-name))
void i_ifhasnt(interpreter *i, char *start, char **end)
{
	char *id = parse_id(start, end);
	if(!lookup_instruction(i, id))
		skip_ifs(start, end, 1);
}

void i_else(interpreter *i, char *start, char **end)
{
	skip_ifs(start, end, 0);
}

void i_endif(interpreter *i, char *start, char **end)
{
	*end = start;
}

/// .pool name or .pool "name"
// creates a shared variable pool \em[name]. If it exists already, nothing
// happens. Sets the current pool for the following procedure definitions to
// it (or sets no pool if \em[name] is null or 'none')
void i_pool(interpreter *i, char *start, char **end)
{
	char *name;
	if(*start == '\"' || *start == '\'')
		name = parse_string(start, end);
	else name = parse_id(start, end);
	if(!*name || !strcmp(name, "none"))
	{
		if(cur_pool)
			cur_pool->refcnt--;
		cur_pool = NULL;
	}
	else
	{
		shared_pool *np = list_find(all_pools, name, (list_predicate_t)pool_find);
		if(!np)
		{
			np = ALLOC(shared_pool);
			np->name = xstrdup(name);
			all_pools = list_add(all_pools, np);
		}
		if(cur_pool)
			cur_pool->refcnt--;
		cur_pool = np;
		cur_pool->refcnt++;
	}
}

/// .error error-msg
// generates an error \em[error-msg]
void i_error(interpreter *i, char *start, char **end)
{
	char *t;
	t = strchr(start, '\n');
	recover_error("User error: %*.*s", t - start, t - start, start);
}

/// .env name = value
// is equivalent to setenv(name value) at parse-time
//
// \em[name] and \em[value] may be either identifiers or strings limited by
// \"
void i_env(interpreter *i, char *start, char **end)
{
	char *name;
	char *val;
	if(!lookup_instruction(i, "setenv"))
		recover_error("setting env vars is prohibited");
	name = xstrdup(*start == '\"' ? parse_string(start, &start) : parse_id(start, &start));
	skip_spaces(start, &start);
	if(*start != '=')
		recover_error("missing '=' after '.env'");
	skip_spaces(start + 1, &start);
	val = *start == '\"' ? parse_string(start, end) : parse_id(start, end);
	setenv(name, val, 1);
	free(name);
}
	
/// .do { some-code }
// executes \em[some-code] at parse time
void i_do(interpreter *i, char *start, char **end)
{
	chunk *ch;
	skip_spaces(start, &start);
	if(*start != '{')
		recover_error("'{' expected after '.do'");
	ch = parse_toplevel(i, start + 1, end);
	discard_value(eval_script(i, ch, NULL, cur_pool, NULL)); 
}

/// .module <module-name>
// loads the module \em[module-name] and attaches it to the current
// environment. If module support is off, do nothing. If 'module' builtin is turned off,
// generates an \error[loading modules is prohibited]
void i_module(interpreter *i, char *start, char **end)
{
	if(!lookup_instruction(i, "module"))
		recover_error("loading modules is prohibited");
#ifdef USE_LTDL	
	if(*start != '<')
		recover_error("syntax error at .module");
	*start = '>';
	attach_module(i, parse_string(start, end));
#else
	*start = '>';
	parse_string(start, end);
#endif	
}

/// .pragma module-name pragma
void i_pragma(interpreter *i, char *start, char **end)
{
	char *id;
	if(!lookup_instruction(i, "pragma"))
		recover_error("pragmas are prohibited");
	skip_spaces(start, &start);	
	id = xstrdup(parse_id(start, &start));
	skip_spaces(start, &start);
	if(*start != '\"') recover_error("expected '\"' after '.pragma'");
#ifdef USE_LTDL
	module_pragma(i, id, parse_string(start, end));
#endif
	free(id);
}
/// .declare name declspec
// declares a procedure \em[name] according to \em[declspec]. Declspec has the following form:
// \em[minarg[-maxarg][/defarg][v|p]] 
//
// If \em[maxarg] is missing, it is considered equal to \em2[minarg][.]
// If \em[defarg] is missing, the parethesesless notation for \em[name] is not allowed.
// \em[maxarg] as well as \em[defarg] may be \em2[$][.] that indicates infinity.
//
// \em[v] means `void' showing that a function does not return a value; \em[p] means `pure', i. e.
// a function can be computed at parse-time provided its arguments are constant
void i_declare(interpreter *i, char *start, char **end)
{
	char *id = parse_id(start, &start);
	parse_declaration(i, id, 0, start, end);
}

void i_override(interpreter *i, char *start, char **end)
{
	char *id = parse_id(start, &start);
	parse_declaration(i, id, 1, start, end);
}

void i_line(interpreter *i, char *start, char **end)
{
	set_line_no(strtoul(start, end, 0));
}

static void parse_directive(interpreter *i, char *start, char **end, int safe)
{
	char *id;
	static table_t directives[] = { {"include", i_include}, 
									{"if", i_if},
									{"ifdef", i_ifdef},
									{"ifndef", i_ifndef},
									{"ifhas", i_ifhas},
									{"ifhasnt", i_ifhasnt},
									{"else", i_else},
									{"endif", i_endif},
									{"pool", i_pool},
									{"error", i_error},
									{"env", i_env},
									{"do", i_do},
									{"module", i_module},
									{"declare", i_declare},
									{"override", i_override},
									{"pragma", i_pragma},
									{"line", i_line},
									{NULL, NULL}
								  };
	static table_t safe_directives[] = {{"if", i_if}, {"ifdef", i_ifdef},
										{"ifndef", i_ifndef}, {"ifhas", i_ifhas},
										{"ifhasnt", i_ifhasnt}, {"else", i_else},
										{"endif", i_endif}, {"env", i_env},
										{"line", i_line}, 
										{NULL, NULL}};
	directive_t d;
	
	id = parse_id(start + 1, &start);
	skip_spaces(start, &start);
	d = lookup_table(safe ? safe_directives : directives, id);
	if(!d)
		recover_error("unknown directive: '%s'", id);
	d(i, start, end);
}

static void parse_script(interpreter *i, char *start)
{
	for(;;)
	{
		skip_spaces(start, &start);
		if(!*start)
			break;
		if(*start != '.')
			parse_procedure(i, start, &start);
		else
			parse_directive(i, start, &start, 0);
	}
}

static master_instruction_src builtins[] = {   {"", 0, bi_dummy, 0, 0, 0},
										{"-", 1, bi_ident, 1, 1, 1},
										{"tag", 1, bi_tag, 1, 1, 1},
										{"settag", 0, bi_settag, 2, 2, 2},
										{"settagchar", 0, bi_settagchar, 1, 1, 1},
										{"cleartag", 0, bi_cleartag, 0, 0, 0},
										{"rmtag", 0, bi_rmtag, 1, 1, 1},
										{"passtag", 0, bi_passtag, 1, 1, 1},
										{"var", 1, bi_var, 1, 1, 1},
										{"arg", 1, bi_arg, 1, 1, 1},
										{"passargs", 1, bi_passargs, 1, 1, 1},
#ifdef ENABLE_REF
										{"ref", 1, bi_ref, 1, 1, 1},
#endif
										{"pool", 0, bi_pool, 1, 1, 1},
										{"rmpool", 0, bi_rmpool, 1, 1, 1},
										{"using", 1, bi_using, 0, 1, 1},
										{"eval", 1, bi_eval, 1, 1, 1},
										{"void", 0, bi_dummy, 1, 1, 1},
										{"if", 0, bi_if, 1, 1, 1},
										{"endif", 0, bi_endif, 0, 0, 0},
										{"else", 0, bi_else, 0, 0, 0},
										{"elsif", 0, bi_elsif, 1, 1, 1},
										{"switch", 0, bi_switch, 1, 1, 1},
										{"case", 0, bi_case, 1, 1, 1},
										{"caserange", 0, bi_caserange, 2, 2, 2},
										{"when", 0, bi_when, 1, 1, 1},
										{"breaksw", 0, bi_breaksw, 0, 0, 0},
										{"endsw", 0, bi_endsw, 0, 0, 0},
										{"while", 0, bi_while, 0, 1, 1},
										{"foreach", 0, bi_foreach, 2, 2, 2},
										{"loop", 0, bi_loop, 0, 1, 0},
										{"break", 0, bi_break, 0, 0, 0},
										{"continue", 0, bi_continue, 0, 0, 0},
										{"return", 0, bi_return, 0, 1, 0},
										{".return", 0, bi_preturn, 0, 1, 0},
										{"once", 1, bi_once, 1, 1, 1},
										{"Exit", 0, bi_exit, 0, 0, 0},
										{"exit", 0, bi_realexit, 0, 1, 0},
										{"export", 0, bi_export, 3, 3, 3},
										{"embed", 0, bi_embed, 4, 4, 4},
										{"isexport", 1, bi_isexport, 2, 2, 2},
										{"import", 0, bi_import, 3, 3, 3},
										{"allow", 0, bi_allow, 3, INT_MAX, INT_MAX},
										{"deny", 0, bi_deny, 3, INT_MAX, INT_MAX},
										{"rename", 0, bi_rename, 3, 4, 3},
										{"listof", 1, bi_listof, 1, 2, 1},
										{"declare", 0, bi_declare, 2, 2, 2},
										{"override", 0, bi_override, 2, 2, 2},
										{"undeclare", 0, bi_undeclare, 1, 2, 1},
										{"declof", 1,  bi_declof, 1, 1, 1},
										{"release", 0, bi_release, 2, 2, 2},
										{"gensym", 1, bi_gensym, 0, 0, 0},
										{"load", 0, bi_load, 1, 1, 1},
										{"subload", 1, bi_subload, 2, 2, 2},
										{"module", 0, bi_module, 1, 1, 1},
										{"pragma", 0, bi_pragma, 2, 2, 2},
										{"callin", 1, bi_callin, 2, INT_MAX, INT_MAX},
										{"restrict", 0, bi_restrict, 2, INT_MAX, INT_MAX},
										{"revoke", 0, bi_revoke, 2, INT_MAX, INT_MAX},
										{"dispatch", 0, bi_dispatch, 0, 1, 0},
										{"prefix", 1, bi_prefix, 0, 1, 1},
										{"echo", 0, bi_echo, 1, INT_MAX, 1},
										{"error", 0, bi_error, 1, INT_MAX, 1},
										{"fatal-error", 0, bi_fatal_error, 1, INT_MAX, 1},
										{"pass-error", 0, bi_pass_error, 1, 1, 1},
										{"onerror", 1, bi_onerror, 0, 1, 1},
										{"cancel", 0, bi_cancel, 0, 1, 0},
										{"catch", 1, bi_catch, 1, 1, 1},
										{"throw", 0, bi_throw, 1, 1, 1},
										{"trap", 0, bi_trap, 2, 2, 2},
										{"stop", 0, bi_stop, 0, 0, 0},
										{"write", 0, bi_write, 1, INT_MAX, 1},
										{"writeto", 0, bi_writeto, 2, INT_MAX, 2},
										{"writetag", 0, bi_writetag, 0, INT_MAX, INT_MAX},
										{"writetagto", 0, bi_writetagto, 1, INT_MAX, INT_MAX},
										{"ioctl", 1, bi_ioctlin, 2, 2, 2},
										{"getpos", 1, bi_getpos, 1, 1, 1},
										{"setpos", 0, bi_setpos, 2, 2, 2},
										{"rewind", 0, bi_rewind, 1, 1, 1},
										{"read", 1, bi_read, 0, 1, 0},
										{"readfrom", 1, bi_readfrom, 1, 2, 1},
										{"def", 0, bi_def, 1, INT_MAX, INT_MAX},
										{"global", 0, bi_global, 1, INT_MAX, INT_MAX},
										{"globalin", 0, bi_globalin, 2, INT_MAX, INT_MAX},
										{"shared", 0, bi_shared, 1, INT_MAX, INT_MAX},
										{"undef", 0, bi_undef, 1, INT_MAX, INT_MAX},
										{"defined", 1, bi_defined, 1, 1, 1},
										{"isglobal", 1, bi_isglobal, 1, 1, 1},
										{"readonly", 0, bi_readonly, 1, INT_MAX, INT_MAX},
										{"set", 1, bi_set, 2, 3, 3},
										{"push", 1, bi_push, 1, 3, 1},
										{"pop", 1, bi_pop, 1, 1, 1},
										{"clear", 0, bi_clear, 1, 1, 1},
										{"depth", 1, bi_depth, 1, 1, 1},
										{"copy", 0, bi_copy, 2, 2, 2},
										{"exch", 0, bi_exch, 1, 1, 1},
										{"revstack", 0, bi_revstack, 1, 1, 1},
										{"iseof", 1, bi_iseof, 0, 0, 0},
										{"readline", 1, bi_readline, 0, 1, 0},
										{"isint", 1, bi_isint, 1, 1, 1, 1},
										{"add", 1, bi_add, 2, INT_MAX, 2, 1},
										{"incr", 1, bi_incr, 2, 2, 2},
										{"postincr", 1, bi_postincr, 2, 2, 2},
										{"mul", 1, bi_mul, 2, INT_MAX, 2, 1},
										{"sub", 1, bi_sub, 2, 2, 2, 1},
										{"decr", 1, bi_decr, 2, 2, 2},
										{"postdecr", 1, bi_postdecr, 2, 2, 2},
										{"div", 1, bi_div, 2, 2, 2, 1},
										{"rem", 1, bi_rem, 2, 2, 2, 1},
										{"neg", 1, bi_neg, 1, 1, 1, 1},
										{"abs", 1, bi_abs, 1, 1, 1, 1},
										{"bitand", 1, bi_bitand, 2, INT_MAX, 2, 1},
										{"bitor", 1, bi_bitor, 2, INT_MAX, 2, 1},
										{"bitxor", 1, bi_bitxor, 2, INT_MAX, 2, 1},
										{"bitinv", 1, bi_bitinv, 1, 1, 1, 1},
										{"bitshift", 1, bi_bitshift, 2, 2, 2, 1},
										{"int", 1, bi_int, 1, 1, 1, 1},
										{"hex", 1, bi_hex, 1, 1, 1, 1},
										{"dec", 1, bi_dec, 1, 1, 1, 1},
										{"oct", 1, bi_oct, 1, 1, 1, 1},
										{"uint8", 1, bi_uint8, 1, 1, 1, 1},
										{"uint16", 1, bi_uint16, 1, 1, 1, 1},
										{"eq", 1, bi_eq, 2, 3, 2, 1},
										{"eqi", 1, bi_eqi, 2, 2, 2, 1},
										{"neq", 1, bi_neq, 2, 3, 2, 1},
										{"less", 1, bi_less, 2, 3, 2, 1},
										{"greater", 1, bi_greater, 2, 3, 2, 1},
										{"le", 1, bi_le, 2, 3, 2, 1},
										{"ge", 1, bi_ge, 2, 3, 2, 1},
										{"expr", 1, bi_expr, 1, 1, 1, 1},
										{"match", 1, bi_match, 2, 2, 2, 1},
										{"matchi", 1, bi_matchi, 2, 2, 2, 1},
										{"between", 1, bi_between, 3, 4, 3, 1},
										{"and", 1, bi_and, 2, INT_MAX, 2, 1},
										{"or", 1, bi_or, 2, INT_MAX, 2, 1},
										{"not", 1, bi_not, 1, 1, 1, 1},
										{"cond", 1, bi_cond, 2, 3, 3, 1},
										{"length", 1, bi_length, 1, 1, 1, 1},
										{"ascii", 1, bi_ascii, 1, 1, 1, 1},
										{"char", 1, bi_char, 1, 1, 1, 1},
										{"trim", 1, bi_trim, 1, 1, 1, 1},
										{"substr", 1, bi_substr, 2, 3, 3, 1},
										{".substr", 1, bi_psubstr, 1, 2, 2},
										{"index", 1, bi_index, 2, 3, 2, 1},
										{"rindex", 1, bi_rindex, 2, 2, 2, 1},
										{"isprefix", 1, bi_isprefix, 2, 2, 2, 1},
										{"issuffix", 1, bi_issuffix, 2, 2, 2, 1},
										{"commonprefix", 1, bi_commonprefix, 2, 2, 2, 1},
										{"auindex", 1, bi_auindex, 2, 3, 3},
										{"auindexmax", 1, bi_auindexmax, 2, 3, 3},
										{"indexi", 1, bi_indexi, 2, 3, 2, 1},
										{"rindexi", 1, bi_rindexi, 2, 2, 2, 1},
										{"replace", 1, bi_replace, 3, 4, 4, 1},
										{".replace", 1, bi_preplace, 2, 3, 3},
										{"repeat", 1, bi_repeat, 2, 2, 2, 1},
										{"split", 1, bi_split, 4, INT_MAX, 4},
										{"stosplit", 1, bi_stosplit, 4, 5, 4},
										{"ausplit", 1, bi_ausplit, 5, INT_MAX, 5},
										{"austosplit", 1, bi_austosplit, 5, 6, 5},
										{"toksplit", 1, bi_toksplit, 4, INT_MAX, 4},
										{"span", 1, bi_span, 2, 2, 2, 1},
										{"reverse", 1, bi_reverse, 1, 1, 1, 1},
										{"pad", 1, bi_pad, 2, 3, 2, 1},
										{"substit", 1, bi_substit, 3, 4, 3, 1},
										{"expand", 1, bi_expand, 1, 3, 1},
										{"array", 1, bi_array, 0, INT_MAX, 1},
										{"table", 1, bi_table, 2, 2, 2},
										{"list", 1, bi_list, 0, INT_MAX, 1 },
										{"sequence", 1, bi_sequence, 1, 1, 1, 1},
										{"iota", 1, bi_iota, 1, 1, 1, 1},
										{"new", 1, bi_new, 1, INT_MAX, 1},
										{"sql", 1, bi_nosql, 1, 2, 2},
										{"local", 1, bi_local, 1, 1, 1},
										{"get", 1, bi_get, 2, 2, 2},
										{"check", 1, bi_check, 2, 2, 2},
										{"put", 0, bi_put, 3, 3, 3},
										{"insert", 0, bi_insert, 3, 3, 3},
										{"remove", 0, bi_remove, 2, 2, 2},
										{"sto.enum", 1, bi_sto_enum, 2, 3, 2},
										{"sto.break", 0, bi_sto_break, 2, 2, 2},
										{"apply", 1, bi_apply, 2, INT_MAX, 2},
										{"mark-remove", 0, bi_mark_remove, 2, 2, 2},
										{"unmark-remove", 0, bi_unmark_remove, 2, 2, 2},
										{"ismark-remove", 1, bi_ismark_remove, 2, 2, 2},
										{"cancel-remove", 0, bi_cancel_remove, 1, 1, 1},
										{"do-remove", 0, bi_do_remove, 1, 1, 1},
										{"count", 1, bi_count, 1, 1, 1},
										{"defautoseq", 1, bi_defautoseq, 2, 2, 2},
										{"rmautoseq", 0, bi_rmautoseq, 1, 1, 1},
										{"au.load", 0, bi_au_load, 1, 1, 1},
										{"au.new", 0, bi_au_new, 1, 1, 1},
										{"au.build", 0, bi_au_build, 2, 2, 2},
										{"translate", 1, bi_translate, 2, 2, 2},
										{"translatez", 1, bi_translatez, 2, 2, 2},
										{"itranslate", 1, bi_itranslate, 2, 2, 2},
										{"itranslatez", 1, bi_itranslatez, 2, 2, 2},
										{"ctranslate", 1, bi_ctranslate, 2, 2, 2},
										{"reset", 1, bi_reset, 1, 1, 1},
										{"state", 1, bi_state, 1, 1, 1},
										{"topstack", 1, bi_topstack, 1, 1, 1},
										{"mapstate", 1, bi_mapstate, 2, 2, 2},
										{"statename", 1, bi_statename, 2, 3, 3},
										{"isfinal", 1, bi_isfinal, 1, 1, 1},
										{"call", 1, bi_call, 1, INT_MAX, INT_MAX},
										{".", 1, bi_pcall, 1, INT_MAX, INT_MAX},
										{"instream", 1, bi_instream, 0, 1, 1},
										{"outstream", 1, bi_outstream, 0, 1, 1},
										{"rmstream", 0, bi_rmstream, 1, 1, 1},
										{"defstream", 1, bi_defstream, 3, 4, 4},
										{"defproc", 0, bi_defproc, 2, 2, 2},
										{"defprocin", 0, bi_defprocin, 3, 3, 3},
										{"isproc", 1, bi_isproc, 1, 1, 1},
										{"iscallback", 1, bi_iscallback, 2, 2, 2},
										{"getenv", 1, bi_getenv, 1, 1, 1},
										{"setenv", 0, bi_setenv, 2, 2, 2},
										{"trace", 0, bi_trace, 0, INT_MAX, 1},
										{"notrace", 0, bi_notrace, 0, INT_MAX, 1},
										{"feature", 1, bi_feature, 1, 1},
										{"version", 1, bi_version, 0, 1, 0},
										{"builddate", 1, bi_builddate, 0, 1, 0},
										{"pid", 1, bi_pid, 0, 0, 0, 1},
										{"id", 1, bi_id, 0, 1, 0},
										{"caller", 1, bi_caller, 0, 1, 0},
										{"sysvar", 1, bi_sysvar, 1, 2, 2},
										{NULL}
};

instruction_src *expand_builtins(instruction_src *src, int delta)
{
	instruction_src *dest = src, *iter;
	int count = 0;
	for(iter = dest; iter->name; iter++, count++)
		;
	dest = xrealloc(dest, (count + delta + 1) * sizeof(*dest));
	memset(dest + count, 0, (delta + 1) * sizeof(*dest));
	return dest;
}

void merge_builtins(interpreter *i,  master_instruction_src *add)
{
	instruction_src *dest = i->builtins, *iter1;
	master_instruction_src *iter;
	int count = 0, count1 = 0;
	if(dest)
	{
		for(iter1 = dest; iter1->name; iter1++, count++)
			;
	}
	for(iter = add; iter->name; iter++, count1++)
		add_declaration_master(i, iter);
	dest = xrealloc(dest, (count + count1 + 1) * sizeof(*dest));
	for(iter = add, iter1 = dest + count; iter->name; iter++, iter1++)
	{
		iter1->name = iter->name;
		iter1->action = iter->action;
	}
	iter1->name   = NULL;
	iter1->action = NULL;
	i->builtins = dest;
}

void override_declaration(interpreter *i, declaration *decl)
{
	declaration *nd = ALLOC(declaration);
	*nd = *decl;
	nd->name = xstrdup(nd->name);
	i->decls = list_add(i->decls, nd);
}

void add_declaration(interpreter *i, declaration *decl)
{
	if(!list_find(i->decls, decl->name, (list_predicate_t)decl_find))
		override_declaration(i, decl);
}

void add_declaration_master(interpreter *i, master_instruction_src *bi)
{
	declaration decl; 
	decl.name =  		 bi->name;
	decl.returns_value = bi->returns_value;
	decl.minarg = 		 bi->minarg;
	decl.maxarg = 		 bi->maxarg;
	decl.defarg = 		 bi->defarg;
	decl.optimizable = 	 bi->optimizable;
	add_declaration(i, &decl);
}

void parse_declaration(interpreter *i, char *name, int override, char *dclstr, char **end)
{
	declaration decl = *lookup_declaration(i, NULL);
	skip_spaces(dclstr, &dclstr);
	decl.name = name;
	decl.minarg = strtoul(dclstr, &dclstr, 10);
	if(*dclstr != '-')
		decl.maxarg = decl.minarg;
	else
	{
		dclstr++;
		if(*dclstr == '$')
		{
			decl.maxarg = INT_MAX;
			dclstr++;
		}
		else
			decl.maxarg = strtoul(dclstr, &dclstr, 10);
	}
	if(*dclstr != '/')
		decl.defarg = -1;
	else
	{
		dclstr++;
		if(*dclstr == '$')
		{
			decl.defarg = INT_MAX;
			dclstr++;
		}
		else
			decl.defarg = strtoul(dclstr, &dclstr, 10);
	}
	switch(*dclstr)
	{
		case 'v':
			decl.returns_value = 0;
			dclstr++;
			break;
		case 'p':
			decl.optimizable = 1;
			dclstr++;
			break;
		case '\0':
		case '\n':
		case ' ':
			break;
		default:
			recover_error("invalid declaration: ...%s", dclstr);
	}
	(override ? override_declaration : add_declaration)(i, &decl);
	if(end)
		*end = dclstr;
}

dstring *stringize_declaration(declaration *decl)
{
	dstring *ds = ds_fromint(decl->minarg, 10, 0);
	if(decl->maxarg != decl->minarg)
	{
		ds_appendch(ds, '-');
		if(decl->maxarg == INT_MAX)
			ds_appendch(ds, '$');
		else
			ds_append(ds, ds_fromint(decl->maxarg, 10, 0));
	}
	if(decl->defarg >= 0)
	{
		ds_appendch(ds, '/');
		if(decl->defarg == INT_MAX)
			ds_appendch(ds, '$');
		else
			ds_append(ds, ds_fromint(decl->defarg, 10, 0));
	}
	if(!decl->returns_value)
		ds_appendch(ds, 'v');
	else if(decl->optimizable)
		ds_appendch(ds, 'p');
	return ds;
}


static storage *ist_nocreate(interpreter *i, char *name, char *arg)
{
	recover_error("cannot create internal storage");
	return NULL; // never reached
}

static void ist_destroy(interpreter *i, storage *s)
{
	// dummy
}

static void ist_nomodify(interpreter *i, storage *s, dstring *key, dstring *val)
{
	// dummy
}

static void ist_noremove(interpreter *i, storage *s, dstring *key)
{
	// dummy
}

static void ist_enditer(interpreter *i, storage *s, void **iter)
{
	// dummy
}

static variable **pget_var_list(interpreter *i, storage *s)
{
	shared_int *si = list_find(i->stos, s, (list_predicate_t)si_find_byso);
	if(!si)
		fatal_error("internal error");
	if(!strcmp(si->intname, "LOCALS"))
		return &pstack->vars;
	if(!strcmp(si->intname, "GLOBALS"))
		return &i->globvars;
	if(!strcmp(si->intname, "SHAREDS"))
		return pstack->cur_pool ? &pstack->cur_pool->vars : NULL;
	fatal_error("internal error");
	return NULL; // never reached
}

static variable *get_var_list(interpreter *i, storage *s)
{
	variable **vp = pget_var_list(i, s);
	return vp ? *vp : NULL;
}

static int vs_check(interpreter *i, storage *s, dstring *key)
{
	variable *vars = get_var_list(i, s);
	return list_find(vars, DS_BODY(key), (list_predicate_t)var_find) != NULL;
}

static dstring *vs_get(interpreter *i, storage *s, dstring *key)
{
	variable *vars = get_var_list(i, s);
	variable *v = list_find(vars, DS_BODY(key), (list_predicate_t)var_find);
	if(!v)
		recover_error("no such variable: %s", DS_BODY(key));
	return v->vals->val;
}

static void vs_put(interpreter *i, storage *s, dstring *key, dstring *val)
{
	variable *vars = get_var_list(i, s);
	variable *v = list_find(vars, DS_BODY(key), (list_predicate_t)var_find);
	if(!v)
		recover_error("no such variable: %s", DS_BODY(key));
	if(v->readonly)
		recover_error("attempt to modify readonly variable %s", DS_BODY(key));
	ds_assign(v->vals->val, val);
}

static void vs_add(interpreter *i, storage *s, dstring *key, dstring *val)
{
	variable **pvars = pget_var_list(i, s);
	variable *v;
	if(!pvars)
		recover_error("no shared pool");
	if(list_find(*pvars, DS_BODY(key), (list_predicate_t)var_find))
		recover_error("variable already defined: %s", DS_BODY(key));
	v = ALLOC(variable);
	v->name = xstrdup(DS_BODY(key));
	v->vals = ALLOC(value);
	v->vals->val = ds_fix(val);
	*pvars = list_add(*pvars, v);
}

static void vs_remove(interpreter *i, storage *s, dstring *key)
{
	variable **pvars = pget_var_list(i, s);
	if(!pvars)
		recover_error("no shared pool");
	*pvars = list_remove(*pvars, DS_BODY(key), (list_predicate_t)var_find, (list_destroy_t)destroy_variable, sizeof(variable));
}

static dstring *vs_first(interpreter *i, storage *s, void **iter)
{
	variable *vars = get_var_list(i, s);
	if(!vars)
		return NULL;
	*iter = list_next(vars);
	return ds_create(vars->name);
}

static dstring *vs_next(interpreter *i, storage *s, void **iter)
{
	variable *var = *iter;
	if(!var)
		return NULL;
	*iter = list_next(var);
	return ds_create(var->name);
}

static int vs_numofitems(interpreter *i, storage *s)
{
	variable *vars = get_var_list(i, s);
	return list_length(vars);
}

char **program_argv = NULL;
int program_argc = 0;

static int avs_check(interpreter *i, storage *s, dstring *key)
{
	int no = convert_to_int(DS_BODY(key));
	return no >=  0 && no < program_argc;
}

static dstring *avs_get(interpreter *i, storage *s, dstring *key)
{
	if(!avs_check(i, s, key))
		return ds_create(NULL);
	return ds_create(program_argv[convert_to_int(DS_BODY(key))]);
}


static dstring *avs_first(interpreter *i, storage *s, void **iter)
{
	if(!program_argv || !*program_argv)
		return NULL;
	*iter = (void *)1;
	return ds_create("0");
}
	
static dstring *avs_next(interpreter *i, storage *s, void **iter)
{
	int count = (int)*iter;
	if(count == program_argc)
		return NULL;
	*iter = (void *)(count + 1);
	return ds_fromint(count, 10, 0);
}
	
static int avs_numofitems(interpreter *i, storage *s)
{
	return program_argc;
}
	
static void create_std_storage(interpreter *i, char *name, storage_class *cls)
{
	storage *ns = ALLOC(storage);
	ns->class = cls;
	add_shared_int(i, init_shared_storage(i, ns, name));
}
									
/// \section[PREDEFINED STORAGES]
// Each environment predefines a set of storages:

/// LOCALS, GLOBALS, SHAREDS
// Those storage provide access to lists of resp. local, global and shared
// (of the current pool) variables. Access to items of the storages is fully
// equivalent to accessing the variables itself. 
// \note There is no builtin to remove global and shared variables; this can
// be done via \em[remove GLOBALS ...] or \em[remove SHAREDS ...] resp. 

static storage_class vars_class = {ist_nocreate, ist_destroy,
									 vs_check, vs_get, vs_put,
									 vs_add, vs_remove,
									 vs_first, vs_next,
									 ist_enditer, vs_numofitems
								    };

/// ARGV
// This storage gives access to command-line options. This storage is
// read-only and silent in the sense that accessing to a missing argument
// yields null, not an error (though \em[check] will return true only for
// arguments actually present)

/// Other types of storages will be added
// \endlist

static storage_class ARGV_class = { ist_nocreate, ist_destroy, 
									avs_check, avs_get, ist_nomodify,
									ist_nomodify, ist_noremove,
									avs_first, avs_next,
									ist_enditer, avs_numofitems,
								  };
							  
									
// static storage_class si_class;

interpreter *load_interpreterx(char *fname, interpreter *parent, stream *ins, stream *outs)
{
	interpreter *ni = ALLOC(interpreter);
	char *body = load_file(fname, getenv(SCRIPT_PATH_ENV), "sc");
	int lno = set_line_no(1);
	char *ofile = set_current_filename(xstrdup(fname));
	instruction_src *iiter;
	
	merge_builtins(ni, builtins);
	reinit_interpreter(ni, ins, outs);
	cur_pool = NULL;
	if(parent)
	{
		for(iiter = ni->builtins; iiter->name; iiter++)
		{
			if(!lookup_instruction(parent, iiter->name))
				iiter->name = "";
		}
	}
	parse_script(ni, body);
	set_line_no(lno);
	set_current_filename(ofile);
	free(body);
	return ni;
}

void reinit_interpreter(interpreter *i, stream *ins, stream *outs)
{
	i->aseqs = init_shared_autoseq(i, ALLOC(autoseqdef), "");
	create_std_storage(i, "GLOBALS", &vars_class);
	create_std_storage(i, "LOCALS", &vars_class);
	create_std_storage(i, "SHAREDS", &vars_class);
	create_std_storage(i, "ARGV", &ARGV_class);
	if(ins)
	{
		i->streams = init_shared_stream(i, ins, "stdin");
		i->instream = ins;
	}
	if(outs)
	{
		shared_int *si = init_shared_stream(i, outs, "stdout");
		i->streams = i->streams ? list_add(i->streams, si) : si;
		i->outstream = outs;
	}
}
